#!/usr/bin/env python3
import os
import sys
from collections import defaultdict
from typing import List, Tuple, Iterable, Optional

from common.errors import printerr
from common.log import LogData
from common.statistics import median
from parser.base import Parser
from parser.client import ClientLogParser
from parser.iftop import IftopLogParser
from parser.latency import LatencyLogParser
from parser.network import NetworkLogParser
from parser.resources import ResourcesLogParser
from parser.server import ServerLogParser
from transformer.apply_base_time import ApplyBaseTimeFilter
from transformer.base import Transformer
from transformer.client import ClientLogFilter
from transformer.client_log_averages import ClientLogAveragesTransformer
from transformer.client_raw_avg_throughput import ClientRawAvgThroughputTransformer
from transformer.client_raw_latencies import ClientRawLatenciesTransformer
from transformer.client_raw_throughput import ClientRawThroughputTransformer
from transformer.cut_log import CutLogFilter
from transformer.latency import LatencyLogFilter
from transformer.output_log_averages import OutputLogAveragesTransformer
from transformer.output_log_throughput import OutputLogThroughputTransformer
from transformer.output_pickle import OutputPickleTransformer
from transformer.output_raw_avg_throughput import OutputRawAvgThroughputTransformer
from transformer.output_raw_latencies import OutputRawLatenciesTransformer
from transformer.output_raw_throughput import OutputRawThroughputTransformer
from transformer.print_warnings import PrintWarningsFilter
from transformer.resources import ResourceLogFilter
from transformer.server import ServerLogFilter
# Mass processing:
# mv */* . && find . -iname '*.DS_Store' -delete && \
# ../heterogeneous-smr/scripts/group-results.py . . rezk \
# readwrite-0,readwrite-0.5,readwrite-1,exists,increment,queue memory,sqlite data128,data1024 && \
# find . -type d -depth 1 -print0 | \
# xargs -0 -L1 -P8 -I {} \
# bash -c 'cd {} && echo {} && ../../heterogeneous-smr/scripts/parse-results.py'
from transformer.server_channel import ServerChannelLogFilter

AUTO_RANGE_SLACK_AT_END = 5


class LogError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


def setup_parsers() -> Iterable[Parser]:
    return [
        ClientLogParser(),
        IftopLogParser(),
        LatencyLogParser(),
        NetworkLogParser(),
        ResourcesLogParser(),
        ServerLogParser(),
    ]


def setup_transformers(path: str) -> Iterable[Transformer]:
    return [
        ApplyBaseTimeFilter(),
        ClientLogFilter(),
        LatencyLogFilter(),
        ResourceLogFilter(),
        ServerLogFilter(),
        ServerChannelLogFilter(),
        PrintWarningsFilter(),
        # FIXME Add transformer to check for required logs
        # OutputJSONTransformer(path),
    ]


def get_group_key(scenario: str) -> str:
    return scenario


def setup_group_transformers(scenario: str) -> Iterable[Transformer]:
    return [
        ClientLogAveragesTransformer(),
        ClientRawLatenciesTransformer(),
        ClientRawThroughputTransformer(),
        ClientRawAvgThroughputTransformer(),
        CutLogFilter(),
        # OutputJSONTransformer("parsed-" + scenario),
        OutputPickleTransformer("parsed-" + scenario),
        OutputLogAveragesTransformer(),
        OutputLogThroughputTransformer(),
        OutputRawLatenciesTransformer(),
        OutputRawThroughputTransformer(),
        OutputRawAvgThroughputTransformer(),
    ]


def walk_folders(folder_name: str, relative_time_range: Optional[Tuple[float, float]],
                 use_min_log_start: bool = False):
    # split into groups
    groups = defaultdict(lambda: [])
    for root, dirs, _ in os.walk(folder_name):
        for dirname in dirs:
            scenario, client_count = parse_folder_name(dirname)
            if scenario is None:
                printerr("Skipping folder {}".format(dirname))
                continue

            group = get_group_key(scenario)
            groups[group].append(dirname)
        break

    if not groups:
        printerr("No results found!")
        sys.exit(1)

    # process group by group
    effective_time_range = None  # type: Optional[Tuple[float, float]]
    for group, dirnames in groups.items():
        bundle = []
        for dirname in sorted(dirnames):
            scenario, client_count = parse_folder_name(dirname)
            log_bundle, used_time_range = analyze_folder(os.path.join(folder_name, dirname), scenario, client_count,
                                                         relative_time_range, use_min_log_start)
            bundle.extend(log_bundle)
            if effective_time_range is None:
                effective_time_range = used_time_range
            else:
                effective_time_range = (
                    max(used_time_range[0], effective_time_range[0]),
                    min(used_time_range[1], effective_time_range[1])
                )

        group_transformers = setup_group_transformers(group)
        for transformer in group_transformers:
            bundle = transformer.transform(bundle, effective_time_range[0], effective_time_range[1])

    print("Recommended time range: {} {}".format(effective_time_range[0], effective_time_range[1]))


def parse_folder_name(name: str) -> Tuple[Optional[str], int]:
    name_chunks = name.split("-refit-", 1)
    if len(name_chunks) < 2:
        return None, 0

    # expected structure: <scenario>-<clients>-<duration>
    chunks = name_chunks[1].rsplit("-", 2)
    scenario = chunks[0]

    try:
        client_count = int(chunks[-2])
    except ValueError:
        scenario += "-" + chunks[-2]
        client_count = 0
    return scenario, client_count


def analyze_folder(path: str, scenario: str, client_count: int,
                   relative_time_range: Optional[Tuple[float, float]] = None,
                   use_min_log_start: bool = False) -> \
        Tuple[Iterable[LogData], Tuple[float, float]]:
    parsers = setup_parsers()
    logs = parse_logs(parsers, path, scenario, client_count)
    log_base_time, auto_start_time, auto_end_time = auto_range(logs, use_min_log_start)
    if relative_time_range is None:
        print("Auto range {:.3} {:.3} for {}".format(auto_start_time, auto_end_time, path))
    for log in logs:
        log.set_base_time(log_base_time)
    try:
        start_time, end_time = merge_time_range(auto_start_time, auto_end_time, relative_time_range)
    except LogError:
        print("Error while processing {}".format(path))
        raise
    # print("Using range {} {}".format(start_time, end_time))

    transformers = setup_transformers(path)
    for transformer in transformers:
        logs = transformer.transform(logs, start_time, end_time)

    return logs, (start_time, end_time)


def parse_logs(parsers: Iterable[Parser], path: str, scenario: str, client_count: int) -> List[LogData]:
    logs = []  # type: List[LogData]
    for _, _, files in os.walk(path):
        for fn in files:
            log_path = os.path.join(path, fn)
            if not os.path.isfile(log_path):
                continue

            for parser in parsers:
                if not parser.can_handle(fn):
                    continue
                try:
                    log = parser.parse_file(log_path)
                except LogError:
                    printerr("Error while parsing {}".format(log_path))
                    raise
                log.set_scenario(scenario, client_count)
                logs.append(log)
                break
    return logs


def is_auto_range_log(log: LogData) -> bool:
    if "channel-micro" in log.scenario:
        return log.logtype == "server"
    return log.logtype == "client"


def auto_range(logs: List[LogData], use_min_log_start: bool = False) -> Tuple[float, float, float]:
    base = float("inf")
    start = 0
    end = -base
    merge_func = max
    if use_min_log_start:
        start = float("inf")
        merge_func = min

    for log in logs:
        if not is_auto_range_log(log):
            continue

        for entry in log.entries():
            base = min(base, entry["timestamp"])
            end = max(end, entry["timestamp"])

        if log.logtype == "client":
            log_start = find_start(log)
            if log_start is None:
                raise LogError("Failed to find start in log {}".format(log.source_path))
            start = merge_func(start, log_start)
        else:
            start = merge_func(start, base + 5)

    return base, start - base, end - AUTO_RANGE_SLACK_AT_END - base


def find_start(log: LogData) -> Optional[float]:
    data = [e["throughput"] for e in log.entries() if "throughput" in e]
    if not data:
        return None
    warm_throughput = median(data) * 0.9

    for e in log.entries():
        if "throughput" not in e:
            continue
        if e["throughput"] > warm_throughput:
            return e["timestamp"]

    return None


def merge_time_range(auto_start_time: float, auto_end_time: float,
                     relative_time_range: Optional[Tuple[float, float]]) -> Tuple[float, float]:
    if relative_time_range is not None:
        start_time, end_time = relative_time_range[0], relative_time_range[1]
    else:
        start_time, end_time = auto_start_time, auto_end_time
    if not (auto_start_time <= start_time < auto_end_time and auto_start_time < end_time <= auto_end_time):
        raise LogError("Requested time interval outside auto range interval. "
                       "Auto: [{};{}[, Requested: [{};{}[".format(auto_start_time, auto_end_time,
                                                                  start_time, end_time))
    if end_time < start_time + 10:
        raise LogError("Selected time interval is less than 10 seconds: "
                       "Interval: [{};{}[".format(start_time, end_time))
    return start_time, end_time


def run():
    import argparse
    parser = argparse.ArgumentParser(description='Parse, check and process results')
    parser.add_argument('--range', type=int, nargs=2, default=None,
                        help='Use range between start and end from results', metavar=('START', 'END'))
    parser.add_argument('--min-log-start', help="Use the minimal log start as range limit instead of the maximum. "
                                                "This is useful when not all clients are active at the same time",
                        action="store_true")

    args = parser.parse_args()
    relative_time_range = args.range
    use_min_log_start = args.min_log_start
    walk_folders(".", relative_time_range, use_min_log_start)


if __name__ == '__main__':
    run()
