import os
import re
from typing import Optional

from common.log import LogData
from .base import Parser


class ClientLogParser(Parser):
    @staticmethod
    def extract_id(fn: str) -> Optional[str]:
        match = re.match(r"(client-\d+)\.log", fn)
        if match is None:
            return None
        return match.group(1)

    def can_handle(self, fn: str) -> bool:
        return self.extract_id(fn) is not None

    def parse_file(self, path: str) -> LogData:
        ld = LogData("client", self.extract_id(os.path.basename(path)), path)

        with open(path, "r") as f:
            at_start = True
            timestamp = None
            for line in f:
                line = line.strip()
                if not line:
                    continue

                split_line = line.split(None, 1)
                if split_line[0].isdigit():
                    timestamp = float(int(split_line[0])) * 1e-6
                    # remove timestamp at line start
                    line = split_line[1]
                # keep last timestamp when encountering a line without

                if at_start:
                    if line.startswith("main [EVENT] BENCH: Startup delay"):
                        at_start = False
                        # ignore printed delay and use the log timestamp instead
                        ld.add_data(timestamp, started=1)
                    continue
                elif line.startswith("main [EVENT] BENCH: END: "):
                    break

                split_line = line.split(None, 3)
                if len(split_line) >= 4 and split_line[0].isdigit() \
                        and split_line[1].isdigit() and (split_line[2].isdigit() or split_line[2] == "NaN"):
                    # assume this is measurement data
                    idx = int(split_line[0])
                    throughput = int(split_line[1])
                    if split_line[2] == "NaN":
                        latency = min_latency = max_latency = -1
                    else:
                        latency = int(split_line[2])
                        remainder = split_line[3]
                        groups = re.match(r"\( *(\d+)/ *(\d+)\)", remainder).groups()
                        min_latency = int(groups[0])
                        max_latency = int(groups[1])
                    ld.add_data(timestamp, throughput=throughput, latency=latency, latency_min=min_latency,
                                latency_max=max_latency)
                else:
                    ld.add_warning(timestamp, line)

        return ld

# gibt's auch fuer server
# client-0-iftop.log
# client-0-network.log

# nur client
# client-0.latency
