import math

from common.log import LogData
from .base import Filter, PatternChecker, Rule, FatalLogIssue

CLIENT_PATTERN = r"CLNT\d+-1?\d "

IGNORE_LINES = [
    Rule(r"BENCH\d+ \[EVENT\] BENCH: OK", 1),
    Rule(r"BENCH\d+ \[EVENT\] BENCH: Start time: [0-9]+.[0-9]+s", 1),
    Rule(CLIENT_PATTERN + r"\[WARNG\] LIB(BYZ|SPD|STW)\[\d+\]: retry \{REQUEST\|\d+\|\[\d+\-\d+\]\} "
                          r"\(\d+ replies received\)", 0, True),
    Rule(CLIENT_PATTERN + r"\[DEBUG\] RPNET: Connecting to node \d+@.+", 0, True, True),
    Rule(CLIENT_PATTERN + r"\[WARNG\] CONWK\d+: Connection to \d+@.+ failed: "
                          r"java.net.ConnectException: Connection refused", 0, True, True),
    Rule(CLIENT_PATTERN + r"\[WARNG\] RPNET: Lost connection to \d+@.+", 0, True, True),
    Rule(CLIENT_PATTERN + r"\[WARNG\] CONWK\d+: Connection was closed \d+@.+", 0, True, True),
    Rule(CLIENT_PATTERN + r"\[WARNG\] CONWK\d+: Connection setup timed out \d+@.+", 0, True),
    Rule(r"hang \[ERROR\] HangDetector: Detected hang of " + CLIENT_PATTERN + r"for 1?\d\d\d\d\d\dus", 0, True),
    Rule(r"\[\d+\.\d\d\ds\]\[info\]\[gc\] GC\(\d+\) Pause Young \(Normal\) "
         r"\(G1 Evacuation Pause\) \d+M->\d+M\(\d+M\) \d?\d?\d.\d\d\dms", float("inf")),
    Rule(r"BENCH\d+ \[WARNG\] LOGC-0: Got 1?[0-9] entries; base \d+", float("inf")),
]


class ClientLogFilter(Filter):
    def can_handle(self, log: LogData) -> bool:
        return log.logtype == "client"

    def filter(self, log: LogData, start_time: float, end_time: float) -> LogData:
        out = log.derive()
        checker = PatternChecker(IGNORE_LINES, start_time, end_time)
        has_start = False
        data_points = 0

        for entry in log.entries():
            if "warning" in entry:
                if not checker.matches(entry["timestamp"], entry["warning"]):
                    out.add_entry(entry)
            else:
                out.add_entry(entry)
                # if "latency" in entry and entry["latency"] == -1 and start_time <= entry["timestamp"] < end_time:
                #     out.add_warning(entry["timestamp"], "Unexpected latency spike (> 1 second!)")
                if "started" in entry and entry["timestamp"] < start_time:
                    # ignore start confirmation if it is too late
                    has_start = True
                if "throughput" in entry and start_time <= entry["timestamp"] < end_time:
                    data_points += 1

        if not has_start:
            raise FatalLogIssue("No or too late log start")

        # fatal error if not enough data points during time range
        expected_data = int(math.ceil(end_time - start_time))
        if abs(expected_data - data_points) > 1:
            raise FatalLogIssue("Unexpected log length, got {} expected {}".format(data_points, expected_data))

        return out
