from common.log import LogData
from .base import Filter, PatternChecker, Rule, FatalLogIssue

REPLICA_PATTERN = r"RPLC[12]?[0-9]-[0-4] "

IGNORE_LINES = [
    Rule(REPLICA_PATTERN + r"\[EVENT\] EXCTR: Start time: [0-9]+.[0-9]+s", 1),
    Rule(REPLICA_PATTERN + r"\[EVENT\] EXCTR:  *\d+ @ \d+", float("inf")),
    Rule(REPLICA_PATTERN + r"\[EVENT\] EXCTR: change checkpoint-creation "
                           r"setting to \"(regular|lightweight)\"", 0, True),
    Rule(REPLICA_PATTERN + r"\[EVENT\] EXCTR: change update-creation setting to \"(disabled|enabled)\"", 0, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] EXECT: TODO: implement checkpoint fetching", float("inf")),
    Rule(REPLICA_PATTERN + r"\[EVENT\] ORDER: switch to REFIT(Steward|PBFT(Passive)?|Weave)Protocol \(view \d\)", 0,
         True),
    Rule(REPLICA_PATTERN + r"\[EVENT\] ORDER: \d is now the contact replica", 0, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] ORDER: No free slots in the agreement window. completed \d+/\d+", 0, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] TT: Update delay between \d and \d to \d+ ms", float("inf")),
    Rule(REPLICA_PATTERN + r"\[WARNG\] (EXCLI|CLINT\[\d\]): resend reply"
                           r" \{REPLY\|\d+\|\[\d+-[01]\]\|true\|\d\|\d+\|(READ|WRITE)\|\d+ bytes\}", 0, True),
    Rule(REPLICA_PATTERN + r"\[EVENT\] CLINT\[(-1|\d)\]: Configuration update"
                           r" \(send replies: (true|false) contactReplica: \d\)", 0, True),
    Rule(REPLICA_PATTERN + r"\[EVENT\] OCLINT: Configuration update", 0, True),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] RPNET: Connecting to node \d+@.+", 0, True, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] RPNET: Lost connection to (\d+@.+|null)", 0, True, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] CONWK\d+: Connection to \d+@.+ failed: "
                           r"java.net.ConnectException: Connection refused", 0, True, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] CONWK\d+: Dropping connection from \d+", 0, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] CONWK\d+: Connection was closed null", 0, True),
    Rule(REPLICA_PATTERN + r"\[WARNG\] CNCTN\[Socket\[addr=/.+,port=\d+,localport=\d+\]\]: Broken connection",
         0, False, True),
    Rule(r"hang \[ERROR\] HangDetector: Detected hang of " + REPLICA_PATTERN + r"for 1?\d\d\d\d\d\dus", 0, True),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] PRBSS: (Active leaders|"
                           r"Raw (replies|prepares \d|commits \d)) \[[0-9, \-]+\]", float("inf")),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] PRBSS: (Raw probe from client \d+ in \d+ with leader \d|"
                           r"Recommend leader \d with \d+|Last \d with \d+)", float("inf")),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] PRBSS: (Passive leader \d percentiles \[[0-9, \-]+\]|"
                           r"Recommend passive replicas \[\d\])", float("inf")),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] EXCTR: (Requesting fuzzy checkpoint start @\d+|Next offset: \d+|FCP took \d+ "
                           r"seqs|getting changelist for fuzzy collection|getting changes|clearing changes"
                           r"|\d+ms until \d+ objects of checkpoint data retrieved from application"
                           r"|Notifying about finished interval up to \d+)", float("inf")),
    Rule(REPLICA_PATTERN + r"\[DEBUG\] EXECT: Notified about finished collection while @ \d+", float("inf")),
    Rule(REPLICA_PATTERN + r"\[FCP\] FCP: (Fuzzy collection started @\d+|Finishing checkpoint @\d+"
                           r"|Collecting full checkpoint|Collection done after \d{4}ms"
                           r"|Serializing and broadcasting done for \d+ objects of checkpoint @ \d+ after \d{2,3}ms)",
         float("inf")),
    Rule(r"RPLC[0-3]-[0-4] \[WARNG\] CH-SRC: Setting channel group 4 to status (false|true)", 2),
    Rule(r"RPLC1[6-8]-0 \[WARNG\] CHKPT: Got cp", 2),
    Rule(r"\[\d+\.\d\d\ds\]\[info\]\[gc\] GC\(\d+\) Pause (Remark|Cleanup|Young "
         r"\((Normal|Concurrent Start|Prepare Mixed|Mixed)\) "
         r"\((G1 Evacuation Pause|GCLocker Initiated GC|G1 Humongous Allocation)\)) "
         r"\d+M->\d+M\(\d+M\) 1?\d?\d?\d.\d\d\dms", float("inf")),
    Rule(r"\[\d+\.\d\d\ds\]\[info\]\[gc\] GC\(\d+\) Concurrent Cycle( \d?\d?\d.\d\d\dms)?", float("inf")),
]

MUST_HAVE_LINES = [
    # Rule(REPLICA_PATTERN + r"\[EVENT\] CLINT: Configuration update \(send replies: (true|false)\)"),
]


class ServerLogFilter(Filter):
    def can_handle(self, log: LogData) -> bool:
        return log.logtype == "server"

    def filter(self, log: LogData, start_time: float, end_time: float) -> LogData:
        out = log.derive()
        ignores = PatternChecker(IGNORE_LINES, start_time, end_time)
        must_haves = PatternChecker(MUST_HAVE_LINES, start_time, end_time)
        has_start = False

        for entry in log.entries():
            if "warning" in entry:
                must_haves.matches(entry["timestamp"], entry["warning"])
                if not ignores.matches(entry["timestamp"], entry["warning"]):
                    out.add_entry(entry)
            else:
                out.add_entry(entry)
                if "started" in entry and entry["timestamp"] < start_time:
                    # ignore start confirmation if it is too late
                    has_start = True

        if not has_start:
            raise FatalLogIssue("No or too late log start")

        unmatched_rules = must_haves.unmatched_rules()
        if unmatched_rules:
            raise FatalLogIssue("Did not find expected patterns:\n    "
                                + "\n    ".join([str(r.pattern.pattern) for r in unmatched_rules]))

        return out
