package refit.agreement.pbft;

import java.nio.ByteBuffer;

import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTCommit;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTMessage;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrePrepare;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrepare;
import refit.config.REFITConfig;
import refit.message.*;
import refit.util.REFITBallotBox;

public class REFITPBFTCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private int PREPARED_THRESHOLD;
	private int COMMITTED_THRESHOLD;
	private boolean[] legitimateSenders;

	private long instanceID;
	private int viewID;

	public REFITPBFTPrePrepare prePrepare;
	public REFITOrderRequest message;
	private REFITBallotBox<Short, ByteBuffer, REFITPBFTMessage> prepares;
	private REFITBallotBox<Short, ByteBuffer, REFITPBFTCommit> commits;


	public REFITPBFTCertificate(int groupSize) {
		int[] weights = new int[groupSize];
		if (REFITConfig.WHEAT_WEIGHTS.length > 1) {
			if (REFITConfig.WHEAT_WEIGHTS.length != groupSize + 1) {
				throw new AssertionError("Broken order weights!");
			}
			PREPARED_THRESHOLD = REFITConfig.WHEAT_WEIGHTS[REFITConfig.WHEAT_WEIGHTS.length - 1];
			COMMITTED_THRESHOLD = PREPARED_THRESHOLD;
			System.arraycopy(REFITConfig.WHEAT_WEIGHTS, 0, weights, 0, groupSize);
		} else {
			// calculates ceil((n+f+1)/2)
			PREPARED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
			COMMITTED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
			weights = null;
		}
		if (groupSize < PREPARED_THRESHOLD || groupSize < COMMITTED_THRESHOLD) {
			throw new AssertionError(
					String.format("Group size %d is too small for the expected number of faults %d",
							groupSize, REFITConfig.FAULTS_TO_TOLERATE));
		}
		this.prepares = new REFITBallotBox<>(PREPARED_THRESHOLD, weights);
		this.commits = new REFITBallotBox<>(COMMITTED_THRESHOLD, weights);
	}


	@Override
	public String toString() {
		return "{" + viewID + " " + instanceID + ": " + isPrepared() + " " + isCommitted() + "}";
	}

	public void init(long instanceID, int viewID, boolean[] legitimateSenders) {
		this.instanceID = instanceID;
		this.viewID = viewID;
		this.prePrepare = null;
		this.message = null;
		this.prepared = false;
		this.committed = false;
		this.legitimateSenders = legitimateSenders;

		prepares.clear();
		commits.clear();
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public boolean addPrePrepare(REFITPBFTPrePrepare prePrepare, REFITMessageAuthentication mac) {
		if (prePrepare == null) return false;
		if (this.prePrepare != null) return false;
		if (message != null) return false;
		if (!mac.verifyMulticastMAC(prePrepare, legitimateSenders)) return false;
		REFITOrderRequest proposal = prePrepare.message;
		if (!REFITConfig.SIGNED_REQUESTS && !mac.verifyMulticastMACRequest(proposal)
				|| REFITConfig.SIGNED_REQUESTS && !mac.verifySignatureRequest(proposal)) return false;

		// check that request may be proposed yet
		if (proposal.type == REFITMessageType.REQUEST) {
			if (((REFITRequest) proposal).executeAfter >= prePrepare.uid.seqNr) return false;
		} else if (proposal.type == REFITMessageType.BATCH) {
			for (REFITRequest request : ((REFITBatch) proposal).requests) {
				if (request.executeAfter >= prePrepare.uid.seqNr) return false;
			}
		}

		this.prePrepare = prePrepare;
		this.message = proposal;
		prepares.add(prePrepare.from, ByteBuffer.wrap(this.message.getHash()), prePrepare);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[prePrepare.from] = PARTICIPATION_INTERVAL;

		checkPrepared();
		return true;
	}

	public boolean addPrepare(REFITPBFTPrepare prepare, REFITMessageAuthentication mac) {
		if (prepare == null) return false;
		if (prepares.hasVoted(prepare.from)) return false;
		if (!mac.verifyMulticastMAC(prepare, legitimateSenders)) return false;
		prepares.add(prepare.from, prepare.getCommand(), prepare);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[prepare.from] = PARTICIPATION_INTERVAL;

		checkPrepared();
		return true;
	}

	public AddResult addCommit(REFITPBFTCommit commit, REFITMessageAuthentication mac) {
		if (commit == null) return AddResult.INVALID;
		if (commits.hasVoted(commit.from)) return AddResult.DUPLICATE;
		if (!mac.verifyMulticastMAC(commit, legitimateSenders)) return AddResult.INVALID;
		commits.add(commit.from, commit.getCommand(), commit);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[commit.from] = PARTICIPATION_INTERVAL;

		checkCommitted();
		return AddResult.ADDED;
	}


	// ######################
	// # CERTIFICATE STATES #
	// ######################

	private boolean prepared;
	private boolean committed;


	private static final int PARTICIPATION_INTERVAL = 2 * REFITConfig.CHECKPOINT_INTERVAL;
	private static final int[] PARTICIPATION = new int[REFITConfig.TOTAL_NR_OF_REPLICAS];

	static {
		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i] = PARTICIPATION_INTERVAL;
	}


	public boolean isPrepared() {
		return prepared;
	}

	public boolean preparedVotesDiffer() {
		return prepares.votesDiffer();
	}

	public boolean isCommitted() {
		return committed;
	}

	private void checkPrepared() {
		if (prepared) return;
		if (prePrepare == null) return;
		ByteBuffer decision = prepares.getDecision();
		if (decision == null) return;
		if (!prepares.getDecidingBallots().contains(prePrepare)) return;
		prepared = true;
	}

	private void checkCommitted() {
		if (committed) return;
		if (!prepared) return;
		if (prePrepare == null) return;

//		if(!REFITConfig.USE_PASSIVE_REPLICATION) {
//			for(int i = 0; i < PARTICIPATION.length; i++) {
//				if(PARTICIPATION[i] <= 0) return;
//			}
//		}

		ByteBuffer decision = commits.getDecision();
		if (decision == null) return;
		// In the current implementation, a non-faulty replica that
		// has been provided with a faulty PRE_PREPARE will never
		// reach the status "committed" in the corresponding instance,
		// even if it is able to obtain 2f+1 correct COMMITs; see the
		// following code line. This is not a safety problem but
		// implementing a solution (which would require obtaining
		// the correct message) would lead to a more robust protocol.
		if (!decision.equals(ByteBuffer.wrap(prePrepare.message.getHash()))) return;
		committed = true;
	}


	// ##################
	// # PROGRESS PROOF #
	// ##################

	public REFITPBFTMessage[] getProof() {
		if (isPrepared()) return createProof(prepares, PREPARED_THRESHOLD);
		return new REFITPBFTMessage[0];
	}

	private static REFITPBFTMessage[] createProof(REFITBallotBox<?, ?, ? extends REFITPBFTMessage> ballotBox, int proofSize) {
		REFITPBFTMessage[] proof = new REFITPBFTMessage[proofSize];
		int i = 0;
		for (REFITPBFTMessage message : ballotBox.getDecidingBallots()) {
			proof[i] = message;
			if (++i >= proofSize) break;
		}
		return proof;
	}
}
