package refit.agreement.pbft;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Collection;

import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTMessage;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrePrepare;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrepare;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTViewChange;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.util.REFITBallotBox;


public class REFITPBFTTransitionCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private final int groupSize;
	private final REFITBallotBox<Short, ByteBuffer, REFITPBFTViewChange> ballot;
	private final REFITBallotBox<Short, ByteBuffer, REFITPBFTViewChange> nopBallot;

	private REFITOrderRequest result;
	private Collection<REFITPBFTViewChange> proof;
	private final boolean[] legitimateSenders;


	public REFITPBFTTransitionCertificate(int decisionThreshold, int nopDecisionThreshold, int groupSize, boolean[] legitimateSenders) {
		this.groupSize = groupSize;
		this.ballot = new REFITBallotBox<>(decisionThreshold);
		this.nopBallot = new REFITBallotBox<>(nopDecisionThreshold);
		this.legitimateSenders = legitimateSenders;

		result = null;
		proof = null;
	}


	@Override
	public String toString() {
		return "TCERT";
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public AddResult addViewChange(REFITPBFTViewChange viewChange, REFITMessageAuthentication mac) {
		// Check VIEW_CHANGE
		if (viewChange == null) return AddResult.INVALID;
		if (ballot.hasVoted(viewChange.from) || nopBallot.hasVoted(viewChange.from)) return AddResult.DUPLICATE;
		if (!mac.verifySignature(viewChange, legitimateSenders)) return AddResult.INVALID;

		if (viewChange.message.type == REFITMessageType.REQUEST_NO_OP) {
			nopBallot.add(viewChange.from, ByteBuffer.wrap(viewChange.message.getHash()), viewChange);
			if (nopBallot.getDecision() != null) {
				result = REFITNoOpRequest.INSTANCE;
				proof = nopBallot.getDecidingBallots();
			}

			return AddResult.ADDED;
		} else if (isValidViewChange(viewChange, mac)) {
			ballot.add(viewChange.from, ByteBuffer.wrap(viewChange.message.getHash()), viewChange);
			if (ballot.getDecision() != null) {
				result = ballot.getDecidingBallots().get(0).message;
				proof = ballot.getDecidingBallots();
			}

			return AddResult.ADDED;
		}
		return AddResult.INVALID;
	}

	private boolean isValidViewChange(REFITPBFTViewChange viewChange, REFITMessageAuthentication mac) {
		// Recreate agreement process
		if (viewChange.message.type == REFITMessageType.REQUEST_NO_OP) {
			return true;
		}

		REFITPBFTCertificate certificate = new REFITPBFTCertificate(groupSize);
		certificate.init(viewChange.uid.seqNr, viewChange.viewID, legitimateSenders);
		for (REFITPBFTMessage pbftMessage : viewChange.proof) {
			switch (pbftMessage.type) {
			case PBFT_PRE_PREPARE:
				certificate.addPrePrepare((REFITPBFTPrePrepare) pbftMessage, mac);
				break;
			case PBFT_PREPARE:
				certificate.addPrepare((REFITPBFTPrepare) pbftMessage, mac);
				break;
			default:
				REFITLogger.logWarning(this, "unexpected message type in VIEW_CHANGE proof (" + pbftMessage.type + ")");
			}
		}

		// Only accept VIEW_CHANGE if the certificate proves progress and does not contain junk
		if (certificate.preparedVotesDiffer() || !certificate.isPrepared()) {
			return false;
		}
		return Arrays.equals(certificate.message.getHash(), viewChange.message.getHash());
	}


	// ######################
	// # CERTIFICATE STATES #
	// ######################

	public boolean isStable() {
		return result != null;
	}

	public REFITOrderRequest getDecision() {
		return result;
	}

	public Collection<REFITPBFTViewChange> getProof() {
		return proof;
	}
}
