package refit.application.queue;

import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.Set;

import refit.application.REFITApplicationServer;
import refit.application.REFITResult;
import refit.message.REFITUniqueID;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointContainer;


public class REFITQueueServer implements REFITApplicationServer {

	private final LinkedList<byte[]> queue;


	public REFITQueueServer() {
		this.queue = new LinkedList<>();
	}


	@Override
	public void init() {
		// Do nothing
	}

	@Override
	public boolean checkReadOnly(ByteBuffer payload) {
		return false;
	}

	@Override
	public REFITResult processRequest(REFITUniqueID uid, long agreementSeqNr, ByteBuffer payload, boolean createUpdate) {
		REFITQueueRequest msg = new REFITQueueRequest(payload);

		REFITQueueReply reply;
		if (msg.isEnqueue) {
			queue.add(msg.data);
			reply = new REFITQueueReply();
		} else {
			int count = Math.min(queue.size(), msg.count);
			byte[][] data = new byte[count][];
			for (int i = 0; i < count; i++) {
				data[i] = queue.removeFirst();
			}

			reply = new REFITQueueReply(data);
		}

		reply.serializeMessage();

		// Return result
		return new REFITResult(reply.getBuffer(), createUpdate ? createUpdate(msg) : null);
	}

	private ByteBuffer createUpdate(REFITQueueRequest req) {
		ByteBuffer bb;
		if (req.isEnqueue) {
			bb = ByteBuffer.allocate(req.data.length);
			bb.put(req.data);
			bb.flip();
		} else {
			bb = ByteBuffer.allocate(0);
		}
		return bb;
	}

	@Override
	public void applyUpdate(ByteBuffer update) {
		if (update.remaining() == 0) return;

		update.mark();
		byte[] data = new byte[update.remaining()];
		update.get(data);
		update.reset();

		queue.add(data);
	}

	@Override
	public String toString() {
		return "QUEUE";
	}

	@Override
	public REFITCheckpointObject[] createCheckpoint() {
		int len = Integer.BYTES;
		for (byte[] bytes : queue) {
			len += Short.BYTES + bytes.length;
		}

		ByteBuffer checkpoint = ByteBuffer.allocate(len);
		checkpoint.putInt(queue.size());
		for (byte[] bytes : queue) {
			checkpoint.putShort((short) bytes.length);
			checkpoint.put(bytes);
		}
		return new REFITCheckpointObject[]{new REFITCheckpointObject(checkpoint.array())};
	}

	@Override
	public void applyCheckpoint(final REFITCheckpointObject[] checkpoint) {
//		System.out.println("Apply checkpoint");
		ByteBuffer checkpointBuffer = checkpoint[0].getBuffer();
		int queueLen = checkpointBuffer.getInt();
		queue.clear();
		for (int i = 0; i < queueLen; i++) {
			byte[] bytes = new byte[checkpointBuffer.getShort()];
			checkpointBuffer.get(bytes);
			queue.add(bytes);
		}
	}

	@Override
	public Set<?> takeChangelist() {
		throw new UnsupportedOperationException("Not implemented yet");
	}

	@Override
	public REFITCheckpointObject getObject(Object key) {
		throw new UnsupportedOperationException("Not implemented yet");
	}

	@Override
	public REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> createFuzzyCheckpoint() {
		throw new UnsupportedOperationException("Not implemented yet");
	}

}
