package refit.application.rezk;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import refit.application.rezk.common.*;
import refit.application.rezk.util.REFITByteString;
import refit.client.REFITClientLibrary;
import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;


public class REZKClient {

	private final REFITClientLibrary service;


	public REZKClient(REFITClientLibrary service) {
		this.service = service;
	}


	// #####################
	// # SERVICE INTERFACE #
	// #####################

	public String create(String path, byte[] data, boolean isSequential) throws REZKException {
		// Check data size
		if ((data != null) && (data.length > REZKServer.MAXIMUM_DATA_LENGTH))
			throw new REZKException("Data size " + data.length + " exceeds maximum of " + REZKServer.MAXIMUM_DATA_LENGTH);

		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.CREATE, new REFITByteString(path));
		request.setData(data);
		request.setSequential(isSequential);

		// Invoke operation
		REZKReply reply = invoke(request, false, true);
		return reply.path.toString();
	}

	public void delete(String path) throws REZKException {
		REZKRequest request = new REZKRequest(REZKOperation.DELETE, new REFITByteString(path));
		invoke(request, false, true);
	}

	public boolean exists(String path, boolean watch, boolean isTotalOrder) throws REZKException {
		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.EXISTS, new REFITByteString(path));
		request.setWatch(watch);

		// Invoke operation
		REZKReply reply = invoke(request, true, isTotalOrder);
		return reply.exists;
	}

	public List<String> getChildren(String path, boolean watch, REZKStat stat, boolean isTotalOrder) throws REZKException {
		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.GET_CHILDREN, new REFITByteString(path));
		request.setWatch(watch);

		// Invoke operation
		REZKReply reply = invoke(request, true, isTotalOrder);
		if (stat != null) stat.copyFrom(reply.stat);

		List<String> children = new ArrayList<>(reply.children.size());
		for (REFITByteString child : reply.children) {
			children.add(child.toString());
		}
		return children;
	}

	public byte[] getData(String path, boolean watch, REZKStat stat, boolean isTotalOrder) throws REZKException {
		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.GET_DATA, new REFITByteString(path));
		request.setWatch(watch);

		// Invoke operation
		REZKReply reply = invoke(request, !REFITConfig.UPDATE_LAST_ACCESS, REFITConfig.UPDATE_LAST_ACCESS || isTotalOrder);
		if (stat != null) stat.copyFrom(reply.stat);
		return reply.data;
	}

	public REZKStat setData(String path, byte[] data, int version) throws REZKException {
		// Check data size
		if (data.length > REZKServer.MAXIMUM_DATA_LENGTH)
			throw new REZKException("Data size " + data.length + " exceeds maximum of " + REZKServer.MAXIMUM_DATA_LENGTH);

		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.SET_DATA, new REFITByteString(path));
		request.setData(data);
		request.setVersion(version);

		// Invoke operation
		REZKReply reply = invoke(request, false, true);
		return reply.stat;
	}

	public int increment(String path) throws REZKException {
		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.INCREMENT, new REFITByteString(path));

		// Invoke operation
		REZKReply reply = invoke(request, false, true);
		ByteBuffer buffer = ByteBuffer.wrap(reply.data);
		return buffer.getInt();
	}

	public byte[] dequeue(String path, REZKStat stat) throws REZKException {
		// Create request
		REZKRequest request = new REZKRequest(REZKOperation.DEQUEUE, new REFITByteString(path));

		// Invoke operation
		REZKReply reply = invoke(request, false, true);
		if (stat != null) stat.copyFrom(reply.stat);
		return reply.data;
	}


	// ##############
	// # INVOCATION #
	// ##############

	private REZKReply invoke(REZKRequest request, boolean isReadOnly, boolean isTotalOrder) throws REZKException {
		request.serializeMessage();
		REZKReply reply;
		try {
			ByteBuffer replyBytes = service.invoke(request.getBuffer(), isReadOnly ? (isTotalOrder ? RequestMode.READ : RequestMode.READ_WEAK) : RequestMode.WRITE);
			reply = new REZKReply(replyBytes);
		} catch (Exception e) {
			throw new REZKException("Invoke failed ", e);
		}
		if (reply.message != null) throw new REZKException(reply.message.toString());
		return reply;
	}

}
