package refit.application.rezk;

import java.nio.ByteBuffer;
import java.util.*;

import refit.application.REFITApplicationServer;
import refit.application.REFITResult;
import refit.application.rezk.common.*;
import refit.application.rezk.common.REZKUpdate.REZKUpdateType;
import refit.application.rezk.datatree.REZKDataTree;
import refit.application.rezk.util.REFITByteString;
import refit.config.REFITConfig;
import refit.message.REFITUniqueID;
import refit.replica.checkpoint.REFITCheckpointObject;


public class REZKServer implements REFITApplicationServer {
	private static final ByteBuffer EMPTY_UPDATE_PAYLOAD = ByteBuffer.allocate(0);
	public static final int MAXIMUM_DATA_LENGTH = 4096;

	private final REZKDataTree tree;
	private long sequenceCounter;
	private static final REFITByteString SLASH = new REFITByteString("/");
	private static final REFITByteString EMPTY = new REFITByteString("");

	public REZKServer(REZKDataTree tree) {
		this.tree = tree;
		this.sequenceCounter = 0L;
		this.changes = new HashMap<>();
	}

	@Override
	public void init() {
		try {
			// Initialize tree
			tree.init();

			// Create root
			REZKStat stat = new REZKStat();
			stat.setVersion(0);
			stat.setNrOfChildren(0);
			tree.create(SLASH, null, stat);
		} catch (REZKException e) {
			throw new RuntimeException("Failed to initialize application", e);
		}
	}

	public void shutdown() throws REZKException {
		tree.shutdown();
	}

	@Override
	public boolean checkReadOnly(ByteBuffer payload) {
		payload.mark();
		REZKRequest request = new REZKRequest(payload);
		payload.rewind();

		switch (request.operation) {
		case CREATE:
		case DELETE:
		case SET_DATA:
		case INCREMENT:
		case DEQUEUE:
			return false;
		case EXISTS:
		case GET_CHILDREN:
			return true;
		case GET_DATA:
			return !REFITConfig.UPDATE_LAST_ACCESS;
		default:
			throw new IllegalArgumentException("Unexpected operation type");
		}
	}

	@Override
	public REFITResult processRequest(REFITUniqueID uid, long agreementSeqNr, ByteBuffer payload, boolean createUpdate) {
		REZKRequest request = new REZKRequest(payload);
		configureUpdates(createUpdate);

		REZKReply reply = invoke(request, agreementSeqNr);
		reply.serializeMessage();

		ByteBuffer update = null;
		if (REFITConfig.ENABLE_FUZZY_CHECKPOINTS && REFITConfig.UPDATE_BASED_CHECKPOINTING || createUpdate) {
			int size = getUpdatesSize();
			if (size == 0) {
				update = EMPTY_UPDATE_PAYLOAD;
			} else {
				update = ByteBuffer.allocate(size);
				collectUpdates(update);
				update.flip();
			}
		}

		return new REFITResult(reply.getBuffer(), update);
	}

	private REZKReply invoke(REZKRequest request, long agreementSeqNr) {
		REZKReply reply = new REZKReply(request.operation);
		try {
			switch (request.operation) {
			case CREATE:
				create(request, reply, agreementSeqNr);
				break;
			case DELETE:
				delete(request);
				break;
			case EXISTS:
				exists(request, reply);
				break;
			case GET_CHILDREN:
				getChildren(request, reply);
				break;
			case GET_DATA:
				getData(request, reply, agreementSeqNr);
				break;
			case SET_DATA:
				setData(request, reply, agreementSeqNr);
				break;
			case INCREMENT:
				increment(request, reply, agreementSeqNr);
				break;
			case DEQUEUE:
				dequeue(request, reply, agreementSeqNr);
				break;
			default:
				throw new REZKException("Unexpected operation: " + request.operation);
			}
		} catch (REZKException rzke) {
			reply.setException(rzke);
		}
		return reply;
	}

	public static REFITByteString getParentPath(REFITByteString path) {
		if (SLASH.equals(path)) return new REFITByteString();
		int lastSlashIndex = path.lastIndexOf((byte) '/');
		return (lastSlashIndex == 0) ? new REFITByteString("/") : path.substring(0, lastSlashIndex);
	}

	public static REFITByteString getChildPath(REFITByteString path) {
		return path.substring(path.lastIndexOf((byte) '/') + 1);
	}


	// ###################
	// # REZKClient OPERATIONS #
	// ###################

	private void create(REZKRequest request, REZKReply reply, long agreementSeqNr) throws REZKException {
		// Check parent
		if (SLASH.equals(request.path)) throw new REZKException("Cannot create root");
		REFITByteString parentPath = getParentPath(request.path);
		REZKStat parentStat = tree.getStat(parentPath);
		if (parentStat == null) throw new REZKException("Parent " + parentPath + " does not exist");

		// Check node
		REFITByteString nodePath;
		if (request.isSequential) {
			// Sequential node
			nodePath = new REFITByteString(String.format("%s%010d", request.path, sequenceCounter));
			sequenceCounter++;
			if (tree.getStat(nodePath) != null) {
				throw new RuntimeException();
			}
		} else {
			// Regular node
			nodePath = request.path;
			if (tree.getStat(nodePath) != null) throw new REZKException("Node " + nodePath + " already exists");
		}

		// Create node
		REZKStat stat = new REZKStat();
		stat.setVersion(0);
		if (request.data != null) stat.setDataLength(request.data.length);
		stat.setNrOfChildren(0);
		stat.setLastAccess(agreementSeqNr);
		tree.create(nodePath, request.data, stat);

		// Update parent stat
		parentStat.setNrOfChildren(parentStat.nrOfChildren + 1);
		parentStat.setLastAccess(agreementSeqNr);
		tree.updateStat(parentPath, parentStat);

		// Update reply
		reply.setPath(nodePath);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addCreateUpdate(nodePath, stat, request.data);
			addStatUpdate(parentPath, parentStat);
			if (request.isSequential) {
				addSequenceUpdate();
			}
		} else {
			changedObjects.add(nodePath);
			changedObjects.add(parentPath);
			if (request.isSequential) {
				changedObjects.add(EMPTY);
			}
		}
	}

	private void delete(REZKRequest request) throws REZKException {
		// Perform checks
		if (SLASH.equals(request.path)) throw new REZKException("Cannot delete root");
		REZKStat stat = tree.getStat(request.path);
		if (stat == null) throw new REZKException("Node " + request.path + " does not exist");
		if (stat.nrOfChildren != 0) throw new REZKException("Node " + request.path + " has children");

		// Delete node
		tree.delete(request.path);

		// Update parent stat
		REFITByteString parentPath = getParentPath(request.path);
		REZKStat parentStat = tree.getStat(parentPath);
		parentStat.setNrOfChildren(parentStat.nrOfChildren - 1);
		tree.updateStat(parentPath, parentStat);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addDeleteUpdate(request.path);
			addStatUpdate(parentPath, parentStat);
		} else {
			changedObjects.add(request.path);
			changedObjects.add(parentPath);
		}
	}

	private void exists(REZKRequest request, REZKReply reply) throws REZKException {
		// Check whether node exists
		REZKStat stat = tree.getStat(request.path);
		reply.setExists(stat != null);
	}

	private void getChildren(REZKRequest request, REZKReply reply) throws REZKException {
		// Get node
		REZKStat stat = tree.getStat(request.path);
		if (stat == null) throw new REZKException("Node " + request.path + " does not exist");

		// Get children
		Set<REFITByteString> children = tree.getChildren(request.path);

		// Update reply
		reply.setStat(stat);
		reply.setChildren(children);
	}

	private void getData(REZKRequest request, REZKReply reply, long agreementSeqNr) throws REZKException {
		// Get node
		REZKStat stat = new REZKStat();
		byte[] data = tree.getData(request.path, stat);
		if (stat.version < 0) throw new REZKException("Node " + request.path + " does not exist");

		// Update reply
		reply.setStat(stat);
		reply.setData(data);

		if (!REFITConfig.UPDATE_LAST_ACCESS) return;

		stat.setLastAccess(agreementSeqNr);
		tree.updateStat(request.path, stat);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addStatUpdate(request.path, stat);
		} else {
			changedObjects.add(request.path);
		}
	}

	private void setData(REZKRequest request, REZKReply reply, long agreementSeqNr) throws REZKException {
		// Get node
		REZKStat stat = tree.getStat(request.path);
		if (stat == null) throw new REZKException("Node " + request.path + " does not exist");

		// Check version
		if ((request.version >= 0) && (request.version != stat.version))
			throw new REZKException("Version mismatch: " + request.version + " vs. " + stat.version);

		// Set data
		stat.setVersion(stat.version + 1);
		stat.setDataLength((request.data != null) ? request.data.length : -1);
		stat.setLastAccess(agreementSeqNr);
		tree.setData(request.path, request.data, stat);

		// Update reply
		reply.setStat(stat);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addDataUpdate(request.path, stat, request.data);
		} else {
			changedObjects.add(request.path);
		}
	}

	private void increment(REZKRequest request, REZKReply reply, long agreementSeqNr) throws REZKException {
		// Get node
		REZKStat stat = new REZKStat();
		byte[] data = tree.getData(request.path, stat);
		if (stat.version < 0) throw new REZKException("Node " + request.path + " does not exist");

		// Update value
		ByteBuffer buffer = ByteBuffer.wrap(data);
		int newValue = buffer.getInt(0) + 1;
		buffer.putInt(0, newValue);
		stat.setVersion(stat.version + 1);
		stat.setLastAccess(agreementSeqNr);
		tree.setData(request.path, data, stat);

		// Update reply
		reply.setStat(stat);
		reply.setData(data);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addDataUpdate(request.path, stat, data);
		} else {
			changedObjects.add(request.path);
		}
	}

	private void dequeue(REZKRequest request, REZKReply reply, long agreementSeqNr) throws REZKException {
		// Get node
		REZKStat queueStat = tree.getStat(request.path);
		if (queueStat == null) throw new REZKException("Node " + request.path + " does not exist");

		// Get children
		Set<REFITByteString> children = tree.getChildren(request.path);
		if (children.isEmpty()) throw new REZKException("Queue " + request.path + " is empty");

		// Determine head element
		REFITByteString headPath = null;
		int headIndex = Integer.MAX_VALUE;
		for (REFITByteString child : children) {
			int index = Integer.parseInt(child.substring(child.length() - 10).toString());
			if ((headPath == null) || (index < headIndex)) {
				headPath = child;
				headIndex = index;
			}
		}

		// Retrieve head element
		REZKStat headStat = new REZKStat();
		REFITByteString fullPath = request.path.concat(SLASH.concat(headPath));
		byte[] headData = tree.getData(fullPath, headStat);
		tree.delete(fullPath);

		// Update queue stat
		queueStat.setNrOfChildren(queueStat.nrOfChildren - 1);
		queueStat.setLastAccess(agreementSeqNr);
		tree.updateStat(request.path, queueStat);

		// Update reply
		reply.setStat(headStat);
		reply.setData(headData);

		// Handle update
		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING && !updatesEnabled) {
			return;
		}

		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			addDeleteUpdate(fullPath);
			addStatUpdate(request.path, queueStat);
		} else {
			changedObjects.add(request.path);
			changedObjects.add(fullPath);
		}
	}

	// ###########
	// # WATCHES #
	// ###########

	// TODO
	// Data watches: Node created, node deleted, data changed
	// Child watches: New child, child deleted


	// ###########
	// # UPDATES #
	// ###########

	private final Map<REFITByteString, REZKUpdate> changes;
	private boolean updatesEnabled;


	private void configureUpdates(boolean enable) {
		updatesEnabled = enable;
	}

	private int getUpdatesSize() {
		int size = 0;
		for (REZKUpdate update : changes.values()) {
			size += update.calculatePayloadSize();
		}
		return size;
	}

	private void collectUpdates(ByteBuffer buffer) {
		if (!REFITConfig.UPDATE_BASED_CHECKPOINTING) throw new AssertionError("No updates for you!");
		for (REZKUpdate update : changes.values()) {
			update.serializeMessage();
			buffer.put(update.getBuffer());
		}
		changes.clear();
	}

	@Override
	public void applyUpdate(ByteBuffer buffer) {
		Collection<REZKUpdate> updates = new LinkedList<>();
		while (buffer.hasRemaining()) {
			REZKUpdate update = new REZKUpdate(buffer.slice());
			if (update.type == REZKUpdateType.SET_SEQ) {
				sequenceCounter = update.sequenceCounter;
			} else {
				updates.add(update);
			}
			buffer.position(buffer.position() + update.getMessageSize());
		}
		try {
			tree.applyUpdates(updates);
		} catch (REZKException rzke) {
			rzke.printStackTrace();
		}
	}

	private void addStubUpdate(REFITByteString path) {
		REZKUpdate update = new REZKUpdate(REZKUpdateType.STUB, path);
		changes.put(path, update);
	}

	private void addStatStubUpdate(REFITByteString path) {
		REZKUpdate update = new REZKUpdate(REZKUpdateType.STAT_STUB, path);
		changes.put(path, update);
	}

	private void addCreateUpdate(REFITByteString path, REZKStat stat, byte[] data) {
		if (!updatesEnabled) {
			addStubUpdate(path);
			return;
		}
		REZKUpdate update = new REZKUpdate(REZKUpdateType.CREATE, path);
		update.setStat(stat);
		update.setData(data);
		changes.put(path, update);
	}

	private void addDeleteUpdate(REFITByteString path) {
		if (!updatesEnabled) {
			addStubUpdate(path);
			return;
		}
		// Insert DELETE update if this is the first recent operation on the node
		REZKUpdate update = changes.get(path);
		if (update == null) {
			update = new REZKUpdate(REZKUpdateType.DELETE, path);
			changes.put(path, update);
			return;
		}

		// Remove update if the node has only been created recently
		if (update.type == REZKUpdateType.CREATE) {
			changes.remove(path);
			return;
		}

		// Transform update
		update.setType(REZKUpdateType.DELETE);
		update.setData(null);
	}

	private void addDataUpdate(REFITByteString path, REZKStat stat, byte[] data) {
		if (!updatesEnabled) {
			addStubUpdate(path);
			return;
		}
		// Get update
		REZKUpdate update = changes.get(path);
		if (update == null) {
			update = new REZKUpdate(REZKUpdateType.SET_DATA, path);
			changes.put(path, update);
		} else if (update.type == REZKUpdateType.SET_STAT) {
			// Transform update
			update.setType(REZKUpdateType.SET_DATA);
		}

		// Update contents
		update.setStat(stat);
		update.setData(data);
	}

	private void addStatUpdate(REFITByteString path, REZKStat stat) {
		if (!updatesEnabled) {
			addStatStubUpdate(path);
			return;
		}
		REZKUpdate update = changes.get(path);
		if (update == null) {
			update = new REZKUpdate(REZKUpdateType.SET_STAT, path);
			changes.put(path, update);
		}
		update.setStat(stat);
	}

	private void addSequenceUpdate() {
		if (!updatesEnabled) {
			addStubUpdate(EMPTY);
			return;
		}
		REZKUpdate update = changes.get(EMPTY);
		if (update == null) {
			update = new REZKUpdate(REZKUpdateType.SET_SEQ, EMPTY);
			changes.put(EMPTY, update);
		}
		update.setSequenceCounter(sequenceCounter);
	}

	// #################
	// # Copy-after-Write #
	// #################

	private Set<REFITByteString> changedObjects = new HashSet<>();

	@Override
	public Set<REFITByteString> takeChangelist() {
		Set<REFITByteString> changed = changedObjects;
		changedObjects = new HashSet<>();
		return changed;
	}

	@Override
	public REFITCheckpointObject getObject(Object key) {
		REZKUpdate update;
		if (key.equals(EMPTY)) {
			update = new REZKUpdate(REZKUpdateType.SET_SEQ, EMPTY);
			update.setSequenceCounter(sequenceCounter);
		} else {
			// Get node
			REZKStat stat = new REZKStat();
			byte[] data;
			try {
				data = tree.getData((REFITByteString) key, stat);
			} catch (REZKException e) {
				throw new RuntimeException("O.o");
			}

			if (stat.version < 0) {
				return null;
			}

			update = new REZKUpdate(REZKUpdateType.CREATE, ((REFITByteString) key));
			update.setStat(stat);
			update.setData(data);
		}
		update.serializeMessage();
		return update;
	}

	// ##############
	// # CHECKPOINT #
	// ##############

	@Override
	public REFITCheckpointObject[] createCheckpoint() {
		try {
			Queue<ByteBuffer> objects = new LinkedList<>();
			tree.createCheckpoint(objects);

			int blocksize = 100 * 1024 * 1024;
			LinkedList<REFITCheckpointObject> blocks = new LinkedList<>();

			while (true) {
				int size = Long.BYTES;
				int parts = 0;
				Iterator<ByteBuffer> iterator = objects.iterator();
				while (iterator.hasNext() && size < blocksize) {
					ByteBuffer object = iterator.next();
					size += object.remaining();
					parts++;
				}

				ByteBuffer buffer = ByteBuffer.allocate(size);
				buffer.putLong(sequenceCounter);

				iterator = objects.iterator();
				for (int i = 0; i < parts && iterator.hasNext(); i++) {
					ByteBuffer object = iterator.next();
					buffer.put(object);
					iterator.remove();
				}
				buffer.flip();
				blocks.add(new REFITCheckpointObject(buffer));

				if (!iterator.hasNext()) {
					break;
				}
			}

			return blocks.toArray(new REFITCheckpointObject[0]);
		} catch (REZKException rzke) {
			rzke.printStackTrace();
			return null;
		}
	}

	@Override
	public REZKFuzzyCheckpointContainer createFuzzyCheckpoint() {
		try {
			REZKFuzzyCheckpointContainer checkpoint = new REZKFuzzyCheckpointContainer();
			tree.createFuzzyCheckpoint(checkpoint);

			REZKUpdate seqUpdate = new REZKUpdate(REZKUpdate.REZKUpdateType.SET_SEQ, EMPTY);
			seqUpdate.setSequenceCounter(sequenceCounter);
			checkpoint.addData(seqUpdate);

			return checkpoint;
		} catch (REZKException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public Map<Object, REFITCheckpointObject> createFuzzyDiffCheckpoint(Set<?> changelist) {
		try {
			Map<Object, REFITCheckpointObject> changedObjects = new HashMap<>();
			tree.createFuzzyDiffCheckpoint(changelist, changedObjects);

			if (changelist.contains(EMPTY)) {
				changedObjects.put(EMPTY, getObject(EMPTY));
			}

			return changedObjects;
		} catch (REZKException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public Collection<REFITCheckpointObject> createFuzzyUpdateCheckpoint(Map<?, REFITCheckpointObject> changelist) {
		try {
			Collection<REFITCheckpointObject> changedObjects = new ArrayList<>(changelist.size());
			tree.createFuzzyUpdateCheckpoint(changelist, changedObjects);

			if (changelist.containsKey(EMPTY)) {
				changedObjects.add(getObject(EMPTY));
			}

			return changedObjects;
		} catch (REZKException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public void applyCheckpoint(REFITCheckpointObject[] checkpoint) {
		try {
			if (checkpoint.length == 1) {
				ByteBuffer buffer = checkpoint[0].getBuffer();
				sequenceCounter = buffer.getLong();
				tree.applyCheckpoint(buffer);
				changes.clear();
			} else {
				// object based checkpoint
				tree.shutdown();
				tree.init();
				for (REFITCheckpointObject object : checkpoint) {
					applyUpdate(object.getBuffer());
				}
			}
		} catch (REZKException rzke) {
			throw new InternalError(rzke);
		}
	}
}
