package refit.application.rezk.common;

import java.nio.ByteBuffer;
import java.util.Collection;
import java.util.Set;

import refit.application.rezk.util.REFITByteString;
import refit.application.rezk.util.REFITSerialization;
import refit.application.rezk.util.REFITSerialization.REFITSerializationSize;
import refit.message.REFITBaseMessage;


public class REZKReply extends REFITBaseMessage {

	public final REZKOperation operation;
	public REFITByteString message;

	public byte[] data;
	public REZKStat stat;
	public REFITByteString path;
	public boolean exists;
	public Collection<REFITByteString> children;


	public REZKReply(REZKOperation operation) {
		this.operation = operation;
	}


	// ###########
	// # SETTERS #
	// ###########

	public void setException(REZKException exception) {
		message = new REFITByteString(exception.getMessage());
	}

	public void setData(byte[] data) {
		this.data = data;
	}

	public void setStat(REZKStat stat) {
		this.stat = stat;
	}

	public void setPath(REFITByteString path) {
		this.path = path;
	}

	public void setExists(boolean exists) {
		this.exists = exists;
	}

	public void setChildren(Set<REFITByteString> children) {
		this.children = children;
	}


	// ###################################
	// # SERIALIZATION & DESERIALIZATION #
	// ###################################

	public REZKReply(ByteBuffer buffer) {
		super(buffer);
		// Deserialize common attributes
		this.operation = REZKOperation.values[buffer.get()];
		this.message = REFITSerialization.getByteString(buffer, REFITSerializationSize.TINY);
		if (message != null) return;

		// Deserialize specific attributes
		switch (operation) {
		case CREATE:
			this.path = REFITSerialization.getByteString(buffer, REFITSerializationSize.TINY);
			break;
		case DELETE:
			// Nothing more to do
			break;
		case EXISTS:
			this.exists = REFITSerialization.getBoolean(buffer);
			break;
		case GET_CHILDREN:
			this.stat = new REZKStat(buffer);
			this.children = REFITSerialization.getStringCollection(buffer, REFITSerializationSize.SMALL, REFITSerializationSize.TINY);
			break;
		case GET_DATA:
		case DEQUEUE:
			this.stat = new REZKStat(buffer);
			this.data = REFITSerialization.getByteArray(buffer, REFITSerializationSize.SMALL);
			break;
		case SET_DATA:
			this.stat = new REZKStat(buffer);
			break;
		case INCREMENT:
			this.data = REFITSerialization.getByteArray(buffer, REFITSerializationSize.TINY);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		// Serialize common attributes
		REFITSerialization.putByte(buffer, operation.ordinal());
		REFITSerialization.putByteString(buffer, message, REFITSerializationSize.TINY);
		if (message != null) return;

		// Serialize specific attributes
		switch (operation) {
		case CREATE:
			REFITSerialization.putByteString(buffer, path, REFITSerializationSize.TINY);
			break;
		case DELETE:
			// Nothing more to do
			break;
		case EXISTS:
			REFITSerialization.putBoolean(buffer, exists);
			break;
		case GET_CHILDREN:
			stat.serialize(buffer);
			REFITSerialization.putStringCollection(buffer, children, REFITSerializationSize.SMALL,
					REFITSerializationSize.TINY);
			break;
		case GET_DATA:
		case DEQUEUE:
			stat.serialize(buffer);
			REFITSerialization.putByteArray(buffer, data, REFITSerializationSize.SMALL);
			break;
		case SET_DATA:
			stat.serialize(buffer);
			break;
		case INCREMENT:
			REFITSerialization.putByteArray(buffer, data, REFITSerializationSize.TINY);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
	}

	@Override
	protected int getHeaderSize() {
		int size = super.getHeaderSize() + Byte.BYTES
				+ REFITSerialization.getByteStringSize(message, REFITSerializationSize.TINY);
		if (message != null) {
			return size;
		}

		switch (operation) {
		case CREATE:
			size += REFITSerialization.getByteStringSize(path, REFITSerializationSize.TINY);
			break;
		case DELETE:
			break;
		case EXISTS:
			size += REFITSerialization.getBooleanSize();
			break;
		case GET_CHILDREN:
			size += stat.getSerializedSize() + REFITSerialization.getStringCollectionSize(children,
					REFITSerializationSize.SMALL, REFITSerializationSize.TINY);
			break;
		case GET_DATA:
		case DEQUEUE:
			size += stat.getSerializedSize()
					+ REFITSerialization.getByteArraySize(data, REFITSerializationSize.SMALL);
			break;
		case SET_DATA:
			size += stat.getSerializedSize();
			break;
		case INCREMENT:
			size += REFITSerialization.getByteArraySize(data, REFITSerializationSize.TINY);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
		return size;
	}

	@Override
	protected int calculatePayloadSize() {
		return getHeaderSize();
	}

}
