package refit.application.rezk.common;

import java.nio.ByteBuffer;

import refit.application.rezk.util.REFITByteString;
import refit.application.rezk.util.REFITSerialization;
import refit.application.rezk.util.REFITSerialization.REFITSerializationSize;
import refit.replica.checkpoint.REFITCheckpointObject;

public class REZKUpdate extends REFITCheckpointObject {

	public final REFITByteString path;
	public final REZKStat stat;
	public REZKUpdateType type;
	public long sequenceCounter = -1L;
	private ByteBuffer data;

	public REZKUpdate(final REZKUpdateType type, final REFITByteString path) {
		this.type = type;
		this.path = path;
		this.stat = new REZKStat();
		this.data = null;
	}


	// ########
	// # TYPE #
	// ########

	public enum REZKUpdateType {

		CREATE,
		DELETE,
		SET_DATA,
		SET_STAT,
		SET_SEQ,
		STUB,
		STAT_STUB;

		public static final REZKUpdateType[] values = values();

	}


	// ###########
	// # SETTERS #
	// ###########

	public void setType(REZKUpdateType type) {
		this.type = type;
	}

	public void setSequenceCounter(final long sequenceCounter) {
		this.sequenceCounter = sequenceCounter;
	}

	public void setStat(final REZKStat stat) {
		this.stat.copyFrom(stat);
	}

	public void setData(final byte[] data) {
		this.data = (data != null) ? ByteBuffer.wrap(data) : null;
	}

	public void setDataFrom(REZKUpdate update) {
		if (update.data == null) {
			this.data = null;
		} else {
			this.data = update.data.slice();
		}
	}

	public byte[] dataArray() {
		if (data == null) return null;
		byte[] array = new byte[data.remaining()];
		data.mark();
		data.get(array);
		data.reset();
		return array;
	}

	// ###################################
	// # SERIALIZATION & DESERIALIZATION #
	// ###################################

	public REZKUpdate(ByteBuffer buffer) {
		super(buffer);
		this.type = REZKUpdateType.values[buffer.get()];
		this.path = REFITSerialization.getByteString(buffer, REFITSerializationSize.TINY);
		if (type == REZKUpdateType.DELETE || type == REZKUpdateType.STUB || type == REZKUpdateType.STAT_STUB) {
			this.stat = null;
			this.data = null;
		} else if (type == REZKUpdateType.SET_SEQ) {
			this.stat = null;
			this.data = null;
			this.sequenceCounter = buffer.getLong();
		} else {
			this.stat = new REZKStat(buffer);
			this.data = REFITSerialization.getByteBuffer(buffer, REFITSerializationSize.SMALL);
		}

		messageSize = buffer.position();
		serializationBuffer.flip();
	}

	@Override
	protected void serialize(final ByteBuffer buffer) {
		REFITSerialization.putByte(buffer, type.ordinal());
		REFITSerialization.putByteString(buffer, path, REFITSerializationSize.TINY);
		if (type == REZKUpdateType.SET_SEQ) {
			buffer.putLong(sequenceCounter);
		} else if (type != REZKUpdateType.DELETE && type != REZKUpdateType.STUB && type != REZKUpdateType.STAT_STUB) {
			stat.serialize(buffer);
			ByteBuffer dataBuffer = buffer.slice();
			REFITSerialization.putByteBuffer(buffer, data, REFITSerializationSize.SMALL);
			// share data with serialization buffer to avoid duplicate memory usage
			this.data = REFITSerialization.getByteBuffer(dataBuffer, REFITSerializationSize.SMALL);
		}
	}

	protected int getHeaderSize() {
		int size = Byte.BYTES + REFITSerialization.getByteStringSize(path, REFITSerializationSize.TINY);
		if (type == REZKUpdateType.DELETE || type == REZKUpdateType.STUB || type == REZKUpdateType.STAT_STUB) {
			return size;
		} else if (type == REZKUpdateType.SET_SEQ) {
			return size + Long.BYTES;
		}

		return size + stat.getSerializedSize()
				+ REFITSerialization.getByteBufferSize(data, REFITSerializationSize.SMALL);
	}

	@Override
	public int calculatePayloadSize() {
		return getHeaderSize();
	}

	@Override
	public String toString() {
		return "REZKUpdate [path=" + path + ", stat=" + stat + ", type=" + type + ", sequenceCounter=" + sequenceCounter
				+ ", data len=" + ((data != null) ? data.remaining() : "") + "]";
	}
}
