package refit.application.rezk.datatree;

import java.nio.ByteBuffer;
import java.util.*;

import refit.application.rezk.REZKFuzzyCheckpointContainer;
import refit.application.rezk.REZKServer;
import refit.application.rezk.common.REZKException;
import refit.application.rezk.common.REZKStat;
import refit.application.rezk.common.REZKUpdate;
import refit.application.rezk.util.REFITByteString;
import refit.replica.checkpoint.REFITCheckpointObject;


public class REZKInMemoryDataTree implements REZKDataTree {

	private final Map<REFITByteString, REZKInMemoryDataNode> nodes;
	private static final REFITByteString SLASH = new REFITByteString("/");


	public REZKInMemoryDataTree() {
		this.nodes = new LinkedHashMap<>();
	}


	// ########
	// # TREE #
	// ########

	@Override
	public void init() {
		// Nothing to be done
	}

	@Override
	public void shutdown() throws REZKException {
		// Nothing to be done
	}


	// ###################
	// # REZKClient OPERATIONS #
	// ###################

	@Override
	public void create(REFITByteString path, byte[] data, REZKStat stat) {
		// Create node
		REZKInMemoryDataNode node = new REZKInMemoryDataNode(path);
		node.setStat(stat);
		node.setData(data);
		nodes.put(node.path, node);

		// Add child to parent
		if (SLASH.equals(path)) return;
		REFITByteString parentPath = REZKServer.getParentPath(path);
		REZKInMemoryDataNode parent = nodes.get(parentPath);
		parent.addChild(path);
	}

	@Override
	public void delete(REFITByteString path) {
		// Remove node
		nodes.remove(path);

		// Remove child from parent
		if (SLASH.equals(path)) return;
		REFITByteString parentPath = REZKServer.getParentPath(path);
		REZKInMemoryDataNode parent = nodes.get(parentPath);
		parent.removeChild(path);
	}

	@Override
	public Set<REFITByteString> getChildren(REFITByteString path) {
		REZKInMemoryDataNode node = nodes.get(path);
		return node.children;
	}

	@Override
	public byte[] getData(REFITByteString path, REZKStat stat) {
		REZKInMemoryDataNode node = nodes.get(path);
		if (node == null) return null;
		stat.copyFrom(node.stat);
		return node.data;
	}

	@Override
	public void setData(REFITByteString path, byte[] data, REZKStat stat) {
		REZKInMemoryDataNode node = nodes.get(path);
		node.setStat(stat);
		node.setData(data);
	}


	// ###########
	// # HELPERS #
	// ###########

	@Override
	public REZKStat getStat(REFITByteString path) {
		REZKInMemoryDataNode node = nodes.get(path);
		return (node != null) ? node.stat : null;
	}

	@Override
	public void updateStat(REFITByteString path, REZKStat stat) {
		REZKInMemoryDataNode node = nodes.get(path);
		node.setStat(stat);
	}


	// ###########
	// # UPDATES #
	// ###########

	@Override
	public void applyUpdates(Collection<REZKUpdate> updates) {
		for (REZKUpdate update : updates) {
			switch (update.type) {
			case CREATE:
				create(update.path, update.dataArray(), update.stat);
				break;
			case DELETE:
				delete(update.path);
				break;
			case SET_DATA:
				setData(update.path, update.dataArray(), update.stat);
				break;
			case SET_STAT:
				updateStat(update.path, update.stat);
				break;
			default:
				System.err.println("Unexpected update type: " + update.type);
			}
		}
	}


	// ###############
	// # CHECKPOINTS #
	// ###############

	@Override
	public void createCheckpoint(final Queue<ByteBuffer> queue) {
		for (final REZKInMemoryDataNode node : nodes.values()) {
			queue.add(node.serialize());
		}
	}

	@Override
	public void createFuzzyCheckpoint(REZKFuzzyCheckpointContainer checkpoint) throws REZKException {
		// FIXME implement
		throw new UnsupportedOperationException("Not implemented yet");
	}

	@Override
	public void createFuzzyDiffCheckpoint(Set<?> changelist, Map<Object, REFITCheckpointObject> changedObjects) throws REZKException {
		// FIXME implement
		throw new UnsupportedOperationException("Not implemented yet");
	}

	@Override
	public void createFuzzyUpdateCheckpoint(Map<?, REFITCheckpointObject> changelist, Collection<REFITCheckpointObject> changedObjects) throws REZKException {
		// FIXME implement
		throw new UnsupportedOperationException("Not implemented yet");
	}

	@Override
	public Map<REFITByteString, REZKStat> applyCheckpoint(ByteBuffer buffer) {
		nodes.clear();
		Map<REFITByteString, REZKStat> stats = new HashMap<>();
		while (buffer.hasRemaining()) {
			REZKInMemoryDataNode node = new REZKInMemoryDataNode(buffer);
			create(node.path, node.data, node.stat);
			stats.put(node.path, node.stat);
		}
		return stats;
	}

}
