package refit.application.rezk.util;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collection;


public class REFITSerialization {

	// ################
	// # ELEMENT SIZE #
	// ################

	public enum REFITSerializationSize {
		TINY,
		SMALL,
		REGULAR
	}


	// ########
	// # BYTE #
	// ########

	public static void putByte(ByteBuffer buffer, short value) {
		if (value != ((byte) value)) throw new RuntimeException("Invalid value: " + value);
		buffer.put((byte) value);
	}

	public static void putByte(ByteBuffer buffer, int value) {
		if (value != ((byte) value)) throw new RuntimeException("Invalid value: " + value);
		buffer.put((byte) value);
	}


	// #########
	// # SHORT #
	// #########

	public static void putShort(ByteBuffer buffer, int value) {
		if (value != ((short) value)) throw new RuntimeException("Invalid value: " + value);
		buffer.putShort((short) value);
	}


	// ###########
	// # INTEGER #
	// ###########

	public static void put(ByteBuffer buffer, int value, REFITSerializationSize size) {
		switch (size) {
		case TINY:
			putByte(buffer, value);
			break;
		case SMALL:
			putShort(buffer, value);
			break;
		case REGULAR:
			buffer.putInt(value);
			break;
		default:
			throw new RuntimeException("Unexpected size: " + size);
		}
	}

	public static int get(ByteBuffer buffer, REFITSerializationSize size) {
		switch (size) {
		case TINY:
			return buffer.get();
		case SMALL:
			return buffer.getShort();
		case REGULAR:
			return buffer.getInt();
		default:
			throw new RuntimeException("Unexpected size: " + size);
		}
	}

	public static int getSize(REFITSerializationSize size) {
		switch (size) {
		case TINY:
			return Byte.BYTES;
		case SMALL:
			return Short.BYTES;
		case REGULAR:
			return Integer.BYTES;
		default:
			throw new RuntimeException("Unexpected size: " + size);
		}
	}


	// ###########
	// # BOOLEAN #
	// ###########

	public static void putBoolean(ByteBuffer buffer, boolean value) {
		putByte(buffer, value ? 1 : 0);
	}

	public static boolean getBoolean(ByteBuffer buffer) {
		return (buffer.get() > 0);
	}

	public static int getBooleanSize() {
		return 1;
	}


	// ##############
	// # BYTE ARRAY #
	// ##############

	public static void putByteArray(ByteBuffer buffer, byte[] array, REFITSerializationSize arraySize) {
		int length = (array != null) ? array.length : -1;
		put(buffer, length, arraySize);
		if (array != null) buffer.put(array);
	}

	public static byte[] getByteArray(ByteBuffer buffer, REFITSerializationSize arraySize) {
		int length = get(buffer, arraySize);
		if (length < 0) return null;
		byte[] array = new byte[length];
		buffer.get(array);
		return array;
	}

	public static int getByteArraySize(byte[] array, REFITSerializationSize arraySize) {
		return getSize(arraySize) + ((array != null) ? array.length : 0);
	}

	// ###############
	// # BYTE BUFFER #
	// ###############

	public static void putByteBuffer(final ByteBuffer buffer, final ByteBuffer array,
	                                 final REFITSerializationSize arraySize) {
		final int length = (array != null) ? array.remaining() : -1;
		put(buffer, length, arraySize);
		if (array != null) {
			array.mark();
			buffer.put(array);
			array.reset();
		}
	}

	public static ByteBuffer getByteBuffer(final ByteBuffer buffer, final REFITSerializationSize arraySize) {
		final int length = get(buffer, arraySize);
		if (length < 0) {
			return null;
		}
		final ByteBuffer array = buffer.slice();
		array.limit(length);
		buffer.position(buffer.position() + length);
		return array;
	}

	public static int getByteBufferSize(final ByteBuffer array, final REFITSerializationSize arraySize) {
		return getSize(arraySize) + ((array != null) ? array.remaining() : 0);
	}

	// ###############
	// # BYTE STRING #
	// ###############

	public static void putByteString(ByteBuffer buffer, REFITByteString string, REFITSerializationSize stringSize) {
		putByteArray(buffer, (string != null) ? string.getBytes() : null, stringSize);
	}

	public static REFITByteString getByteString(ByteBuffer buffer, REFITSerializationSize stringSize) {
		byte[] stringBytes = getByteArray(buffer, stringSize);
		return (stringBytes != null) ? new REFITByteString(stringBytes) : null;
	}

	public static int getByteStringSize(REFITByteString string, REFITSerializationSize stringSize) {
		return getSize(stringSize) + ((string != null) ? string.length() : 0);
	}


	// #####################
	// # STRING COLLECTION #
	// #####################

	public static void putStringCollection(ByteBuffer buffer, Collection<REFITByteString> collection, REFITSerializationSize listSize, REFITSerializationSize stringSize) {
		int length = (collection != null) ? collection.size() : -1;
		put(buffer, length, listSize);
		if (collection != null) {
			for (REFITByteString string : collection) putByteString(buffer, string, stringSize);
		}
	}

	public static Collection<REFITByteString> getStringCollection(ByteBuffer buffer, REFITSerializationSize listSize, REFITSerializationSize stringSize) {
		int length = get(buffer, listSize);
		if (length < 0) return null;
		Collection<REFITByteString> collection = new ArrayList<>(length);
		for (int i = 0; i < length; i++) {
			REFITByteString string = getByteString(buffer, stringSize);
			collection.add(string);
		}
		return collection;
	}

	public static int getStringCollectionSize(Collection<REFITByteString> collection, REFITSerializationSize listSize, REFITSerializationSize stringSize) {
		int size = getSize(listSize);
		for (REFITByteString string : collection) {
			size += getByteStringSize(string, stringSize);
		}
		return size;
	}

}
