package refit.client;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import refit.client.policy.REFITClientPolicy;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITInitialConnectionsCompleted;
import refit.message.REFITInstruction.REFITInvocationInstruction;
import refit.replica.REFITReplicaGroups;
import refit.replica.order.REFITOrderGroups;
import refit.scheduler.REFITSchedulerGroup;
import refit.util.REFITIntervalStatistics;
import refit.util.REFITTime;


public class REFITLibByz extends REFITClientLibrary {

	private short replyReplicaID;
	private final REFITMessageAuthentication messageAuthentication;
	private final REFITReplyCertificate certificate;
	private ResultState resultState;
	private final Semaphore stableReplyLock;
	private boolean readWriteConflict;

	private boolean isConnected;
	private REFITInvocationInstruction delayedInstruction;

	private short primaryID;
	private final short groupID;
	private final int myLocation;
	private final int[] executorViews;
	private final boolean[] retryGroup;
	private final short retryGroupSize;

	private long nextSeqNr;
	private REFITRequest request;

	private long probeCounter;
	private final REFITProbeCertificate probeCert;


	public REFITLibByz(short clientID, REFITSchedulerGroup group, REFITIntervalStatistics statistics,
	                   REFITClientPolicy policy, boolean isFirstClient, LinkedBlockingQueue<Object> sidechannel) {
		super(clientID, group, statistics, policy, isFirstClient, sidechannel);

		short replyOffset = (short) (clientID % (REFITConfig.TOTAL_NR_OF_REPLICAS - (REFITConfig.USE_PASSIVE_REPLICATION ? REFITConfig.FAULTS_TO_TOLERATE : 0)));
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			boolean isPassive = false;
			for (short passiveReplica : REFITConfig.PASSIVE_REPLICAS) {
				if (i == passiveReplica) {
					isPassive = true;
					break;
				}
			}
			if (isPassive) continue;

			if (replyOffset > 0) {
				replyOffset--;
			} else {
				this.replyReplicaID = i;
				break;
			}
		}

		this.replyReplicaID = replyOffset;
		this.messageAuthentication = new REFITMessageAuthentication(clientID);
		myLocation = REFITReplicaGroups.getClientLocation(clientID);
//		REFITLogger.logEvent(this, "Created client at location " + myLocation);

		this.certificate = new REFITReplyCertificate(messageAuthentication, statistics, REFITReplicaGroups.getExecutionReplicas(0));
		this.stableReplyLock = new Semaphore(0);

		groupID = (REFITConfig.LOCATION_BASED_ORDER_GROUPS) ? REFITOrderGroups.getClientGroup(clientID)
				: weightedGroupSelect(REFITConfig.ORDER_CLIENT_WEIGHTS);
		primaryID = REFITOrderGroups.getGroupPrimary(groupID, REFITConfig.INITIAL_VIEW);
		if (REFITConfig.LOCATION_BASED_ORDER_GROUPS) {
			REFITLogger.logEvent(this, "Using group " + groupID);
			replyReplicaID = primaryID;
		}
		executorViews = new int[REFITConfig.EXEC_GROUP_SIZE];
		// this is usually identical to the normal group of ordering replicas
		retryGroup = new boolean[REFITConfig.TOTAL_NR_OF_REPLICAS];
		short rgs = 0;
		boolean[] orderGroup = REFITOrderGroups.getGroupByID(groupID);
		boolean[] execGroup = REFITReplicaGroups.getExecutionReplicas(0);
		for (int i = 0; i < retryGroup.length; i++) {
			retryGroup[i] = orderGroup[i] | execGroup[i];
			rgs += (retryGroup[i]) ? 1 : 0;
		}
		retryGroupSize = rgs;
		resultState = ResultState.NONE;

		this.nextSeqNr = 0L;
		this.invocationTimeStamp = -1L;
		this.isConnected = false;

		probeCert = new REFITProbeCertificate(messageAuthentication, REFITReplicaGroups.getExecutionReplicas(0));
		this.probeCounter = 0L;
	}

	private short weightedGroupSelect(short[] weights) {
		int totalWeights = 0;
		for (short weight : weights) {
			totalWeights += weight;
		}

		int pick = getNodeID() % totalWeights;
		for (short i = 0; i < weights.length; i++) {
			if (pick < weights[i]) {
				return i;
			}
			pick -= weights[i];
		}
		return 0;
	}

//	public void init() {
//		// Find out the initial sequence number for this client
//		try {
//			REFITRequest initRequest = prepareRequest(new byte[0]);
//			sendRequest(initRequest, true);
//			
//			long invocationTime = REFITTime.currentTimeMillis.getAsLong();
//			for(int i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
//				REFITReply reply = receiveReply(invocationTime + REFITConfig.CLIENT_REQUEST_TIMEOUT - REFITTime.currentTimeMillis.getAsLong());
//				if(reply == null) break;
//				if(!certificate.add(reply)) continue;
//				nextSeqNr = Math.max(nextSeqNr, ByteBuffer.wrap(reply.payload).getLong());
//			}
//		} catch(IOException ioe) {
//			ioe.printStackTrace();
//		}
//		nextSeqNr++;
//		if(REFITLogger.LOG_CLIENT) REFITLogger.logClient(this, "initial sequence number set to " + nextSeqNr);
//	}


	@Override
	public long getProgress() {
		return nextSeqNr;
	}

	@Override
	public String toString() {
		return String.format("LIBBYZ[%03d]", getNodeID());
	}


	// ######################
	// # CLIENT INTERACTION #
	// ######################

	private long invocationTimeStamp;

	@Override
	public ByteBuffer invoke(ByteBuffer request, long executeAfter, RequestMode mode) throws InterruptedException {
		// Insert invocation request
		invokeAsync(request, REFITConfig.ONLY_CONSISTENT ? -1 : executeAfter, REFITConfig.ONLY_CONSISTENT ? mode.toStrongMode() : mode);

		// Wait until reply is stable
		while (true) {
			boolean success = stableReplyLock.tryAcquire(REFITConfig.CLIENT_REQUEST_TIMEOUT, TimeUnit.MILLISECONDS);
			if (success) break;

			// Retry
			REFITInstruction invocation = new REFITInvocationInstruction();
			insertMessage(invocation);
		}

		// Return result; return null if just acked
		return (resultState.equals(ResultState.STABLE)) ? ((REFITReply) certificate.result).getPayload() : null;
	}

	@Override
	public void invokeAsync(ByteBuffer request, long executeAfter, RequestMode mode) {
		REFITInstruction invocation = new REFITInvocationInstruction(request, executeAfter, mode);
		insertMessage(invocation);
		invocationTimeStamp = REFITTime.currentTimeMillis.getAsLong();
	}

	@Override
	public REFITProbeData invokeProbe() throws InterruptedException {
		// Insert invocation request
		byte[] probeRiddle = messageAuthentication.generateOrderRiddle();
		short oldPrimaryID = primaryID;
		// FIXME alternatives: probes in parallel, ...
		primaryID = (short) ((probeCounter++) % REFITConfig.TOTAL_NR_OF_REPLICAS);

		invokeAsyncProbe(ByteBuffer.wrap(probeRiddle));

		// Wait until probe times out
		stableReplyLock.tryAcquire(2 * REFITConfig.PROBE_TIMEOUT, TimeUnit.MILLISECONDS);

		// FIXME this would be the place/time to retry probes with invalid messages
		REFITProbeData probeData = probeCert.getProbeData();

		primaryID = oldPrimaryID;
		return probeData;
	}

	private void invokeAsyncProbe(ByteBuffer request) {
		REFITInstruction invocation = new REFITInvocationInstruction(request, -1, RequestMode.PROBE);
		insertMessage(invocation);
		invocationTimeStamp = System.currentTimeMillis();
	}

	@Override
	public boolean isResultStable() {
		// Check whether result is stable
		boolean resultStable = certificate.isStable();
		if (resultStable) return true;

		// Check for retry
		long now = REFITTime.currentTimeMillis.getAsLong();
		if ((now - invocationTimeStamp) > REFITConfig.CLIENT_REQUEST_TIMEOUT) {
			// Retry
			REFITInstruction invocation = new REFITInvocationInstruction();
			insertMessage(invocation);
		}
		return false;
	}

	@Override
	public ByteBuffer getResult() {
		stableReplyLock.acquireUninterruptibly();
		// Return result; return null if just acked
		return (resultState.equals(ResultState.STABLE)) ? ((REFITReply) certificate.result).getPayload() : null;
	}

	@Override
	public boolean isReadWriteConflict() {
		return readWriteConflict;
	}

	@Override
	public long getAssignedSeqNr() {
		switch (resultState) {
		case NONE:
			return -1;
		case ACK:
		case STABLE:
			return certificate.result.sequenceNumber();
		}
		throw new InternalError("should never happen");
	}

	// #######################
	// # REPLICA INTERACTION #
	// #######################

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REPLY:
			handleReply((REFITReply) message);
			break;
		case ACK_REPLY:
			handleAck((REFITAckReply) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleReply(REFITReply reply) {
		// Ignore replies that arrive after the result has become stable
		if (request == null || !request.mode.equals(RequestMode.PROBE) && certificate.isStable()
				|| request.mode.equals(RequestMode.PROBE) && probeCert.isStable())
			return;

		if (!request.mode.equals(RequestMode.PROBE)) {
			// Add reply to certificate
			certificate.add(reply);
			if (!certificate.isStable()) {
				if (!certificate.isCompletable()) {
					// immediately trigger retry if the request can't be completed
					REFITInstruction invocation = new REFITInvocationInstruction();
					insertMessage(invocation);
				}
				return;
			}

			updateReplyReplica(certificate.getReplyTimes());
			if (!request.mode.equals(RequestMode.READ) && ((REFITReply) certificate.result).groupID == groupID) {
				updatePrimary(certificate.getPerExecutorViewId());
			}
		} else {
			// Add reply to certificate
			probeCert.add(reply);
			if (!probeCert.isStable()) return;
		}
		// Release lock
		resultState = ResultState.STABLE;
		stableReplyLock.release();
		if (client != null) client.progress();
	}

	private void updateReplyReplica(short[] replyTimes) {
		short fastestReplicaID = 0;
		short fastestReplyTime = replyTimes[0];
		for (short i = 0; i < replyTimes.length; i++) {
			if (replyTimes[i] < fastestReplyTime) {
				fastestReplicaID = i;
				fastestReplyTime = replyTimes[i];
			}
		}

		if (fastestReplyTime < REFITConfig.PROBE_TIMEOUT) {
			replyReplicaID = fastestReplicaID;
		}
	}

	private void updatePrimary(int[] perExecutorViewId) {
		for (int i = 0; i < perExecutorViewId.length; i++) {
			executorViews[i] = Math.max(executorViews[i], perExecutorViewId[i]);
		}

		int[] viewIds = Arrays.copyOf(executorViews, executorViews.length);
		Arrays.sort(viewIds);
		int mergedViewId = viewIds[viewIds.length / 2];
		short oldPrimary = primaryID;
		primaryID = REFITOrderGroups.getGroupPrimary(groupID, mergedViewId);
		if (primaryID != oldPrimary)
			REFITLogger.logWarning(this, String.format("Setting leader of group %s to %s%n", groupID, primaryID));
	}

	private void handleAck(REFITAckReply ack) {
		// Ignore ACKs once we have enough
		if (certificate.isStable()) return;

		// Add ACK to certificate
		certificate.add(ack);
		if (!certificate.isStable()) return;

		// Release lock
		resultState = ResultState.ACK;
		stableReplyLock.release();
		if (client != null) client.progress();
	}


	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case INVOCATION:
			handeInvocation(((REFITInvocationInstruction) instruction));
			break;
		case INITIAL_CONNECTIONS_COMPLETED:
			REFITInitialConnectionsCompleted init = (REFITInitialConnectionsCompleted) instruction;
			if (init.nodeID == getNodeID()) {
				handleInitCompleted();
			} else {
				completeDelayedConnect();
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	private void handleInitCompleted() {
		isConnected = true;
		if (delayedInstruction != null) {
			insertMessage(delayedInstruction);
			delayedInstruction = null;
		}
	}

	private void handeInvocation(REFITInvocationInstruction instruction) {
		// block instructions until we are connected
		if (!isConnected) {
			if (delayedInstruction == null) {
				delayedInstruction = instruction;
			}
			return;
		}

		if (instruction.request != null) {
			resultState = ResultState.NONE;
			// Prepare request
			REFITUniqueID uid = new REFITUniqueID(getNodeID(), nextSeqNr++);
			request = new REFITRequest(uid, getNodeID(), groupID, instruction.executeAfter, instruction.request, replyReplicaID, instruction.mode);

			if (instruction.mode.equals(RequestMode.PROBE)) {
				probeCert.init(uid, primaryID);
				probeCert.setRiddle(instruction.request.array());
			} else {
				certificate.init(uid, request.mode.isTotalOrder(), request.mode.equals(RequestMode.WRITE_ACK));
			}
			readWriteConflict = false;

			if (request.mode.isReadOnly() && (REFITConfig.USE_PBFT_READ_OPTIMIZATION || !request.mode.isTotalOrder())) {
				boolean[] executionReplicas = REFITReplicaGroups.getExecutionReplicas(0);
				multicastRequest(request, executionReplicas, REFITConfig.EXEC_GROUP_SIZE);
			} else {
				unicastRequestToPrimary(request);
			}
		} else if (request != null) {
			if (request.mode.equals(RequestMode.PROBE)) {
				throw new InternalError("Can't retry a probe! " + request);
			}
			// Retry; just resend normal requests
			if (!REFITConfig.USE_PBFT_READ_OPTIMIZATION && !request.mode.equals(RequestMode.READ))
				REFITLogger.logWarning(this, String.format("retry %s (%d replies received)", request, certificate.replies.getVoteCount()));

			if (!request.mode.isTotalOrder() || REFITConfig.USE_PBFT_READ_OPTIMIZATION && request.mode.equals(RequestMode.READ)) {
				// Send read-only as normal request instead, if we couldn't get a reply in time - doesn't need to be marked panic the first time
				REFITUniqueID uid = new REFITUniqueID(getNodeID(), nextSeqNr++);
				REFITLogger.logWarning(this, "Failed receiving reply to read-only request " + request + ", resending as read-write");
				// Reinitialize certificates
				certificate.init(uid, true, false);

				request = new REFITRequest(uid, getNodeID(), groupID, request.executeAfter, request.getPayload(), (short) -1, RequestMode.WRITE);
				readWriteConflict = true;

				// send first retry after mode change only to the primary
				unicastRequestToPrimary(request);
			} else {
				request = new REFITRequest(request.uid, getNodeID(), groupID, request.executeAfter, request.getPayload(), (short) -1, request.mode);
				request.markPanic();

				if (!REFITConfig.SIGNED_REQUESTS) {
					multicastRequest(request, retryGroup, retryGroupSize);
				} else {
					multicastRequest(request, REFITReplicaGroups.getExecutionReplicas(0), REFITConfig.EXEC_GROUP_SIZE);
				}
			}
		}
		progress();
	}

	private void unicastRequestToPrimary(REFITRequest request) {
		if (!REFITConfig.SIGNED_REQUESTS) {
			// Append MAC
			request.setRecipientCount(REFITConfig.ORDER_GROUP_SIZE);
			request.serializeMessage();
			messageAuthentication.appendMulticastMAC(request, REFITOrderGroups.getGroupByID(request.groupID));
			if (!REFITConfig.USE_SPINNING_LEADER) {
				// Send request to primary
				sendToReplica(request, primaryID);
			} else {
				short[] group = REFITConfig.ORDER_GROUPS[request.groupID];
				short replicaID = group[getNodeID() % group.length];
				sendToReplica(request, replicaID);
			}
		} else {
			request.serializeMessage();
			// use signatures to prevent executors from damaging the request's authenticator
			messageAuthentication.appendSignature(request);

			// wrap
			REFITRequestWrapper wrapper = new REFITRequestWrapper(request.uid, getNodeID(), request);
			wrapper.serializeMessage();
			messageAuthentication.appendUnicastMAC(wrapper, primaryID);

			sendToReplica(wrapper, primaryID);
		}
	}

	private void multicastRequest(REFITRequest request, boolean[] executionReplicas, short recipientCount) {
		if (!REFITConfig.SIGNED_REQUESTS) {
			// Append MAC, special case for executors
			request.setRecipientCount(recipientCount);
			request.serializeMessage();
			messageAuthentication.appendMulticastMAC(request, executionReplicas);
			// Send directly to the execution replicas
			sendToReplicas(request, executionReplicas);
		} else {
			// using MACs is sufficient as every executor processes the command on its own
			// add MAC to wrapper to unify code paths
			request.serializeMessage();
			messageAuthentication.appendNullMAC(request);
			for (short i = 0; i < executionReplicas.length; i++) {
				if (!executionReplicas[i]) continue;
				// wrap
				REFITRequestWrapper wrapper = new REFITRequestWrapper(request.uid, getNodeID(), request);
				wrapper.serializeMessage();
				messageAuthentication.appendUnicastMAC(wrapper, i);

				sendToReplica(wrapper, i);
			}
		}
	}

}
