package refit.client;

import java.nio.ByteBuffer;
import java.util.Arrays;

import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.message.*;
import refit.util.REFITPayload;


public class REFITProbeCertificate {

	private final REFITProbeResult[] receivedProbes = new REFITProbeResult[REFITConfig.TOTAL_NR_OF_REPLICAS];
	private final REFITMessageAuthentication messageAuthentication;
	private REFITUniqueID uid;
	private final boolean[] legitimateSenders;
	private long requestTime;
	private short[] replyTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
	private short leader;

	public REFITProbeCertificate(REFITMessageAuthentication messageAuthentication, boolean[] legitimateSenders) {
		this.messageAuthentication = messageAuthentication;
		this.legitimateSenders = legitimateSenders;
	}

	public void init(REFITUniqueID uid, short leader) {
		this.uid = uid;
		this.leader = leader;
		requestTime = System.currentTimeMillis();
		replyTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		Arrays.fill(replyTimes, REFITConfig.PROBE_TIMEOUT);
		Arrays.fill(receivedProbes, null);
	}

	public boolean add(REFITReply reply) {
		long replyTime = System.currentTimeMillis();
		// Check reply
		if (reply == null) return false;
		if (!uid.equals(reply.uid)) return false;
		if (!messageAuthentication.verifyUnicastMAC(reply, legitimateSenders)) return false;
		if (!reply.mode.equals(RequestMode.PROBE)) return false;

		REFITProbeResult probe = new REFITProbeResult(reply.getPayload());
		if (!verifyProbeResult(probe, reply.from)) return false;

		REFITProbeResult currentProbe = receivedProbes[reply.from];
		// verify stragglers
		if (currentProbe != null) {
			if (!verifyProbeUpdate(probe, currentProbe)) {
				return false;
			} else {
				receivedProbes[reply.from] = probe;
				return true;
			}
		}

		receivedProbes[reply.from] = probe;
		replyTimes[reply.from] = (short) (replyTime - requestTime);
		return true;
	}

	public short[] getReplyTimes() {
		short[] replyTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		short[] rawTimes = this.replyTimes;
		for (int i = 0; i < replyTimes.length; i++) {
			replyTimes[i] = (short) Math.max(0, Math.min(rawTimes[i], REFITConfig.PROBE_TIMEOUT));
		}
		return replyTimes;
	}

	public boolean isStable() {
		for (REFITProbeResult probe : receivedProbes) {
			if (probe == null || !probe.isComplete()) {
				return false;
			}
		}
		return true;
	}

	public REFITProbeData getProbeData() {
		short[] replyTimes = getReplyTimes();
		short[][] preparedTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS][];
		short[][] committedTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS][];

		for (int i = 0; i < receivedProbes.length; i++) {
			REFITProbeResult probe = receivedProbes[i];
			if (probe == null) {
				continue;
			}
			preparedTimes[i] = probe.preparedTimes;
			committedTimes[i] = probe.commitTimes;
		}

		short[] placeholder = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		Arrays.fill(placeholder, REFITConfig.PROBE_TIMEOUT);

		for (int i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (preparedTimes[i] == null) {
				preparedTimes[i] = placeholder;
			}
			if (committedTimes[i] == null) {
				committedTimes[i] = placeholder;
			}
		}

		return new REFITProbeData(leader, replyTimes, preparedTimes, committedTimes);
	}


	private byte[] probeRiddle;
	private byte[][] prepareRiddles;

	public void setRiddle(byte[] probeRiddle) {
		this.probeRiddle = probeRiddle;

		byte[] probeStepKey = messageAuthentication.calcStepKey(probeRiddle, 0, uid.nodeID, leader);
		byte[] preprepareRiddle = messageAuthentication.deriveDirectHashChain(probeStepKey, probeRiddle);

		prepareRiddles = new byte[REFITConfig.TOTAL_NR_OF_REPLICAS][];
		for (int i = 0; i < prepareRiddles.length; i++) {
			if (i == leader) {
				prepareRiddles[i] = preprepareRiddle;
			} else {
				byte[] prepareStepKey = messageAuthentication.calcStepKey(probeRiddle, 1, uid.nodeID, (short) i);
				prepareRiddles[i] = messageAuthentication.deriveDirectHashChain(prepareStepKey, preprepareRiddle);
			}
		}
	}

	private boolean verifyProbeResult(REFITProbeResult result, short i) {
		return verifyRiddle(result, i) && verifyProbeTimes(result, i);
	}

	private boolean verifyRiddle(REFITProbeResult result, short i) {
		int PREPARED_THRESHOLD = (REFITConfig.TOTAL_NR_OF_REPLICAS + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		int COMMITTED_THRESHOLD = (REFITConfig.TOTAL_NR_OF_REPLICAS + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;

		byte[] executeRiddle = result.executeRiddle;
		if (executeRiddle.length != REFITPayload.HASH_SIZE + (REFITConfig.TOTAL_NR_OF_REPLICAS + 1) * (Short.SIZE / 8)) {
			return false;
		}

		ByteBuffer bb = ByteBuffer.wrap(executeRiddle);
		bb.position(executeRiddle.length - 2);

		short commitChunkFlags = bb.getShort();
		if (!messageAuthentication.isReplicaInFlags(i, commitChunkFlags)) {
			// replica must not skip phases, thus it must use its own commit message
			return false;
		}
		int commitCount = messageAuthentication.countChunkFlags(commitChunkFlags);
		if (commitCount < COMMITTED_THRESHOLD) {
			return false;
		}

		bb.position(executeRiddle.length - 2 - REFITConfig.TOTAL_NR_OF_REPLICAS * 2);
		byte[][] commitRiddles = new byte[REFITConfig.TOTAL_NR_OF_REPLICAS][];
		for (short j = 0; j < commitRiddles.length; j++) {
			short prepareChunkFlags = bb.getShort();
			if ((commitChunkFlags & (1 << j)) == 0) {
				continue;
			}

			if (!messageAuthentication.isReplicaInFlags(leader, prepareChunkFlags)
					|| !messageAuthentication.isReplicaInFlags(j, prepareChunkFlags)) {
				return false;
			}
			int prepareCount = messageAuthentication.countChunkFlags(prepareChunkFlags);
			if (prepareCount < PREPARED_THRESHOLD) {
				return false;
			}

			byte[][] prepareInputRiddles = messageAuthentication.selectUsedRiddles(prepareRiddles, prepareChunkFlags);
			byte[] commitStepKey = messageAuthentication.calcStepKey(probeRiddle, 2, uid.nodeID, j);
			commitRiddles[j] = messageAuthentication.deriveHashChain(commitStepKey, prepareInputRiddles);
		}

		byte[] executeStepKey = messageAuthentication.calcStepKey(probeRiddle, 3, uid.nodeID, i);
		byte[] expectedExecuteRiddle = messageAuthentication.deriveHashChain(executeStepKey, commitRiddles);

		return Arrays.equals(executeRiddle, expectedExecuteRiddle);
	}

	private boolean verifyProbeTimes(REFITProbeResult result, short i) {
		byte[] executeRiddle = result.executeRiddle;
		ByteBuffer bb = ByteBuffer.wrap(executeRiddle);
		bb.position(executeRiddle.length - 2);
		short commitChunkFlags = bb.getShort();

		bb.position(executeRiddle.length - 2 - REFITConfig.TOTAL_NR_OF_REPLICAS * 2 + i * 2);
		short prepareChunkFlags = bb.getShort();

		for (short j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
			// verify that only prepare and commit times covered by riddles are included
			if (result.commitTimes[j] != REFITConfig.PROBE_TIMEOUT
					&& !messageAuthentication.isReplicaInFlags(j, commitChunkFlags)) {
				return false;
			}
			if (result.preparedTimes[j] != REFITConfig.PROBE_TIMEOUT
					&& !messageAuthentication.isReplicaInFlags(j, prepareChunkFlags)) {
				return false;
			}
		}
		return true;
	}

	private boolean verifyProbeUpdate(REFITProbeResult newResult, REFITProbeResult currentResult) {
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			// verify that only prepare and commit times covered by riddles are included
			if (currentResult.commitTimes[i] != newResult.commitTimes[i]
					&& currentResult.commitTimes[i] != REFITConfig.PROBE_TIMEOUT) {
				return false;
			}
			if (currentResult.preparedTimes[i] != newResult.preparedTimes[i]
					&& currentResult.preparedTimes[i] != REFITConfig.PROBE_TIMEOUT) {
				return false;
			}
		}
		return true;
	}
}
