package refit.config;

import java.io.File;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.*;

import refit.agreement.pbft.REFITPBFT.REFITPBFTProtocol;
import refit.agreement.pbft.REFITPBFTPassive.REFITPBFTPassiveProtocol;
import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.client.REFITBenchmarkRunnable;
import refit.client.REFITClientLibrary;
import refit.client.policy.REFITClientPolicy;
import refit.replica.REFITBaseReplica;
import refit.replica.order.REFITOrderProtocol;
import refit.util.REFITAssert;


public class REFITConfig {
	private static final REFITProperties config;

	static {
		REFITProperties defaults = REFITProperties.loadFile("scripts/config/refit-defaults");
		config = REFITProperties.loadFile("scripts/config/refit-overrides", defaults);
	}

	// Just for numbering experiment repetitions
	public static final int RUN = config.getInt("run_number");

	// Cell
	public static final short FAULTS_TO_TOLERATE = config.getShort("system.faults");
	public static final short TOTAL_NR_OF_REPLICAS = config.getShort("replica.count");
	public static final short ORDER_GROUP_SIZE;
	public static final short EXEC_GROUP_SIZE;
	public static final short TOTAL_NR_OF_CLIENTS = config.getShort("replica.max_clients");
	public static final short[] NR_OF_CLIENTS = config.getShortArray("client.count");
	public static final short CLIENT_FORCE_GROUP = config.getShort("client.force_group");
	// Spider; must run before INITIAL_ORDER_PROTOCOL is evaluated
	public static final boolean SEPARATE_EXECUTION_REPLICAS = config.getBoolean("system.spider.separate_execution_replicas");

	static {
		short orderGroupSize = config.getShort("replica.order_size");
		short execGroupSize = config.getShort("replica.exec_size");
		ORDER_GROUP_SIZE = (orderGroupSize > 0) ? orderGroupSize : TOTAL_NR_OF_REPLICAS;
		EXEC_GROUP_SIZE = (execGroupSize > 0) ? execGroupSize : TOTAL_NR_OF_REPLICAS;

		int sum = 0;
		for (short nr : NR_OF_CLIENTS) {
			sum += nr;
		}
		if (sum > TOTAL_NR_OF_CLIENTS) {
			throw new AssertionError("Trying to use more clients than available");
		}
	}

	// Intervals and timeouts
	public static final int CLIENT_REQUEST_TIMEOUT = config.getInt("client.timeout.request");
	public static final int REPLICA_REQUEST_TIMEOUT = config.getInt("replica.timeout.request");
	public static final int REPLICA_PANIC_INTERVAL = config.getInt("replica.timeout.panic_interval");

	// Client-replica & replica-replica communication
	public static final boolean AUTHENTICATE_MESSAGES = config.getBoolean("system.crypto.authenticate");
	public static final boolean USE_ED25519 = config.getBoolean("system.crypto.ed25519");
	public static final boolean SIGNED_REQUESTS = config.getBoolean("system.signed_requests");
	public static final boolean USE_HASHED_REPLIES = config.getBoolean("client.replies.hashed");
	public static final boolean USE_PBFT_READ_OPTIMIZATION = config.getBoolean("client.pbft_read_opt");
	public static final boolean AUTONOMOUS_PANICS = config.getBoolean("replica.client.autonomous_panics");

	// Passive replication
	public static final boolean USE_PASSIVE_REPLICATION = config.getBoolean("replica.passive");
	public static final short[] PASSIVE_REPLICAS = config.getShortArray("replica.passive.replicas");
	public static final boolean USE_HASHED_UPDATES = config.getBoolean("replica.updates.hashed");
	public static final int UPDATE_BATCH_INTERVAL = config.getInt("replica.updates.batch_interval");
	public static final int UPDATE_STABILITY_THRESHOLD = REFITConfig.FAULTS_TO_TOLERATE + 1;
	public static final REFITOrderProtocol INITIAL_ORDER_PROTOCOL = USE_PASSIVE_REPLICATION ? REFITPBFTPassiveProtocol.instance : REFITPBFTProtocol.instance;

	// Probe processing
	public static final int PROBE_PERCENTILE = config.getInt("replica.probe.percentile");
	public static final int PROBE_MAX_AGE;
	public static final long PROBE_WARMUP;
	public static final boolean APPLY_PROBING_RESULTS = config.getBoolean("replica.probe.apply");
	public static final short PROBING_DECISION_HYSTERESIS = config.getShort("replica.probe.hysteresis");
	public static final short PROBE_TIMEOUT = config.getShort("replica.timeout.probe");
	public static final short PROBE_INTERVAL = config.getShort("replica.timeout.probe_interval");

	static {
		PROBE_MAX_AGE = ((NR_OF_CLIENTS.length > 0) ? NR_OF_CLIENTS[0] : TOTAL_NR_OF_CLIENTS) * TOTAL_NR_OF_REPLICAS * config.getInt("replica.probe.age_factor");
		PROBE_WARMUP = ((NR_OF_CLIENTS.length > 0) ? NR_OF_CLIENTS[0] : TOTAL_NR_OF_CLIENTS) * TOTAL_NR_OF_REPLICAS * config.getInt("replica.probe.warmup_factor");
		if (REFITConfig.APPLY_PROBING_RESULTS && REFITConfig.USE_PASSIVE_REPLICATION) {
			throw new AssertionError("Automatic leader switching not yet supported for passive replication");
		}
	}

	// Checkpoints
	public static final int CHECKPOINT_INTERVAL = config.getInt("replica.checkpoint.interval");
	public static final int LIGHTWEIGHT_CHECKPOINT_STABILITY_THRESHOLD = REFITConfig.EXEC_GROUP_SIZE;
	// with signatures only f+1 checkpoints are required or when using the upright trick
	public static final int REGULAR_CHECKPOINT_STABILITY_THRESHOLD;
	public static final boolean ENABLE_FUZZY_CHECKPOINTS = config.getBoolean("replica.checkpoint.fuzzy");
	public static final boolean ENABLE_DIFFERENTIAL_CHECKPOINTING = config.getBoolean("replica.checkpoint.differential");
	// the standard checkpoint uses updates for passive replication
	public static final boolean UPDATE_BASED_CHECKPOINTING = config.getBoolean("replica.checkpoint.update_based")
			|| !ENABLE_FUZZY_CHECKPOINTS && !ENABLE_DIFFERENTIAL_CHECKPOINTING;
	public static final boolean DYNAMIC_FUZZY_CHECKPOINT_OFFSET = config.getBoolean("replica.checkpoint.dynamic_offset");
	public static final int FUZZY_CHECKPOINT_BUFFER = config.getInt("replica.checkpoint.fuzzy_buffer");
	public static final float FUZZY_CHECKPOINT_BUFFER_SCALE = config.getFloat("replica.checkpoint.fuzzy_scale");
	public static final int FUZZY_CHECKPOINT_OFFSET = (int) (CHECKPOINT_INTERVAL * config.getFloat("replica.checkpoint.fuzzy_rel_offset"));

	static {
		int checkpointThreshold = config.getBoolean("replica.checkpoint.quick_stability")
				? REFITConfig.FAULTS_TO_TOLERATE + 1 : 2 * REFITConfig.FAULTS_TO_TOLERATE + 1;
		if (!AUTHENTICATE_MESSAGES) checkpointThreshold = 1;
		REGULAR_CHECKPOINT_STABILITY_THRESHOLD = checkpointThreshold;
		if (!ENABLE_FUZZY_CHECKPOINTS && ENABLE_DIFFERENTIAL_CHECKPOINTING && UPDATE_BASED_CHECKPOINTING) {
			throw new AssertionError("UPDATE_BASED_CHECKPOINTING must be set to false");
		}
		if ((ENABLE_FUZZY_CHECKPOINTS || ENABLE_DIFFERENTIAL_CHECKPOINTING) && USE_PASSIVE_REPLICATION) {
			throw new AssertionError("Not implemented feature combination");
		}
	}

	// Coupling between order and execution stage
	public static final int EXECUTION_STABILITY_THRESHOLD = REFITConfig.FAULTS_TO_TOLERATE + 1;
	public static final boolean USE_HASHED_ORDERED_REQUEST = config.getBoolean("replica.coupling.hashed");
	public static final int NO_BATCHING_DELAY_WINDOW = config.getInt("replica.coupling.no_batching_delay_window");
	public static final long FLUSH_TIMEOUT = config.getInt("replica.coupling.flush_timeout");
	public static final boolean OPTIMIZED_AGREEMENT_AND_EXECUTION = config.getBoolean("replica.coupling.optimized");
	public static final boolean USE_SPINNING_LEADER = config.getBoolean("replica.coupling.spin_leader");

	// Order stage
	public static final short[][] ORDER_GROUPS;
	public static final short[] ORDER_GROUP_LEADER_LIMIT;
	public static final short[] ORDER_CLIENT_WEIGHTS;
	public static final int INITIAL_VIEW = config.getInt("replica.order.initial_view");
	public static final int MAXIMUM_ORDER_INSTANCES_IN_PROGRESS = (int) (CHECKPOINT_INTERVAL * config.getFloat("replica.order.instance_factor"));
	public static final int INSTANCES_AFTER_CHECKPOINT = (int) (CHECKPOINT_INTERVAL * config.getFloat("replica.order.instance_slack_factor"));
	public static final int MINIMUM_BATCH_SIZE = config.getInt("replica.order.min_batch");
	public static final int[] MAXIMUM_BATCH_SIZE = config.getIntArray("replica.order.max_batch");
	public static final int MAXIMUM_BATCH_BYTES = config.getInt("replica.order.max_batch_bytes");
	public static final boolean PREFER_FULL_BATCHES = config.getBoolean("replica.order.prefer_full_batches");
	public static final boolean BATCH_ALWAYS = config.getBoolean("replica.order.batch_always");
	public static final int ACTIVE_PROPOSALS = config.getInt("replica.order.active_proposals");
	public static final int[] WHEAT_WEIGHTS = config.getIntArray("replica.order.wheat_weights");

	static {
		if (INSTANCES_AFTER_CHECKPOINT <= CHECKPOINT_INTERVAL || INSTANCES_AFTER_CHECKPOINT > MAXIMUM_ORDER_INSTANCES_IN_PROGRESS) {
			throw new AssertionError("Invalid order window size");
		}
	}

	// Server throttling
	public static final boolean ENABLE_THROTTLING = config.getBoolean("replica.mac_throttling");
	// costs : 2 + (10f+2)/batch per Request
	public static final int[] MAX_REQUESTS_PER_SECOND = config.getIntArray("replica.mac_limit_per_second");

	// WEAVE
	public static final boolean LOCATION_BASED_ORDER_GROUPS = config.getBoolean("system.location_based_groups");
	public static final int PROPOSALS_INTERVAL_MS = config.getInt("replica.weave.proposal_interval_ms");
	public static final int NETWORK_PING_ITERATIONS = config.getInt("replica.weave.ping_iterations");
	public static final boolean MENCIUS_MODE = config.getBoolean("system.mencius_mode");

	// Spider
	public static final int MAXIMUM_EXECUTION_INSTANCES_IN_PROGRESS;
	public static final boolean IRMC_SENDER_COLLECTION = config.getBoolean("replica.spider.irmc_sender_collection");

	static {
		float execScale = config.getFloat("replica.execution.instance_factor");
		MAXIMUM_EXECUTION_INSTANCES_IN_PROGRESS = (execScale < 0) ? MAXIMUM_ORDER_INSTANCES_IN_PROGRESS : (int) (CHECKPOINT_INTERVAL * execScale);
	}

	// Debugging and statistics
	public static final boolean ENABLE_DEBUG_CHECKS = config.getBoolean("system.debug_checks");
	public static final boolean TRACE_AUTH = config.getBoolean("system.trace_auth");
	public static final boolean ENABLE_MESSAGE_TRACES = config.getBoolean("system.trace_messages");
	public static final boolean TRACK_SCHEDULER_HANGS = config.getBoolean("system.track_scheduler_hangs");
	public static final boolean TRACK_SLOW_STAGES = config.getBoolean("system.track_slow_stages");
	public static final boolean COLLECT_STAGE_STATISTICS = config.getBoolean("system.statistics.stage");
	public static final boolean COLLECT_MESSAGE_STATISTICS = config.getBoolean("replica.statistics.message");
	public static final boolean CLIENT_REPLY_STATISTICS = config.getBoolean("client.statistics.repliers");
	public static final boolean CLIENT_LATENCY_STATISTICS = config.getBoolean("client.statistics.latency");
	public static final int CLIENT_WARM_UP_SECONDS = config.getInt("client.statistics.warm_up_seconds");

	// Schedulers
	public static final int NR_OF_CLIENT_SCHEDULERS = config.getInt("client.schedulers");
	public static final int NR_OF_REPLICA_SCHEDULERS = config.getInt("replica.schedulers");

	// Benchmark workload
	public static final int REQUEST_SIZE = config.getInt("application.request_size");
	public static final int REPLY_SIZE = config.getInt("application.reply_size");
	public static final int UPDATE_SIZE = config.getInt("application.update_size");
	public static final int NR_OF_NODES = config.getInt("application.nodes");
	public static final int ACTIVE_NODES = (int) (NR_OF_NODES * config.getFloat("application.active_factor"));
	// enable updating of last access field in REZK on getData requests
	public static final boolean UPDATE_LAST_ACCESS = config.getBoolean("application.update_last_access");
	// 0 = write-only, 1 = read-only
	public static final float READ_WRITE_RATIO = config.getFloat("application.read_ratio");
	public static final boolean WAIT_FOR_FULL_REPLY = config.getBoolean("application.wait_write");
	public static final int GROUP_LATE_START_SEQNR = config.getInt("application.group.late_start_seqnr");
	public static final int GROUP_LATE_START_GID = config.getInt("application.group.late_start_group");

	public static final boolean ONLY_CONSISTENT = config.getBoolean("application.only_consistent");
	// Benchmark client execution policy
	public static final short[] POLICY_OPEN_LOOP_RATE = config.getShortArray("policy.open_loop.rate");

	// Counter
	public static final boolean RANDOM_DELAY_AFTER_REQUEST = false;

	// Benchmark applications
	// The benchmark classes MUST be loaded after the benchmark parameters, otherwise these can't be
	// accessed during initialization
	public static final Class<? extends REFITApplicationServer> APPLICATION_SERVER = config.getClass("application.server", REFITApplicationServer.class);
	public static final Class<? extends REFITSystemApplicationServer> SYSTEM_APPLICATION_SERVER = config.getClass("application.system.server", REFITSystemApplicationServer.class);
	public static final Class<? extends REFITBenchmarkRunnable> APPLICATION_CLIENT = config.getClass("application.client", REFITBenchmarkRunnable.class);
	public static final Class<? extends REFITClientLibrary> APPLICATION_CLIENT_LIBRARY = config.getClass("application.library", REFITClientLibrary.class);
	public static final Class<? extends REFITClientPolicy> APPLICATION_CLIENT_POLICY = config.getClass("application.policy", REFITClientPolicy.class);
	public static final Class<? extends REFITBaseReplica> APPLICATION_REPLICA;

	// Communication
	public static final int REPLICA_NETWORK_SEND_BUFFER_SIZE = 1024 * config.getInt("replica.network.buffer.send.kb");
	public static final int REPLICA_NETWORK_RECEIVE_BUFFER_SIZE = 1024 * config.getInt("replica.network.buffer.receive.kb");
	public static final int REPLICA_NETWORK_CONNECTORS = config.getInt("replica.network.connectors");
	public static final int CLIENT_NETWORK_RECEIVE_BUFFER_SIZE = 1024 * config.getInt("replica.network.buffer.client.kb");
	public static final int CLIENT_RECEIVE_BUFFER_SIZE = 1024 * config.getInt("client.network.buffer.receive.kb");
	public static final int CLIENT_SEND_BUFFER_SIZE = 1024 * config.getInt("client.network.buffer.send.kb");
	// use _per_ client
	public static final int CLIENT_INDIVIDUAL_NETWORK_CONNECTORS = config.getInt("client.network.individual_connectors");

	// (Log) output directory
	public static final String OUTPUT_DIRECTORY = REFITSystemHelper.getenv("OUTPUT_DIRECTORY", "tmp");
	// (Log) file prefix name
	public static final String OUTPUT_ID = REFITSystemHelper.getenv("OUTPUT_ID", "client-0");

	static {
		// ensure output directory exists
		//noinspection ResultOfMethodCallIgnored
		new File(OUTPUT_DIRECTORY).mkdir();
	}

	// #########################
	// # NETWORK CONFIGURATION #
	// #########################

	public static final int[][] EXPECTED_GROUP_NETWORK_DELAYS_MS;

	public static final SocketAddress[] ADDRESSES = new SocketAddress[TOTAL_NR_OF_REPLICAS + TOTAL_NR_OF_CLIENTS];
	public static final SocketAddress[] INT_ADDRESSES = new SocketAddress[TOTAL_NR_OF_REPLICAS + TOTAL_NR_OF_CLIENTS];
	public static final SocketAddress[] BIND_ADDRESSES = new SocketAddress[TOTAL_NR_OF_REPLICAS];
	public static final int[] LOCATION = new int[TOTAL_NR_OF_REPLICAS + TOTAL_NR_OF_CLIENTS];

	private static final String[] CLIENT_RAW_ADDRESSES = config.getStringArray("client.network.addresses");
	private static final String[] SERVER_RAW_LIST = config.getStringArray("replica.network.addresses");
	private static final String BIND_RAW_ADDRESS = config.getString("replica.network.bind_address");

	public static final boolean FAKE_WAN_DELAYS = config.getBoolean("system.fake_wan_delays");
	public static final int[][] INTER_LOCATION_DELAYS_MS;

	static {
		REFITProperties serversMap = REFITProperties.loadFile("scripts/config/servers");

		REFITAssert.assertTrue(SERVER_RAW_LIST.length >= TOTAL_NR_OF_REPLICAS, "Not enough replicas");
		for (int i = 0; i < TOTAL_NR_OF_REPLICAS; i++) {
			String rawAddress = serversMap.getString(SERVER_RAW_LIST[i]);
			ADDRESSES[i] = new InetSocketAddress(rawAddress, 8170 + i);
			String intAddress = serversMap.getOptionalString(SERVER_RAW_LIST[i] + "int");
			INT_ADDRESSES[i] = (intAddress != null) ? new InetSocketAddress(intAddress, 8170 + i) : ADDRESSES[i];
			BIND_ADDRESSES[i] = new InetSocketAddress(BIND_RAW_ADDRESS, 8170 + i);
			String loc = serversMap.getOptionalString(SERVER_RAW_LIST[i] + "loc");
			LOCATION[i] = (loc != null) ? Integer.parseInt(loc) : 0;

			if (REFITConfig.FAKE_WAN_DELAYS) {
				if (!rawAddress.equals("localhost") || (intAddress != null && !intAddress.equals("localhost")))
					throw new AssertionError("FAKE_WAN_DELAYS are ONLY allow for use with a local test setup!!!");
			}
		}

		for (String clientRawAddress : CLIENT_RAW_ADDRESSES) {
			serversMap.getString(clientRawAddress);
		}

		for (int i = 0; i < (ADDRESSES.length - TOTAL_NR_OF_REPLICAS); i++) {
			ADDRESSES[i + TOTAL_NR_OF_REPLICAS] = new InetSocketAddress(12345 + i);
		}

		int ofs = TOTAL_NR_OF_REPLICAS;
		for (int i = 0; i < NR_OF_CLIENTS.length; ++i) {
			String loc = serversMap.getOptionalString(CLIENT_RAW_ADDRESSES[i] + "loc");
			for (int j = 0; j < NR_OF_CLIENTS[i]; ++j) {
				LOCATION[ofs + j] = (loc != null) ? Integer.parseInt(loc) : 0;
			}
			ofs += NR_OF_CLIENTS[i];
		}

		Set<Integer> locations = new HashSet<>();
		int maxLoc = -1;
		for (int value : LOCATION) {
			REFITAssert.assertTrue(value >= 0 && value < 16, "Location value must be between 0 and 16 (exclusive)");
			maxLoc = Math.max(maxLoc, value);
			locations.add(value);
		}

		int[][] delays = null;
		if (FAKE_WAN_DELAYS || LOCATION_BASED_ORDER_GROUPS) {
			delays = new int[maxLoc + 1][maxLoc + 1];

			// Delay from yourself is always 0
			for (int i = 0; i < delays.length; i++) {
				Arrays.fill(delays[i], -1);
				delays[i][i] = 0;
			}

			if (serversMap.getOptionalString("latency_0_1") != null || FAKE_WAN_DELAYS) {
				for (int i = 0; i < delays.length; i++) {
					if (!locations.contains(i)) continue;
					for (int j = i + 1; j < delays.length; j++) {
						if (!locations.contains(j)) continue;
						int latency = Integer.parseInt(serversMap.getString("latency_" + i + "_" + j));
						delays[i][j] = delays[j][i] = latency;
					}
				}
			} else {
				// fill in placeholder values for order groups
				for (int i = 0; i < delays.length; i++) {
					for (int j = 0; j < delays[i].length; j++) {
						if (i != j) {
							delays[i][j] = 200;
						}
					}
				}
			}
		}
		INTER_LOCATION_DELAYS_MS = (FAKE_WAN_DELAYS) ? delays : null;
		EXPECTED_GROUP_NETWORK_DELAYS_MS = (LOCATION_BASED_ORDER_GROUPS) ? delays : null;

		short[][] groups = config.getShortArrayArray("replica.order.groups");
		short[] weights = config.getShortArray("replica.order.weights");
		if (groups.length > 0) {
			REFITAssert.assertTrue(!LOCATION_BASED_ORDER_GROUPS, "either specific order groups or use location-based groups");
			ORDER_GROUPS = groups;
			ORDER_CLIENT_WEIGHTS = weights;
			ORDER_GROUP_LEADER_LIMIT = new short[ORDER_GROUPS.length];
			Arrays.fill(ORDER_GROUP_LEADER_LIMIT, ORDER_GROUP_SIZE);
		} else if (!LOCATION_BASED_ORDER_GROUPS) {
			// create default order group if left empty
			ORDER_GROUPS = new short[][]{new short[ORDER_GROUP_SIZE]};
			for (short i = 0; i < ORDER_GROUP_SIZE; i++) {
				ORDER_GROUPS[0][i] = i;
			}
			ORDER_CLIENT_WEIGHTS = new short[ORDER_GROUPS.length];
			Arrays.fill(ORDER_CLIENT_WEIGHTS, (short) 1);
			ORDER_GROUP_LEADER_LIMIT = new short[ORDER_GROUPS.length];
			Arrays.fill(ORDER_GROUP_LEADER_LIMIT, ORDER_GROUP_SIZE);
		} else {
			if (ORDER_GROUP_SIZE != TOTAL_NR_OF_REPLICAS)
				throw new AssertionError("Location based order groups must contain all replicas");
			// create default order group if left empty
			ORDER_GROUPS = new short[maxLoc + 1][];
			ORDER_GROUP_LEADER_LIMIT = new short[ORDER_GROUPS.length];
			for (int i = 0; i < ORDER_GROUPS.length; i++) {
				ORDER_GROUPS[i] = new short[REFITConfig.ORDER_GROUP_SIZE];
				short ctr = 0;
				// first add replicas of the current location, then fill up with all other
				for (short j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
					if (LOCATION[j] == i) ORDER_GROUPS[i][ctr++] = j;
				}
				ORDER_GROUP_LEADER_LIMIT[i] = ctr;
				if (!REFITConfig.MENCIUS_MODE) {
					// Every group has enough replicas to tolerate f faults
					REFITAssert.assertTrue(ctr == FAULTS_TO_TOLERATE + 1);
				}
				for (short j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
					if (LOCATION[j] != i) ORDER_GROUPS[i][ctr++] = j;
				}
			}
			ORDER_CLIENT_WEIGHTS = new short[ORDER_GROUPS.length];
			Arrays.fill(ORDER_CLIENT_WEIGHTS, (short) 1);
		}

		REFITAssert.assertTrue(!REFITConfig.AUTHENTICATE_MESSAGES || REFITConfig.WAIT_FOR_FULL_REPLY,
				"ack replies are not yet implemented in the BFT protocol");

		APPLICATION_REPLICA = config.getClass("application.replica", REFITBaseReplica.class);
	}


	private static final Map<SocketAddress, Short> addresses = new HashMap<>();

	static {
		for (short i = 0; i < ADDRESSES.length; i++) {
			addresses.put(ADDRESSES[i], i);
		}
		for (short i = 0; i < INT_ADDRESSES.length; i++) {
			addresses.put((INT_ADDRESSES[i] != null) ? INT_ADDRESSES[i] : ADDRESSES[i], i);
		}
	}

	public static short getNodeByAddress(SocketAddress address) {
		return addresses.get(address);
	}

	static {
		config.assertNoUnusedKeys();
	}
}
