package refit.message;

import java.nio.ByteBuffer;

import refit.client.REFITClientLibrary.RequestMode;
import refit.replica.REFITOutbox;
import refit.replica.order.REFITOrderProtocol;

public abstract class REFITInstruction extends REFITMessage {

	// #####################
	// # INSTRUCTION TYPES #
	// #####################

	public enum REFITInstructionType {
		// Client
		INVOCATION,

		// Panic stage
		PANIC_NOTIFICATION,

		// Order stage
		CONFIGURATION_NOTIFICATION,
		EXECUTE_REQUEST,
		INTERNAL_VIEW_CHANGE_NOTIFICATION,

		// Coupling helper
		AWAIT_REQUEST,
		FLUSH_REQUEST,

		// Checkpoint stage
		PROGRESS_NOTIFICATION,

		// Execution stage
		SYSTEM_NOTIFICATION,
		EXECUTION_PROGRESS_NOTIFICATION,

		// Multichannel flow control
		CHANNEL_PROGRESS,
		CHANNEL_SKIP,
		CHANNEL_CONFIG,
		CHECKPOINT_ASSIST,

		// Network
		INITIAL_CONNECTIONS_COMPLETED,
		// Network connection (re-)established
		CONNECTION_RESET_NOTIFICATION,
		// Network back-pressure is gone
		CONNECTION_WRITEABLE_NOTIFICATION,
	}


	// ########################
	// # ABSTRACT INSTRUCTION #
	// ########################

	public final REFITInstructionType instructionType;


	private REFITInstruction(REFITInstructionType instructionType) {
		super(REFITMessageType.INSTRUCTION, null, (short) -1);
		this.instructionType = instructionType;
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.NOT_SERIALIZABLE;
	}

	@Override
	public String toString() {
		return String.format("{%-16s}", instructionType);
	}

	// #########################
	// # SPECIFIC INSTRUCTIONS #
	// #########################

	public static class REFITInvocationInstruction extends REFITInstruction {
		public final ByteBuffer request;
		public final long executeAfter;
		public final RequestMode mode;

		public REFITInvocationInstruction() {
			this(null, -1, null);
		}

		public REFITInvocationInstruction(ByteBuffer request, long executeAfter, RequestMode mode) {
			super(REFITInstructionType.INVOCATION);
			this.request = request;
			this.executeAfter = executeAfter;
			this.mode = mode;
		}
	}

	public static class REFITPanicNotification extends REFITInstruction {
		public REFITPanicNotification() {
			super(REFITInstructionType.PANIC_NOTIFICATION);
		}
	}

	public static class REFITInternalViewChangeNotification extends REFITInstruction {
		public REFITInternalViewChangeNotification() {
			super(REFITInstructionType.INTERNAL_VIEW_CHANGE_NOTIFICATION);
		}
	}

	public static class REFITFlushRequestNotification extends REFITInstruction {
		public final long flushSlotsBeforeSeqNr;
		public final long partialBatchingBeforeSeqNr;
		public final long proposeOnlyBeforeSeqNr;

		public REFITFlushRequestNotification(long flushSlotsBeforeSeqNr, long partialBatchingBeforeSeqNr) {
			this(flushSlotsBeforeSeqNr, partialBatchingBeforeSeqNr, -1);
		}

		public REFITFlushRequestNotification(long flushSlotsBeforeSeqNr, long partialBatchingBeforeSeqNr, long proposeOnlyBeforeSeqNr) {
			super(REFITInstructionType.FLUSH_REQUEST);
			this.flushSlotsBeforeSeqNr = flushSlotsBeforeSeqNr;
			this.partialBatchingBeforeSeqNr = partialBatchingBeforeSeqNr;
			this.proposeOnlyBeforeSeqNr = proposeOnlyBeforeSeqNr;
		}
	}

	public static class REFITConfigurationNotification extends REFITInstruction {
		public final REFITOrderProtocol protocol;
		public final int groupID;
		public final int viewID;

		public REFITConfigurationNotification(REFITOrderProtocol protocol, int groupID, int viewID) {
			super(REFITInstructionType.CONFIGURATION_NOTIFICATION);
			this.protocol = protocol;
			this.groupID = groupID;
			this.viewID = viewID;
		}
	}

	public static class REFITProgressNotification extends REFITInstruction {
		public final long agreementSeqNr;

		public REFITProgressNotification(long agreementSeqNr) {
			super(REFITInstructionType.PROGRESS_NOTIFICATION);
			this.agreementSeqNr = agreementSeqNr;
		}
	}

	public static class REFITSystemNotification extends REFITInstruction {
		public final int targetMinView;
		public final long targetAfterSeq;

		public REFITSystemNotification(int targetMinView, long targetAfterSeq) {
			super(REFITInstructionType.SYSTEM_NOTIFICATION);
			this.targetMinView = targetMinView;
			this.targetAfterSeq = targetAfterSeq;
		}
	}

	public static class REFITAwaitRequestNotification extends REFITInstruction {
		public final long agreementSeqNr;

		public REFITAwaitRequestNotification(long agreementSeqNr) {
			super(REFITInstructionType.AWAIT_REQUEST);
			this.agreementSeqNr = agreementSeqNr;
		}

		@Override
		public String toString() {
			return "{" + instructionType + "|" + agreementSeqNr + "}";
		}
	}

	public static class REFITExecuteRequest extends REFITInstruction {
		public final long agreementSeqNr;
		public final int viewID;
		public final REFITMessage message;
		public final Object attachment;

		public REFITExecuteRequest(long agreementSeqNr, int viewID, REFITMessage message, Object attachment) {
			super(REFITInstructionType.EXECUTE_REQUEST);
			this.agreementSeqNr = agreementSeqNr;
			this.viewID = viewID;
			this.message = message;
			this.attachment = attachment;
		}

		@Override
		public String toString() {
			return "{" + instructionType + "|" + agreementSeqNr + "|" + viewID + "}";
		}
	}

	public static class REFITChannelProgress extends REFITInstruction {
		public final short subChannelID;
		public final long baseSeqNr;
		public final long interestSeqNr;
		public final int tokens;

		// internal flow control back-pressure from receiver
		public REFITChannelProgress(short subChannelID, long baseSeqNr, long interestSeqNr, int tokens) {
			super(REFITInstructionType.CHANNEL_PROGRESS);
			this.subChannelID = subChannelID;
			this.baseSeqNr = baseSeqNr;
			this.interestSeqNr = interestSeqNr;
			this.tokens = tokens;
		}
	}

	public static class REFITChannelSkip extends REFITInstruction {
		public final short subChannelID;
		public final long baseSeqNr;

		// internal flow control sender-side seqNr skip request
		public REFITChannelSkip(short subChannelID, long baseSeqNr) {
			super(REFITInstructionType.CHANNEL_SKIP);
			this.subChannelID = subChannelID;
			this.baseSeqNr = baseSeqNr;
		}
	}

	public static class REFITChannelConfig extends REFITInstruction {
		public final int groupIdx;
		public final boolean isActive;

		public REFITChannelConfig(int groupIdx, boolean isActive) {
			super(REFITInstructionType.CHANNEL_CONFIG);
			this.groupIdx = groupIdx;
			this.isActive = isActive;
		}
	}

	public static class REFITCheckpointAssist extends REFITInstruction {
		public final int groupIdx;
		public final long minCheckpointSeqNr;
		public final short fullSender;

		public REFITCheckpointAssist(int groupIdx, long minCheckpointSeqNr, short fullSender) {
			super(REFITInstructionType.CHECKPOINT_ASSIST);
			this.groupIdx = groupIdx;
			this.minCheckpointSeqNr = minCheckpointSeqNr;
			this.fullSender = fullSender;
		}
	}

	public static class REFITConnectionWriteableNotification extends REFITInstruction {
		public final REFITOutbox outbox;
		public final short id;

		public REFITConnectionWriteableNotification(REFITOutbox outbox, short id) {
			super(REFITInstructionType.CONNECTION_WRITEABLE_NOTIFICATION);
			this.outbox = outbox;
			this.id = id;
		}
	}

	public static class REFITConnectionResetNotification extends REFITInstruction {
		public final short nodeID;

		public REFITConnectionResetNotification(short nodeID) {
			super(REFITInstructionType.CONNECTION_RESET_NOTIFICATION);
			this.nodeID = nodeID;
		}
	}

	public static class REFITInitialConnectionsCompleted extends REFITInstruction {

		public final short nodeID;

		public REFITInitialConnectionsCompleted(short nodeID) {
			super(REFITInstructionType.INITIAL_CONNECTIONS_COMPLETED);
			this.nodeID = nodeID;
		}

		@Override
		public String toString() {
			return "{" + instructionType + "|" + nodeID + "}";
		}
	}

	public static class REFITExecutionProgressNotification extends REFITInstruction {
		public final long nextSeqNr;

		public REFITExecutionProgressNotification(long nextSeqNr) {
			super(REFITInstructionType.EXECUTION_PROGRESS_NOTIFICATION);
			this.nextSeqNr = nextSeqNr;
		}
	}
}
