package refit.message;

import java.nio.ByteBuffer;
import java.security.MessageDigest;

import refit.config.REFITConfig;
import refit.util.REFITAssert;
import refit.util.REFITPayload;


public abstract class REFITMessage extends REFITBaseMessage {

	public enum AuthenticationType {
		MAC_AUTHENTICATOR,
		UNICAST_MAC,
		SIGNATURE,
		NONE,
		NOT_SERIALIZABLE
	}

	public final REFITMessageType type;
	public final REFITUniqueID uid;
	public final short from;
	private short recipientCount = -1;
	private final StackTraceElement[] ste;

	// A uid that refers to an agreement sequence number must either point at a specific request or update for that seq#
	// or in case of a checkpoint refer to the sequence number from which processing will continue
	// More general: a sequence number always points to or before a slot.
	// In a sequence number range the start seq# is included and the end seq# is excluded
	protected REFITMessage(REFITMessageType type, REFITUniqueID uid, short from) {
		// Initialize attributes
		this.type = type;
		this.uid = uid;
		this.from = from;
		this.fromInternal = true;

		ste = (REFITConfig.ENABLE_MESSAGE_TRACES) ? Thread.currentThread().getStackTrace() : null;
	}

	@Override
	public boolean equals(Object object) {
		if (this == object) return true;
		if (object == null || getClass() != object.getClass()) return false;
		REFITMessage other = (REFITMessage) object;
		return from == other.from &&
				type == other.type &&
				uid.equals(other.uid);
	}

	@Override
	public int hashCode() {
		return uid.hashCode();
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%s}", type, from, uid);
	}

	public StackTraceElement[] getSte() {
		return ste;
	}

	// ###################################
	// # SERIALIZATION & DESERIALIZATION #
	// ###################################

	// Header fields :                           type   +   uid.id   + uid.seqNr +    from
	private static final int HEADER_SIZE = Byte.BYTES + Short.BYTES + Long.BYTES + Short.BYTES;
	private static final int TYPE_POSITION = Integer.BYTES;


	protected REFITMessage(REFITMessageType expectedType, ByteBuffer buffer) {
		super(buffer);
		if (REFITConfig.ENABLE_DEBUG_CHECKS && getAuthenticationType().equals(AuthenticationType.NOT_SERIALIZABLE)) {
			throw new AssertionError("Not going to decode a non-serializable message");
		}
		this.type = REFITMessageType.getType(buffer.get());
		if (REFITConfig.ENABLE_DEBUG_CHECKS && type != expectedType) {
			throw new AssertionError("Mismatch between actual and expected message type. got: " + type + " expected: " + expectedType);
		}
		this.uid = new REFITUniqueID(buffer.getShort(), buffer.getLong());
		this.from = buffer.getShort();
		this.fromInternal = false;

		ste = (REFITConfig.ENABLE_MESSAGE_TRACES) ? Thread.currentThread().getStackTrace() : null;
	}

	public REFITMessageType groupType() {
		return type;
	}

	public short recipientCount() {
		return recipientCount;
	}

	public void setRecipientCount(short recipientCount) {
		if (this.recipientCount != -1) {
			throw new IllegalStateException("recipient count is already set");
		}
		this.recipientCount = recipientCount;
	}

	@Override
	public int getPaddingSize() {
		int nrOfPaddingBytes;
		switch (getAuthenticationType()) {
		case SIGNATURE:
			nrOfPaddingBytes = REFITMessageAuthentication.SIGNATURE_SIZE;
			recipientCount = 1;
			break;
		case UNICAST_MAC:
			nrOfPaddingBytes = REFITMessageAuthentication.UNICAST_MAC_SIZE;
			recipientCount = 1;
			break;
		case MAC_AUTHENTICATOR:
			if (recipientCount < 0) {
				recipientCount = REFITConfig.TOTAL_NR_OF_REPLICAS;
			}
			nrOfPaddingBytes = REFITMessageAuthentication.calculateMACSize(recipientCount);
			break;
		case NONE:
			nrOfPaddingBytes = 0;
			recipientCount = 0;
			break;
		case NOT_SERIALIZABLE:
		default:
			throw new InternalError("This message is not serializable!");
		}
		return nrOfPaddingBytes;
	}

	public static REFITMessage createMessage(ByteBuffer buffer) {
		REFITMessageType type = REFITMessageType.getType(buffer.get(buffer.position() + TYPE_POSITION));
		return type.createMessage(buffer);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		buffer.put(type.getMagic());
		buffer.putShort(uid.nodeID);
		buffer.putLong(uid.seqNr);
		buffer.putShort(from);
	}

	@Override
	protected int getHeaderSize() {
		return super.getHeaderSize() + HEADER_SIZE;
	}

	@Override
	protected int calculatePayloadSize() {
		return super.calculatePayloadSize() + HEADER_SIZE;
	}

	public byte[] getPadding() {
		ByteBuffer myBuffer = getBuffer();
		myBuffer.position(getMessageSize() - getPaddingSize());
		byte[] padding = new byte[getPaddingSize()];
		myBuffer.get(padding);
		return padding;
	}

	public void setPadding(byte[] padding) {
		REFITAssert.assertTrue(padding.length == getPaddingSize(), "Padding has wrong length");
		ByteBuffer myBuffer = getBuffer();
		myBuffer.position(getMessageSize() - getPaddingSize());
		myBuffer.put(padding);
	}

	// ##################
	// # AUTHENTICATION #
	// ##################

	private final transient boolean fromInternal;
	private transient boolean isVerified = false;

	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.NOT_SERIALIZABLE;
	}

	public boolean isInternal() {
		return fromInternal;
	}

	public boolean isVerified() {
		return isVerified;
	}

	public void markVerified() {
		this.isVerified = true;
	}


	// ###################
	// # MESSAGE CONTEXT #
	// ###################

	private transient REFITMessageContext context;

	public REFITMessageContext getContext() {
		if (context == null) context = new REFITMessageContext();
		return context;
	}


	// ################
	// # MESSAGE HASH #
	// ################

	private transient byte[] hash;


	public byte[] getHash() {
		if (hash == null) hash = calculateHash();
		return hash;
	}

	protected byte[] calculateHash() {
		// Prepare message buffer
		ByteBuffer messageBuffer = getInternalBuffer();
		messageBuffer.limit(calculatePayloadSize());

		// Calculate hash
		MessageDigest digest = REFITPayload.getDigest();
		digest.update(messageBuffer);
		return digest.digest();
	}

	protected void setHash(byte[] hash) {
		this.hash = hash;
	}
}
