package refit.message;

import java.nio.ByteBuffer;

import refit.agreement.pbft.REFITPBFTMessages.*;
import refit.agreement.steward.REFITStewardMessages;
import refit.agreement.weave.REFITWeaveMessages.REFITWeaveMessageCommit;
import refit.agreement.weave.REFITWeaveMessages.REFITWeaveMessagePrepare;
import refit.config.REFITLogger;
import refit.replica.checkpoint.REFITCheckpoint;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.coupling.REFITOrderedRequest;
import refit.replica.multichannel.REFITChannelAck;
import refit.replica.multichannel.REFITShare;
import refit.replica.multichannel.REFITShareBundle;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbeCommit;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrePrepare;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrepare;
import refit.replica.stewardorder.REFITStewardOrderMessages;
import refit.replica.update.REFITBatchUpdate;
import refit.replica.update.REFITUpdate;


public enum REFITMessageType {
	// NodeID exchange, must only be deserialized manually!
	CONNECT,

	// General types
	REQUEST,
	REQUEST_NO_OP,
	REQUEST_WRAPPER,
	BATCH,
	ACK_REPLY,
	REPLY,

	UPDATE,
	BATCH_UPDATE,
	CHECKPOINT,
	CHECKPOINT_STATE,

	INSTRUCTION,      // Internal message
	ORDER,            // Abstract message type
	FUZZY_CHECKPOINT_NOTIFICATION, // Internal message

	// Types for separating agreement from execution
	ORDERED_REQUEST,
	// Order stage load balancing
	EXECUTOR_FLUSH,

	// PBFT types
	PBFT_PRE_PREPARE,
	PBFT_PREPARE,
	PBFT_COMMIT,
	PBFT_VIEW_CHANGE,
	PBFT_NEW_VIEW,

	PROBE_ORDER, // Abstract message type
	PROBE_PBFT_PRE_PREPARE,
	PROBE_PBFT_PREPARE,
	PROBE_PBFT_COMMIT,

	// Weave types
	WEAVE_PREPARE,
	WEAVE_COMMIT,

	CHANNEL_MSG, // Abstract message type
	CHANNEL_ITEM, // actual item transferred along the irmc
	CHANNEL_ACK, // move_window from receiver-side
	CHANNEL_SENDER_ACK, // move_window from sender-size
	CHANNEL_LOCAL_ACK, // irmc-sc bundling
	CHANNEL_PROGRESS, // irmc-sc progress
	CHANNEL_SHARE,
	CHANNEL_SHARE_PART,
	CHANNEL_SHARE_BUNDLE,

	STEWARD_PRE_PREPARE,
	STEWARD_PREPARE,
	STEWARD_COMMIT,

	STEWARD_WAN, // Abstract message type
	STEWARD_PROPOSAL,
	STEWARD_ACCEPT_SHARE,
	STEWARD_ACCEPT,

	MICRO,
	;

	public static final REFITMessageType[] values = values();

	public static REFITMessageType getType(byte magic) {
		return values[magic];
	}

	public byte getMagic() {
		return (byte) ordinal();
	}

	public REFITMessage createMessage(ByteBuffer buffer) {
		switch (this) {
		case REQUEST:
			return new REFITRequest(buffer);
		case REQUEST_NO_OP:
			return new REFITNoOpRequest(buffer);
		case ACK_REPLY:
			return new REFITAckReply(buffer);
		case REPLY:
			return new REFITReply(buffer);
		case UPDATE:
			return new REFITUpdate(buffer);
		case CHECKPOINT:
			return new REFITCheckpoint(buffer);
		case CHECKPOINT_STATE:
			return new REFITCheckpointState(buffer);
		case BATCH:
			return new REFITBatch(buffer);
		case BATCH_UPDATE:
			return new REFITBatchUpdate(buffer);
		case ORDERED_REQUEST:
			return new REFITOrderedRequest(buffer);
		case EXECUTOR_FLUSH:
			return new REFITExecutorFlushMessage(buffer);
		case PBFT_PRE_PREPARE:
			return new REFITPBFTPrePrepare(buffer);
		case PBFT_PREPARE:
			return new REFITPBFTPrepare(buffer);
		case PBFT_COMMIT:
			return new REFITPBFTCommit(buffer);
		case PBFT_VIEW_CHANGE:
			return new REFITPBFTViewChange(buffer);
		case PBFT_NEW_VIEW:
			return new REFITPBFTNewView(buffer);
		case PROBE_PBFT_PRE_PREPARE:
			return new REFITProbePrePrepare(buffer);
		case PROBE_PBFT_PREPARE:
			return new REFITProbePrepare(buffer);
		case PROBE_PBFT_COMMIT:
			return new REFITProbeCommit(buffer);
		case WEAVE_PREPARE:
			return new REFITWeaveMessagePrepare(buffer);
		case WEAVE_COMMIT:
			return new REFITWeaveMessageCommit(buffer);
		case REQUEST_WRAPPER:
			return new REFITRequestWrapper(buffer);
		case CHANNEL_SHARE:
			return new REFITShare(REFITMessageType.CHANNEL_SHARE, buffer);
		case CHANNEL_SHARE_PART:
			return new REFITShare(REFITMessageType.CHANNEL_SHARE_PART, buffer);
		case CHANNEL_SHARE_BUNDLE:
			return new REFITShareBundle(buffer);
		case CHANNEL_ACK:
			return new REFITChannelAck(REFITMessageType.CHANNEL_ACK, buffer);
		case CHANNEL_LOCAL_ACK:
			return new REFITChannelAck(REFITMessageType.CHANNEL_LOCAL_ACK, buffer);
		case CHANNEL_SENDER_ACK:
			return new REFITChannelAck(REFITMessageType.CHANNEL_SENDER_ACK, buffer);
		case CHANNEL_PROGRESS:
			return new REFITChannelAck(REFITMessageType.CHANNEL_PROGRESS, buffer);
		case STEWARD_PRE_PREPARE:
			return new REFITStewardMessages.REFITStewardPrePrepare(buffer);
		case STEWARD_PREPARE:
			return new REFITStewardMessages.REFITStewardPrepare(buffer);
		case STEWARD_COMMIT:
			return new REFITStewardMessages.REFITStewardCommit(buffer);
		case STEWARD_PROPOSAL:
			return new REFITStewardOrderMessages.REFITStewardProposal(buffer);
		case STEWARD_ACCEPT_SHARE:
			return new REFITStewardOrderMessages.REFITStewardAcceptShare(buffer);
		case STEWARD_ACCEPT:
			return new REFITStewardOrderMessages.REFITStewardAccept(buffer);
		case MICRO:
			return new REFITMicro(buffer);
		default:
			// internal and abstract message types are not listed as they must not be sent over network!
			REFITLogger.logError(REFITMessageType.class.getSimpleName(), "cannot create a message object for type " + this);
			return null;
		}
	}

}
