package refit.message;

import java.nio.ByteBuffer;

import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.util.REFITAssert;

public class REFITRequest extends REFITOrderRequest {

	private static final int REQUEST_HEADER_SIZE = Byte.BYTES + Long.BYTES + Short.BYTES + Short.BYTES + Short.BYTES;

	public final long executeAfter;
	public boolean isPanic;
	public final RequestMode mode;
	public final short groupID;
	public final short replyReplicaID;
	private ByteBuffer payload;

	public REFITRequest(REFITUniqueID uid, short from, short groupID, long executeAfter, ByteBuffer payload, short replyReplicaID, RequestMode mode) {
		super(REFITMessageType.REQUEST, uid, from);
		this.isPanic = false;
		this.mode = mode;
		this.groupID = groupID;
		this.executeAfter = executeAfter;
		this.replyReplicaID = replyReplicaID;
		this.payload = payload;
	}

	public REFITRequest(ByteBuffer buffer) {
		super(REFITMessageType.REQUEST, buffer);
		byte flags = buffer.get();
		this.isPanic = ((flags & 1) != 0);
		int modeIdx = (flags >> 1) & 0xf;
		this.mode = RequestMode.values()[modeIdx];
		this.groupID = buffer.getShort();
		this.executeAfter = buffer.getLong();
		this.replyReplicaID = buffer.getShort();
		short payloadSize = buffer.getShort();
		this.payload = buffer.slice();
		this.payload.limit(payloadSize);
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		if (REFITConfig.SIGNED_REQUESTS) {
			// getAuthenticationType() is called during the initialization when mode is not set yet
			if (mode == null) return AuthenticationType.NONE;
			if (mode.isTotalOrder()) {
				return AuthenticationType.SIGNATURE;
			} else {
				return AuthenticationType.NONE;
			}
		} else {
			return AuthenticationType.MAC_AUTHENTICATOR;
		}
	}

	public void markPanic() {
		isPanic = true;
	}

	public ByteBuffer getPayload() {
		return payload.slice();
	}

	@Override
	protected byte[] calculateHash() {
		return super.calculateHash();
	}

	@Override
	public boolean equals(Object object) {
		if (this == object) return true;
		if (object == null || getClass() != object.getClass()) return false;
		if (!super.equals(object)) return false;
		REFITRequest other = (REFITRequest) object;
		if (mode != other.mode) return false;
		return payload.equals(other.payload);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		int modeIdx = mode.ordinal();
		REFITAssert.assertTrue(modeIdx < 16, "Only up to 16 request modes are supported");
		buffer.put((byte) ((isPanic ? 1 : 0) + (modeIdx << 1)));
		buffer.putShort(groupID);
		buffer.putLong(executeAfter);
		buffer.putShort(replyReplicaID);
		int payloadSize = payload.remaining();
		buffer.putShort((short) payloadSize);
		// replace command ByteBuffer after serialization to store the data only once
		ByteBuffer newPayload = buffer.slice();
		buffer.put(payload);
		newPayload.limit(payloadSize);
		payload = newPayload;
	}

	@Override
	public void relocateTo(ByteBuffer target) {
		super.relocateTo(target);
		ByteBuffer buffer = getInternalBuffer();
		buffer.position(getHeaderSize());
		ByteBuffer newPayload = buffer.slice();
		newPayload.limit(payload.limit());
		payload = newPayload;
	}

	@Override
	protected int getHeaderSize() {
		return (super.getHeaderSize() + REQUEST_HEADER_SIZE);
	}

	@Override
	protected int calculatePayloadSize() {
		return (super.calculatePayloadSize() + REQUEST_HEADER_SIZE + payload.remaining());
	}
}
