package refit.replica;

import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.message.REFITMessageType;
import refit.replica.checkpoint.REFITCheckpointStage;
import refit.replica.client.REFITClientStage;
import refit.replica.execution.REFITExecutionStage;
import refit.replica.execution.REFITExecutionStageBase;
import refit.replica.execution.REFITExecutionStubStage;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointStage;
import refit.replica.order.REFITOrderGroups;
import refit.replica.order.REFITOrderStage;
import refit.replica.probe.REFITProbeStage;
import refit.replica.update.REFITUpdateStage;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITReplica extends REFITBaseReplica {
	private final REFITClientStage clientStage;
	private final REFITOrderStage orderStage;
	private final REFITProbeStage probeStage;

	private final REFITUpdateStage updateStage;
	private final REFITCheckpointStage checkpointStage;
	private final REFITExecutionStageBase executionStage;
	private final REFITFuzzyCheckpointStage fuzzyCheckpointStage;

	static {
		if (REFITOrderGroups.COUNT > 1) {
			throw new AssertionError("The base configuration only supports a single order group!");
		}
	}

	public REFITReplica(short id, boolean isSingleThreaded) {
		super(id, isSingleThreaded, true);

		// Create global stages
		if (id < REFITConfig.EXEC_GROUP_SIZE) {
			// create and initialize application
			final REFITApplicationServer application = createApplication();
			application.init();
			final REFITSystemApplicationServer systemApplication = createSystemApplication();
			systemApplication.init();

			this.executionStage = new REFITExecutionStage(this, application, systemApplication);
			this.fuzzyCheckpointStage = (REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || REFITConfig.ENABLE_FUZZY_CHECKPOINTS)
					? new REFITFuzzyCheckpointStage(this, application) : null;
		} else {
			this.executionStage = new REFITExecutionStubStage(this);
			this.fuzzyCheckpointStage = null;
		}
		this.updateStage = REFITConfig.USE_PASSIVE_REPLICATION ? new REFITUpdateStage(this, REFITReplicaGroups.getExecutionReplicas(0).clone()) : null;
		if (updateStage != null) {
			addToInbound(REFITMessageType.UPDATE, updateStage);
			addToInbound(REFITMessageType.BATCH_UPDATE, updateStage);
		}
		this.checkpointStage = new REFITCheckpointStage(this, REFITReplicaGroups.getExecutionReplicas(0),
				REFITReplicaGroups.getAllReplicas(), null);
		addToInbound(REFITMessageType.CHECKPOINT, checkpointStage);
		addToInbound(REFITMessageType.CHECKPOINT_STATE, checkpointStage);
		this.probeStage = new REFITProbeStage(this);
		addToInbound(REFITMessageType.PROBE_ORDER, probeStage);

		// also setup client and coupling stage on execution replicas
		clientStage = new REFITClientStage(this, 0);
		addToClientInbound(clientStage);
		executionStage.replySenders[0].addListener(clientStage);

		orderStage = new REFITOrderStage(this, REFITConfig.INITIAL_ORDER_PROTOCOL, (short) 0, null);
		addToInbound(REFITMessageType.ORDER, orderStage);
		executionStage.systemSender.addListener(orderStage);

		clientStage.requestSender.addListener(orderStage);
		clientStage.readRequestSender.addListener(executionStage);
		clientStage.panicSender.addListener(orderStage);
		clientStage.probeSender.addListener(probeStage);
		executionStage.lastSeqNrSender.addListener(clientStage);

		orderStage.requestSender.addListener(executionStage);
		orderStage.ackSender.addListener(clientStage);

		orderStage.viewChangeSender.addListener(clientStage);
		orderStage.viewChangeSender.addListener(executionStage);
		if (updateStage != null) {
			orderStage.viewChangeSender.addListener(updateStage);
		}
		orderStage.internalViewChangeSender.addListener(clientStage);

		checkpointStage.progressSender.addListener(orderStage);

		executionStage.checkpointSender.addListener(checkpointStage);
		checkpointStage.progressSender.addListener(executionStage);
		checkpointStage.stableCPSender.addListener(executionStage);
		checkpointStage.stableCPStateSender.addListener(executionStage);

		if (fuzzyCheckpointStage != null) {
			executionStage.updateSender.addListener(fuzzyCheckpointStage);
			executionStage.fuzzyCheckpointNotificationSender.addListener(fuzzyCheckpointStage);

			fuzzyCheckpointStage.fuzzyCheckpointNotificationSender.addListener(executionStage);
			fuzzyCheckpointStage.checkpointSender.addListener(checkpointStage);
		}

		if (updateStage != null) {
			checkpointStage.progressSender.addListener(updateStage);
			executionStage.updateSender.addListener(updateStage);
			updateStage.stableUpdateSender.addListener(executionStage);
		}

		// treat message as directly received from a client
		if (REFITConfig.SIGNED_REQUESTS) {
			addClientInbound(REFITMessageType.REQUEST_WRAPPER);
		} else {
			addClientInbound(REFITMessageType.REQUEST);
		}
	}

	@Override
	protected void assignTasks(REFITScheduler[] schedulers) {
		// Assign tasks to schedulers
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.NETWORK_ENDPOINT);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.ORDER_STAGE);
		schedulers[2 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHECKPOINT_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXECUTION_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.UPDATE_STAGE);
		schedulers[4 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.FUZZY_CHECKPOINT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.PROBE_STAGE);
	}
}
