package refit.replica;

import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.message.REFITBatch;
import refit.message.REFITMessageType;
import refit.replica.checkpoint.REFITCheckpointStage;
import refit.replica.client.REFITClientStage;
import refit.replica.execclient.REFITExecClientStage;
import refit.replica.execution.REFITExecutionStage;
import refit.replica.mcclient.REFITMCClientStage;
import refit.replica.mcclient.REFITMCMultiRemoteClientStage;
import refit.replica.mccoupling.REFITMCCouplingStage;
import refit.replica.mccoupling.REFITMCExecutorCouplingStage;
import refit.replica.multichannel.*;
import refit.replica.order.REFITOrderGroups;
import refit.replica.order.REFITOrderStage;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITAssert;

public class REFITSpiderReplica extends REFITBaseReplica {
	private final REFITClientStage clientStage;
	private final REFITCheckpointStage checkpointStage;
	private final REFITOrderStage orderStage;
	private final REFITMCCouplingStage couplingStage;

	private final REFITExecClientStage execClientStage;
	private final REFITMCExecutorCouplingStage execCouplingStage;
	private final REFITExecutionStage execExecutionStage;

	private final REFITAbstractChannelSender requestChannelSenderStage;
	private final REFITAbstractChannelReceiver[] requestChannelReceiverStage;
	private final REFITMCClientStage mcClientStage;
	private final REFITMCMultiRemoteClientStage multiRemoteClientStage;

	private final REFITAbstractChannelSender orderedChannelSenderStage;
	private final REFITAbstractChannelReceiver orderedChannelReceiverStage;

	static {
		// partial configuration sanity check
		REFITAssert.assertTrue(REFITConfig.AUTONOMOUS_PANICS && REFITConfig.SEPARATE_EXECUTION_REPLICAS
						&& REFITConfig.BATCH_ALWAYS && REFITOrderGroups.COUNT == 1 && REFITConfig.SIGNED_REQUESTS,
				"Incompatible configuration");
	}

	public REFITSpiderReplica(short id, boolean isSingleThreaded) {
		super(id, isSingleThreaded, (id >= REFITConfig.ORDER_GROUP_SIZE));

		boolean isExecutor = (id >= REFITConfig.ORDER_GROUP_SIZE);
		int executorID = REFITReplicaGroups.getExecutionGroupId(id);

		final int EXEC_TOKENS = REFITConfig.MAXIMUM_EXECUTION_INSTANCES_IN_PROGRESS;
		final int EXEC_HISTORY = 0; // REFITConfig.CHECKPOINT_INTERVAL;
		final int CLIENT_TOKENS = 2; //REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
		final int CLIENT_HISTORY = 0; //REFITConfig.CHECKPOINT_INTERVAL;

		if (isExecutor) {
			// create and initialize application
			final REFITApplicationServer application = createApplication();
			application.init();

			final REFITSystemApplicationServer systemApplication = createSystemApplication();
			systemApplication.init();

			// Create global stages
			execExecutionStage = new REFITExecutionStage(this, application, systemApplication);
			boolean[] execGroup = REFITReplicaGroups.getExecutionReplicas(REFITReplicaGroups.getExecutionGroupId(id));
			boolean[] bootstrapGroup = (REFITReplicaGroups.getExecutionGroupId(id) == REFITConfig.GROUP_LATE_START_GID) ? REFITReplicaGroups.getExecutionReplicas(0) : null;
			checkpointStage = new REFITCheckpointStage(this, execGroup.clone(), execGroup.clone(), bootstrapGroup);
			addToInbound(REFITMessageType.CHECKPOINT, checkpointStage);
			addToInbound(REFITMessageType.CHECKPOINT_STATE, checkpointStage);
			execExecutionStage.checkpointSender.addListener(checkpointStage);
			checkpointStage.progressSender.addListener(execExecutionStage);
			checkpointStage.stableCPSender.addListener(execExecutionStage);
			checkpointStage.stableCPStateSender.addListener(execExecutionStage);

			execCouplingStage = new REFITMCExecutorCouplingStage(this);
			execCouplingStage.requestSender.addListener(execExecutionStage);
			execCouplingStage.configSender.addListener(execExecutionStage);
			checkpointStage.progressSender.addListener(execCouplingStage);
			execCouplingStage.assistSender.addListener(checkpointStage);

			if (REFITConfig.IRMC_SENDER_COLLECTION) {
				orderedChannelReceiverStage = new REFITSourceChannelReceiverStage(this, REFITOrderGroups.getGroupByID(0),
						REFITReplicaGroups.getExecutionGroupId(id), (short) 1, EXEC_TOKENS, REFITConfig.FAULTS_TO_TOLERATE);
			} else {
				orderedChannelReceiverStage = new REFITChannelReceiverStage(this, REFITOrderGroups.getGroupByID(0),
						REFITReplicaGroups.getExecutionGroupId(id), (short) 1, EXEC_TOKENS, REFITConfig.FAULTS_TO_TOLERATE);
			}
			addToInbound(REFITMessageType.CHANNEL_SHARE, orderedChannelReceiverStage);
			addToInbound(REFITMessageType.CHANNEL_SHARE_BUNDLE, orderedChannelReceiverStage);
			addToInbound(REFITMessageType.CHANNEL_PROGRESS, orderedChannelReceiverStage);
			addToInbound(REFITMessageType.CHANNEL_SENDER_ACK, orderedChannelReceiverStage);
			orderedChannelReceiverStage.resultSender.addListener(execCouplingStage);
			orderedChannelReceiverStage.skipSender.addListener(execCouplingStage);
			execCouplingStage.progressSender.addListener(orderedChannelReceiverStage);

			execClientStage = new REFITExecClientStage(this);
			addToClientInbound(execClientStage);
			// treat message as directly received from a client
			addToInbound(REFITMessageType.REQUEST, execClientStage);
			execClientStage.readRequestSender.addListener(execExecutionStage);
			execExecutionStage.replySenders[0].addListener(execClientStage);

			mcClientStage = new REFITMCClientStage(this);
			execClientStage.requestSender.addListener(mcClientStage);

			boolean[][] orderGroups = new boolean[1][];
			orderGroups[0] = REFITOrderGroups.getGroupByID(0);

			if (REFITConfig.IRMC_SENDER_COLLECTION) {
				requestChannelSenderStage = new REFITSourceChannelSenderStage(this, execGroup.clone(), orderGroups, REFITConfig.TOTAL_NR_OF_CLIENTS,
						CLIENT_TOKENS, CLIENT_HISTORY, null, REFITConfig.FAULTS_TO_TOLERATE, REFITConfig.FAULTS_TO_TOLERATE);
			} else {
				requestChannelSenderStage = new REFITChannelSenderStage(this, orderGroups, REFITConfig.TOTAL_NR_OF_CLIENTS,
						CLIENT_TOKENS, CLIENT_HISTORY, null, REFITConfig.FAULTS_TO_TOLERATE);
			}
			addToInbound(REFITMessageType.CHANNEL_ACK, requestChannelSenderStage);
			addToInbound(REFITMessageType.CHANNEL_SHARE_PART, requestChannelSenderStage);
			addToInbound(REFITMessageType.CHANNEL_LOCAL_ACK, requestChannelSenderStage);
			mcClientStage.channelRequestSender.addListener(requestChannelSenderStage);
			mcClientStage.channelSkipSender.addListener(requestChannelSenderStage);
			requestChannelSenderStage.progressSender.addListener(mcClientStage);

			orderStage = null;
			couplingStage = null;
			clientStage = null;

			requestChannelReceiverStage = null;
			orderedChannelSenderStage = null;
			multiRemoteClientStage = null;
		} else {
			// checkpoint stage only communicates withing ordering group
			boolean[] checkpointReceivers = REFITOrderGroups.getGroupByID(0);
			checkpointStage = new REFITCheckpointStage(this, checkpointReceivers, checkpointReceivers, null);
			addToInbound(REFITMessageType.CHECKPOINT, checkpointStage);
			addToInbound(REFITMessageType.CHECKPOINT_STATE, checkpointStage);

			orderStage = new REFITOrderStage(this, REFITConfig.INITIAL_ORDER_PROTOCOL, (short) 0, null);
			addToInbound(REFITMessageType.ORDER, orderStage);

			couplingStage = new REFITMCCouplingStage(this);
			orderStage.requestSender.addListener(couplingStage);
			couplingStage.checkpointSender.addListener(checkpointStage);
			checkpointStage.stableCPStateSender.addListener(couplingStage);
			// order stage is throttled with the progressNotification
			couplingStage.progressSender.addListener(orderStage);

			boolean[][] groups = new boolean[REFITReplicaGroups.EXEC_COUNT][];
			for (int i = 0; i < groups.length; i++) {
				groups[i] = REFITReplicaGroups.getExecutionReplicas(i);
			}
			if (REFITConfig.IRMC_SENDER_COLLECTION) {
				orderedChannelSenderStage = new REFITSourceChannelSenderStage(this, REFITOrderGroups.getGroupByID(0), groups, (short) 1,
						EXEC_TOKENS, EXEC_HISTORY, (bundle, group) -> {
					if (!(bundle.getContent() instanceof REFITBatch)) {
						return bundle;
					}
					int groupID = -1;
					for (int i = 0; i < group.length; i++) {
						if (!group[i]) continue;
						groupID = REFITReplicaGroups.getExecutionGroupId((short) i);
						break;
					}
					final short gid = (short) groupID;
					return bundle.toFilteredBundle((m) -> ((REFITBatch) m).toGroupOnly(gid));
				}, REFITConfig.FAULTS_TO_TOLERATE, REFITConfig.FAULTS_TO_TOLERATE);
			} else {
				orderedChannelSenderStage = new REFITChannelSenderStage(this, groups, (short) 1,
						EXEC_TOKENS, EXEC_HISTORY, (bundle, group) -> {
					if (!(bundle.getContent() instanceof REFITBatch)) {
						return bundle;
					}
					int groupID = -1;
					for (int i = 0; i < group.length; i++) {
						if (!group[i]) continue;
						groupID = REFITReplicaGroups.getExecutionGroupId((short) i);
						break;
					}
					final short gid = (short) groupID;
					return bundle.toFilteredBundle((m) -> ((REFITBatch) m).toGroupOnly(gid));
				}, REFITConfig.FAULTS_TO_TOLERATE);
			}
			addToInbound(REFITMessageType.CHANNEL_ACK, orderedChannelSenderStage);
			addToInbound(REFITMessageType.CHANNEL_SHARE_PART, orderedChannelSenderStage);
			addToInbound(REFITMessageType.CHANNEL_LOCAL_ACK, orderedChannelSenderStage);
			couplingStage.channelRequestSender.addListener(orderedChannelSenderStage);
			couplingStage.channelConfigSender.addListener(orderedChannelSenderStage);
			orderedChannelSenderStage.progressSender.addListener(couplingStage);

			clientStage = new REFITClientStage(this, -1);
			// only listen for requests from other replicas!
			addToInbound(REFITMessageType.REQUEST, clientStage);
			couplingStage.replySender.addListener(clientStage);
			clientStage.requestSender.addListener(orderStage);
			clientStage.panicSender.addListener(orderStage);
			orderStage.viewChangeSender.addListener(clientStage);
			clientStage.requestSender.addListener(couplingStage);

			multiRemoteClientStage = new REFITMCMultiRemoteClientStage(this);
			multiRemoteClientStage.requestSender.addListener(clientStage);
			multiRemoteClientStage.skipSender.addListener(couplingStage);
			couplingStage.channelProgressSender.addListener(multiRemoteClientStage);
			addToInbound(REFITMessageType.CHANNEL_SHARE, multiRemoteClientStage);
			addToInbound(REFITMessageType.CHANNEL_SHARE_BUNDLE, multiRemoteClientStage);
			addToInbound(REFITMessageType.CHANNEL_PROGRESS, multiRemoteClientStage);
			addToInbound(REFITMessageType.CHANNEL_SENDER_ACK, multiRemoteClientStage);

			requestChannelReceiverStage = new REFITAbstractChannelReceiver[REFITReplicaGroups.EXEC_COUNT];
			for (int i = 0; i < requestChannelReceiverStage.length; i++) {
				if (REFITConfig.IRMC_SENDER_COLLECTION) {
					requestChannelReceiverStage[i] = new REFITSourceChannelReceiverStage(this, REFITReplicaGroups.getExecutionReplicas(i),
							(short) i, REFITConfig.TOTAL_NR_OF_CLIENTS, CLIENT_TOKENS, REFITConfig.FAULTS_TO_TOLERATE);
				} else {
					requestChannelReceiverStage[i] = new REFITChannelReceiverStage(this, REFITReplicaGroups.getExecutionReplicas(i),
							(short) i, REFITConfig.TOTAL_NR_OF_CLIENTS, CLIENT_TOKENS, REFITConfig.FAULTS_TO_TOLERATE);
				}

				requestChannelReceiverStage[i].resultSender.addListener(multiRemoteClientStage);
				requestChannelReceiverStage[i].skipSender.addListener(multiRemoteClientStage);
				multiRemoteClientStage.progressSender.addListener(requestChannelReceiverStage[i]);
				multiRemoteClientStage.channelMsgSender[i].addListener(requestChannelReceiverStage[i]);
				couplingStage.channelConfigSender.addListener(requestChannelReceiverStage[i]);
			}

			execExecutionStage = null;
			execCouplingStage = null;
			execClientStage = null;
			mcClientStage = null;

			requestChannelSenderStage = null;
			orderedChannelReceiverStage = null;
		}
	}

	@Override
	protected void assignTasks(REFITScheduler[] schedulers) {
		// Assign tasks to schedulers
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.NETWORK_ENDPOINT);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.ORDER_STAGE);
		schedulers[2 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHECKPOINT_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXECUTION_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.UPDATE_STAGE);
		schedulers[4 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.FUZZY_CHECKPOINT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.PROBE_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.COUPLING_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXEC_COUPLING_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHANNEL_SENDER_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHANNEL_RECEIVER_STAGE);
	}
}
