package refit.replica;

import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;

import refit.config.REFITConfig;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITConnectionResetNotification;
import refit.message.REFITInstruction.REFITConnectionWriteableNotification;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.scheduler.REFITSchedulerGroup;
import refit.scheduler.REFITSchedulerTask;
import refit.scheduler.REFITSchedulerTaskType;


public abstract class REFITStage extends REFITSchedulerTask {

	protected final REFITBaseReplica replica;
	protected final Queue<REFITMessage> incomingMessages;
	protected final Set<REFITOutbox> activeOutboxes;
	protected final List<REFITOutbox> registeredOutboxes;
	protected boolean isInterleaved;


	public REFITStage(REFITSchedulerTaskType taskType, REFITBaseReplica replica) {
		this(taskType, replica, replica.group);
	}

	public REFITStage(REFITSchedulerTaskType taskType, REFITSchedulerGroup group) {
		this(taskType, null, group);
	}

	private REFITStage(REFITSchedulerTaskType taskType, REFITBaseReplica replica, REFITSchedulerGroup group) {
		super(taskType, group);
		this.replica = replica;
		this.incomingMessages = group.isSingleThreaded ? new LinkedList<>() : new ConcurrentLinkedQueue<>();
		this.activeOutboxes = new LinkedHashSet<>();
		this.registeredOutboxes = new ArrayList<>();
		this.isInterleaved = false;
	}

	protected void markInterleaved() {
		isInterleaved = true;
	}


	// ##################
	// # SCHEDULER TASK #
	// ##################

	@Override
	public void execute() {
		long startTime = (isInterleaved) ? System.nanoTime() : 0;
		// Process messages
		REFITMessage message;
		while ((message = incomingMessages.poll()) != null) {
			if (REFITConfig.ENABLE_MESSAGE_TRACES) {
				try {
					handleMessageInternal(message);
				} catch (Throwable t) {
					Exception e = new Exception("Fake exception to capture message source");
					e.setStackTrace(message.getSte());

					t.addSuppressed(e);
					throw t;
				}
			} else {
				handleMessageInternal(message);
			}
			event();
			if (isInterleaved) {
				long currentTime = System.nanoTime();
				// yield when execution took already more than 1ms
				if (currentTime - startTime > 1000 * 1000 && !incomingMessages.isEmpty()) {
					progress();
					break;
				}
			}
		}

		// Complete stage. Messages inserted into the queue by this method will be processed
		// on the next scheduler sub-iteration
		stageComplete();
		processOutboxes();
	}

	// #################
	// # STAGE METHODS #
	// #################

	public void insertMessage(REFITMessage message) {
		incomingMessages.add(message);
		progress();
	}

	private void handleMessageInternal(REFITMessage message) {
		if (message.type == REFITMessageType.INSTRUCTION) {
			REFITInstruction instruction = (REFITInstruction) message;
			switch (instruction.instructionType) {
			case CONNECTION_WRITEABLE_NOTIFICATION:
				REFITConnectionWriteableNotification writeable = (REFITConnectionWriteableNotification) instruction;
				writeable.outbox.notifyWriteable(writeable.id);
				enqueueOutbox(writeable.outbox);
				return;
			case CONNECTION_RESET_NOTIFICATION:
				REFITConnectionResetNotification reset = (REFITConnectionResetNotification) instruction;
				for (REFITOutbox outbox : registeredOutboxes) {
					outbox.requeue(reset.nodeID);
				}
				return;
			}
		}
		handleMessage(message);
	}

	protected abstract void handleMessage(REFITMessage message);

	protected void stageComplete() {
		// Override in sub classes if necessary
	}

	public void registerOutbox(REFITOutbox outbox) {
		registeredOutboxes.add(outbox);
	}

	public void enqueueOutbox(REFITOutbox outbox) {
		activeOutboxes.add(outbox);
	}

	protected void processOutboxes() {
		for (REFITOutbox outbox : activeOutboxes) {
			outbox.trySend();
		}
		activeOutboxes.clear();
	}

}
