package refit.replica.client;

import java.util.*;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITReplyBase;
import refit.message.REFITRequest;
import refit.message.REFITUniqueID;
import refit.util.REFITTime;


public class REFITReplyCache {

	private final Map<Short, REFITClientRequestReplyCacheEntry> cache;
	private final TreeSet<REFITClientRequestReplyCacheEntry> timeoutQueue;

	public REFITReplyCache() {
		this.cache = new HashMap<>(REFITConfig.TOTAL_NR_OF_CLIENTS);
		this.timeoutQueue = new TreeSet<>(Comparator.comparingLong(o -> o.receiveTime));
	}

	@Override
	public String toString() {
		return "RPYCH";
	}

	public boolean isOldUID(REFITUniqueID uid) {
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) return false;
		return (uid.seqNr < entry.uid.seqNr);
	}

	public boolean isCurrentUID(REFITUniqueID uid) {
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) return false;

		// Make sure that only new UIDs are stored
		return (uid.seqNr == entry.uid.seqNr);
	}

	public boolean registerRequest(REFITUniqueID uid, REFITRequest request) {
		// Get entry for client
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) {
			// Create new entry
			entry = new REFITClientRequestReplyCacheEntry();
			cache.put(uid.nodeID, entry);
		} else {
			// Make sure that only new UIDs are stored
			if (uid.seqNr <= entry.uid.seqNr) return false;
			if (entry.reply == null) {
				timeoutQueue.remove(entry);
			}
		}

		// Update entry
		entry.uid = uid;
		// Keep request for flushIncomplete
		// Only necessary with AUTONOMOUS_PANICS
		entry.request = request;
		entry.reply = null;
		entry.receiveTime = REFITTime.currentTimeMillis.getAsLong();
		timeoutQueue.add(entry);
		return true;
	}

	public void storeReply(REFITUniqueID uid, REFITReplyBase reply) {
		// Get entry for client
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) {
			// Create new entry
			entry = new REFITClientRequestReplyCacheEntry();
			cache.put(uid.nodeID, entry);
		} else {
			// Make sure that only new replies are stored
			if (uid.seqNr < entry.uid.seqNr) return;
		}
		timeoutQueue.remove(entry);

		// Store reply
		entry.uid = uid;
		// No need to keep the request for flushIncomplete once we have a reply
		entry.request = null;
		entry.reply = reply;
	}

	public REFITReplyBase getReply(REFITUniqueID uid) {
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) return null;
		if (uid.seqNr != entry.uid.seqNr) return null;
		return entry.reply;
	}

	public ArrayList<REFITRequest> flushIncomplete() {
		timeoutQueue.clear();
		ArrayList<REFITRequest> requests = new ArrayList<>();
		// automatically resubmit requests without a reply
		for (REFITClientRequestReplyCacheEntry entry : cache.values()) {
			if (entry.reply == null && entry.request != null) {
				requests.add(entry.request);
				entry.receiveTime = REFITTime.currentTimeMillis.getAsLong();
				timeoutQueue.add(entry);
			}
		}
		return requests;
	}

	public void clearIncomplete() {
		timeoutQueue.clear();
		// allow clients of requests without reply to submit them a second time
		for (REFITClientRequestReplyCacheEntry entry : cache.values()) {
			if (entry.reply == null && entry.request != null) {
				entry.request = null;
				entry.uid = new REFITUniqueID(entry.uid.nodeID, entry.uid.seqNr - 1);
			}
		}
	}

	public boolean shouldPanic(int timeout) {
		if (timeoutQueue.isEmpty()) return false;
		REFITClientRequestReplyCacheEntry firstEntry = timeoutQueue.iterator().next();
		long ct = REFITTime.currentTimeMillis.getAsLong();
		boolean shouldPanic = ct > firstEntry.receiveTime + timeout;
		if (shouldPanic) {
			REFITLogger.logWarning(this, "PANIC due to " + firstEntry.uid);
		}
		return shouldPanic;
	}

	public boolean isUIDTimeOut(REFITUniqueID uid, int timeout) {
		REFITClientRequestReplyCacheEntry entry = cache.get(uid.nodeID);
		if (entry == null) return false;
		if (uid.seqNr != entry.uid.seqNr) return false;
		long ct = REFITTime.currentTimeMillis.getAsLong();
		return ct > entry.receiveTime + timeout;
	}


	// ###############
	// # CACHE ENTRY #
	// ###############

	private static class REFITClientRequestReplyCacheEntry {

		public REFITUniqueID uid;
		public REFITRequest request;
		public REFITReplyBase reply;
		public long receiveTime;

		@Override
		public String toString() {
			return uid.toString() + ((request != null) ? "+req" : "") + ((reply != null) ? "+repl" : "");
		}
	}

}
