package refit.replica.client;

import java.util.*;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITReply;
import refit.message.REFITRequest;
import refit.message.REFITUniqueID;
import refit.util.REFITTime;


public class REFITStewardReplyCache {

	private final Map<Short, REFITClientRequestReplyCacheEntry> cache;
	private final TreeSet<REFITClientRequestReplyCacheEntry> timeoutQueue;

	public REFITStewardReplyCache() {
		this.cache = new HashMap<>(REFITConfig.TOTAL_NR_OF_CLIENTS);
		this.timeoutQueue = new TreeSet<>(Comparator.comparingLong(o -> o.receiveTime));
	}


	@Override
	public String toString() {
		return "RPYCH";
	}


	public boolean isOldUID(REFITUniqueID uid, boolean isTotalOrder) {
		short nid = (short) (2 * uid.nodeID + (isTotalOrder ? 0 : 1));
		REFITClientRequestReplyCacheEntry entry = cache.get(nid);
		if (entry == null) return false;
		return (uid.seqNr < entry.uid.seqNr);
	}

	public boolean isCurrentUID(REFITUniqueID uid, boolean isTotalOrder) {
		short nid = (short) (2 * uid.nodeID + (isTotalOrder ? 0 : 1));
		REFITClientRequestReplyCacheEntry entry = cache.get(nid);
		if (entry == null) return false;

		// Make sure that only new UIDs are stored
		return (uid.seqNr == entry.uid.seqNr);
	}

	public boolean registerRequest(REFITRequest request) {
		short nid = (short) (2 * request.uid.nodeID + (request.mode.isTotalOrder() ? 0 : 1));
		// Get entry for client
		REFITClientRequestReplyCacheEntry entry = cache.get(nid);
		if (entry == null) {
			// Create new entry
			entry = new REFITClientRequestReplyCacheEntry();
			cache.put(nid, entry);
		} else {
			// Make sure that only new UIDs are stored
			if (request.uid.seqNr <= entry.uid.seqNr) return false;
			if (entry.reply == null) {
				timeoutQueue.remove(entry);
			}
		}

		// Update entry
		entry.uid = request.uid;
		// Keep request for flushIncomplete
		entry.request = request;
		entry.reply = null;
		entry.receiveTime = REFITTime.currentTimeMillis.getAsLong();
		timeoutQueue.add(entry);
		return true;
	}

	public void storeReply(REFITReply reply) {
		short nid = (short) (2 * reply.uid.nodeID + (reply.mode.isTotalOrder() ? 0 : 1));
		// Get entry for client
		REFITClientRequestReplyCacheEntry entry = cache.get(nid);
		if (entry == null) {
			// Create new entry
			entry = new REFITClientRequestReplyCacheEntry();
			cache.put(nid, entry);
		} else {
			// Make sure that only new replies are stored
			if (reply.uid.seqNr < entry.uid.seqNr) return;
		}
		timeoutQueue.remove(entry);

		// Store reply
		entry.uid = reply.uid;
		// No need to keep the request for flushIncomplete once we have a reply
		entry.request = null;
		entry.reply = reply;
	}

	public REFITReply getReply(REFITUniqueID uid, boolean isTotalOrder) {
		short nid = (short) (2 * uid.nodeID + (isTotalOrder ? 0 : 1));
		REFITClientRequestReplyCacheEntry entry = cache.get(nid);
		if (entry == null) return null;
		if (uid.seqNr != entry.uid.seqNr) return null;
		return entry.reply;
	}

	public ArrayList<REFITRequest> flushIncomplete() {
		timeoutQueue.clear();
		ArrayList<REFITRequest> requests = new ArrayList<>();
		// allow clients of requests without reply to submit them a second time
		for (REFITClientRequestReplyCacheEntry entry : cache.values()) {
			if (entry.reply == null && entry.request != null) {
				requests.add(entry.request);
				entry.receiveTime = REFITTime.currentTimeMillis.getAsLong();
				timeoutQueue.add(entry);
			}
		}
		return requests;
	}

	public boolean shouldPanic() {
		if (timeoutQueue.isEmpty()) return false;
		REFITClientRequestReplyCacheEntry firstEntry = timeoutQueue.iterator().next();
		long ct = REFITTime.currentTimeMillis.getAsLong();
		boolean shouldPanic = ct > firstEntry.receiveTime + REFITConfig.REPLICA_REQUEST_TIMEOUT;
		if (shouldPanic) {
			REFITLogger.logWarning(this, "PANIC due to " + firstEntry.uid);
		}
		return shouldPanic;
	}


	// ###############
	// # CACHE ENTRY #
	// ###############

	private static class REFITClientRequestReplyCacheEntry {

		public REFITUniqueID uid;
		public REFITRequest request;
		public REFITReply reply;
		public long receiveTime;

		@Override
		public String toString() {
			return uid.toString() + ((request != null) ? "+req" : "") + ((reply != null) ? "+repl" : "");
		}

	}

}
