package refit.replica.coupling;

import java.util.Arrays;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITExecutorFlushMessage;
import refit.message.REFITInstruction.REFITAwaitRequestNotification;
import refit.message.REFITInstruction.REFITFlushRequestNotification;
import refit.message.REFITMessage;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.replica.REFITStage;
import refit.replica.order.REFITOrderGroups;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITCouplingHelper extends REFITStage {
	public final REFITObserver<REFITFlushRequestNotification> flushRequestSender;
	public final REFITObserver<REFITAwaitRequestNotification> awaitRequestSender;

	private final long[] flushStatus;
	private final long[] execStatus;


	private long highestOrderedSeqNr;
	private long nextExecSeqNr;
	private long[] nextExecSeqNrs;

	public REFITCouplingHelper(REFITBaseReplica replica) {
		super(REFITSchedulerTaskType.COUPLING_STAGE, replica);
		flushRequestSender = new REFITObserver<>();
		awaitRequestSender = new REFITObserver<>();

		nextExecSeqNrs = new long[REFITOrderGroups.COUNT];
		Arrays.fill(nextExecSeqNrs, 0);
		nextExecSeqNr = 0;

		this.flushStatus = new long[REFITConfig.EXEC_GROUP_SIZE];
		Arrays.fill(flushStatus, -1L);
		this.execStatus = new long[REFITConfig.EXEC_GROUP_SIZE];
		Arrays.fill(execStatus, -1L);

	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.groupType()) {
		case EXECUTOR_FLUSH:
			handleExecFlush((REFITExecutorFlushMessage) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.groupType());
		}
	}

	private void handleExecFlush(REFITExecutorFlushMessage flushMessage) {
		if (flushMessage == null) return;
		if (flushMessage.from >= REFITConfig.EXEC_GROUP_SIZE) return;
		// ignore outdated flush, prevents replaying old flushes
		if (flushMessage.nextOrderedID < flushStatus[flushMessage.from]) return;
		if (flushMessage.nextExecutedID < execStatus[flushMessage.from]) return;
		if (!replica.mac.verifyMulticastMAC(flushMessage, REFITReplicaGroups.getExecutionReplicas(0))) return;

		flushStatus[flushMessage.from] = flushMessage.nextOrderedID;
		execStatus[flushMessage.from] = flushMessage.nextExecutedID;
		// use median of the received values, equals f+1 highest
		long[] sortedFlush = new long[flushStatus.length];
		System.arraycopy(flushStatus, 0, sortedFlush, 0, flushStatus.length);
		Arrays.sort(sortedFlush);
		long nextOrderSeqNr = sortedFlush[sortedFlush.length / 2];

		// use median of the received values
		long[] sortedExec = new long[execStatus.length];
		System.arraycopy(execStatus, 0, sortedExec, 0, execStatus.length);
		Arrays.sort(sortedExec);
		long nextExecSeqNr = sortedExec[sortedExec.length / 2];

		// limits are the result of majority vote thus they can be trusted
		long flushSlotsBeforeSeqNr = nextOrderSeqNr;
		long partialBatchingBeforeSeqNr = nextExecSeqNr + REFITConfig.NO_BATCHING_DELAY_WINDOW;
		flushRequestSender.broadcast(new REFITFlushRequestNotification(flushSlotsBeforeSeqNr, partialBatchingBeforeSeqNr));

		awaitRequestSender.broadcast(new REFITAwaitRequestNotification(flushSlotsBeforeSeqNr - 1));
	}


	public void trackProgress(long highestOrderedSeqNr, long nextExecSeqNr, int groupID) {
		boolean changed = false;
		if (this.highestOrderedSeqNr < highestOrderedSeqNr) {
			this.highestOrderedSeqNr = highestOrderedSeqNr;
			changed = true;
		}

		if (nextExecSeqNrs[groupID] < nextExecSeqNr) {
			nextExecSeqNrs[groupID] = nextExecSeqNr;
			long min = nextExecSeqNrs[0];
			for (int i = 1; i < nextExecSeqNrs.length; i++) {
				min = Math.min(min, nextExecSeqNrs[i]);
			}
			if (min > this.nextExecSeqNr) {
				this.nextExecSeqNr = min;
				changed = true;
			}
		}

		if (changed && !hasTimeout()) {
			setTimeout(REFITConfig.FLUSH_TIMEOUT);
		}
	}

	@Override
	public void timeout() {
		sendFlushMessage();
	}

	private void sendFlushMessage() {
		REFITMessage flush = new REFITExecutorFlushMessage(replica.id, highestOrderedSeqNr + 1, nextExecSeqNr);
		flush.serializeMessage();
		replica.mac.appendMulticastMAC(flush);
		replica.replicaMulticast(flush);
		// send message to myself
		insertMessage(flush);
	}
}
