package refit.replica.coupling;

import java.nio.ByteBuffer;
import java.util.Arrays;

import refit.config.REFITConfig;
import refit.message.REFITInstruction.REFITExecuteRequest;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITUniqueID;
import refit.util.REFITPayload;

public class REFITOrderedRequest extends REFITMessage {

	public static final int ORDERED_REQUEST_HEADER_SIZE = (REFITConfig.USE_SPINNING_LEADER ? 0 : Integer.BYTES) + Byte.BYTES;
	private static final short ORDERED_REQUEST_MAGIC = -32;

	public final int viewID;
	public final byte[] messageHash;
	public REFITMessage message;


	public REFITOrderedRequest(short from, REFITExecuteRequest operation, boolean hashOnly) {
		super(REFITMessageType.ORDERED_REQUEST, new REFITUniqueID(ORDERED_REQUEST_MAGIC, operation.agreementSeqNr), from);
		this.viewID = operation.viewID;
		if (hashOnly && operation.message.type != REFITMessageType.REQUEST_NO_OP) {
			this.message = null;
		} else {
			this.message = operation.message;
		}
		this.messageHash = operation.message.getHash();
	}

	public REFITOrderedRequest(ByteBuffer buffer) {
		super(REFITMessageType.ORDERED_REQUEST, buffer);
		if (!REFITConfig.USE_SPINNING_LEADER) {
			this.viewID = buffer.getInt();
		} else {
			this.viewID = 0;
		}
		// get message
		byte flags = buffer.get();
		boolean isNop = (flags & 1) != 0;
		if (isNop) {
			this.message = REFITNoOpRequest.INSTANCE;
			this.messageHash = message.getHash();
		} else {
			boolean hashOnly = (flags & 2) != 0;
			if (hashOnly) {
				this.message = null;
				this.messageHash = new byte[REFITPayload.HASH_SIZE];
				buffer.get(this.messageHash);
			} else {
				this.message = REFITMessage.createMessage(buffer.slice());
				buffer.position(buffer.position() + this.message.getMessageSize());
				this.messageHash = message.getHash();
			}
		}
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		if (!REFITConfig.USE_SPINNING_LEADER) {
			buffer.putInt(viewID);
		}
		boolean isNop = (message != null && message.type == REFITMessageType.REQUEST_NO_OP);
		buffer.put((byte) ((isNop ? 1 : 0) + (!isNop && message == null ? 2 : 0)));
		if (!isNop) {
			if (message != null) {
				buffer.put(message.getBuffer());
			} else {
				buffer.put(messageHash);
			}
		}
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.MAC_AUTHENTICATOR;
	}

	public boolean setFullMessage(REFITOrderedRequest request) {
		if (request == null || request.message == null) return false;
		if (this.message != null) return false;
		if (!uid.equals(request.uid)) return false;
		if (!Arrays.equals(messageHash, request.messageHash)) return false;
		this.message = request.message;
		return true;
	}

	@Override
	protected int calculatePayloadSize() {
		int size = super.calculatePayloadSize() + ORDERED_REQUEST_HEADER_SIZE;
		if (message == null || message.type != REFITMessageType.REQUEST_NO_OP) {
			if (message != null) {
				size += message.getMessageSize();
			} else {
				size += messageHash.length;
			}
		}
		return size;
	}

	public boolean isExecutable() {
		return message != null;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%d|%s|%s}", type, from, uid.seqNr, (message != null) ? message.uid.toString() : "null", REFITPayload.toString(messageHash));
	}
}
