package refit.replica.execution;

import java.nio.ByteBuffer;
import java.util.*;

import refit.application.REFITApplicationServer;
import refit.application.REFITLeaderRecommendation;
import refit.application.REFITResult;
import refit.application.REFITSystemApplicationServer;
import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITSystemNotification;
import refit.replica.REFITObserver;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification.REFITFuzzyCheckpointFinishNotification;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification.REFITFuzzyCheckpointStartNotification;
import refit.replica.order.REFITOrderGroups;
import refit.replica.update.REFITBatchUpdate;
import refit.replica.update.REFITUpdate;
import refit.replica.update.REFITUpdateMessage;
import refit.util.REFITPayload;
import refit.util.REFITTime;


public class REFITExecutor {

	private final short replicaID;
	private final REFITApplicationServer application;

	private final REFITSystemApplicationServer systemApplication;
	private int targetMinView = REFITConfig.INITIAL_VIEW;
	private long targetAfterSeq = 0;
	private long targetProbeCounter = 0;

	private final REFITObserver<REFITReply>[] replySenders;
	private final REFITObserver<REFITCheckpointState> checkpointSender;
	private final REFITObserver<REFITUpdateMessage> updateSender;
	private final REFITObserver<REFITFuzzyCheckpointNotification> fuzzyCheckpointNotificationSender;
	private final REFITObserver<REFITSystemNotification> systemSender;

	public REFITExecutor(short replicaID, REFITApplicationServer application, REFITSystemApplicationServer systemApplication,
	                     REFITObserver<REFITReply>[] replySenders, REFITObserver<REFITCheckpointState> checkpointSender,
	                     REFITObserver<REFITUpdateMessage> updateSender,
	                     REFITObserver<REFITFuzzyCheckpointNotification> fuzzyCheckpointNotificationSender,
	                     REFITObserver<REFITSystemNotification> systemSender) {
		this.replicaID = replicaID;
		this.application = application;
		this.systemApplication = systemApplication;
		this.enableCheckpoints = false;
		this.createLightWeightCheckpoints = false;
		this.enableUpdates = false;
		this.updateSendQueue = new LinkedList<>();
		this.nodeProgresses = new long[REFITConfig.ADDRESSES.length];
		Arrays.fill(nodeProgresses, -1L);
		this.replyHashes = new byte[REFITConfig.ADDRESSES.length][];
		Arrays.fill(replyHashes, new byte[REFITPayload.HASH_SIZE]);
		this.nodeSeqs = new long[REFITConfig.ADDRESSES.length];
		Arrays.fill(nodeSeqs, -1L);

		this.replySenders = replySenders;
		this.checkpointSender = checkpointSender;
		this.updateSender = updateSender;
		this.fuzzyCheckpointNotificationSender = fuzzyCheckpointNotificationSender;
		this.systemSender = systemSender;
	}


	@Override
	public String toString() {
		return "EXCTR";
	}


	// ##########################
	// # EXACTLY-ONCE EXECUTION #
	// ##########################

	private long[] nodeProgresses;
	private long[] nodeSeqs;


	private boolean ensureExactlyOnce(REFITUniqueID uid) {
		if (uid.seqNr <= nodeProgresses[uid.nodeID]) return false;
		nodeProgresses[uid.nodeID] = uid.seqNr;
		return true;
	}


	// ###############
	// # CHECKPOINTS #
	// ###############

	private byte[][] replyHashes;
	private boolean enableCheckpoints;
	private boolean createLightWeightCheckpoints;

	private boolean fuzzyCheckpointActive;
	private boolean fuzzyCheckpointCriticalSection;
	private long fuzzyCheckpointWaitStart;
	private long fuzzyCheckpointStartSeqNr;
	private static long maxOffset = REFITConfig.CHECKPOINT_INTERVAL / 2;
	private long fuzzyCheckpointOffset = maxOffset;
	private long nextFuzzyCheckpointOffset = maxOffset;

	public void setEnableCheckpoints(boolean enableCheckpoints, boolean createLightWeightCheckpoints, long nextAgreementSeqNr) {
		if (fuzzyCheckpointCriticalSection) REFITLogger.logError(this, "Configuration not allowed in critical section");
		REFITLogger.logEvent(this, "change checkpoint-creation setting to \"" + (enableCheckpoints ? (createLightWeightCheckpoints ? "lightweight" : "regular") : "disabled") + "\"");

		boolean previousLightWeightCheckpoints = this.createLightWeightCheckpoints;
		this.enableCheckpoints = enableCheckpoints;
		this.createLightWeightCheckpoints = createLightWeightCheckpoints;

		// regenerate checkpoint if we only have a lightweight one at the moment if possible
		// this prevents getting stuck if the ordering window is full, but the lightweight checkpointing failed
		if (previousLightWeightCheckpoints && !createLightWeightCheckpoints) {
			if (enableCheckpoints && ((nextAgreementSeqNr % REFITConfig.CHECKPOINT_INTERVAL) == 0))
				createCheckpoint(nextAgreementSeqNr);
		}
	}

	private void createCheckpointIfRequired(long agreementSeqNr) {
		if (!enableCheckpoints) return;

		// Check whether to create a checkpoint
		if (REFITConfig.ENABLE_FUZZY_CHECKPOINTS
				&& (((agreementSeqNr + fuzzyCheckpointOffset) % REFITConfig.CHECKPOINT_INTERVAL) == 0 || agreementSeqNr == 0)) {
			startFuzzyCheckpoint(agreementSeqNr);
		}
		if ((agreementSeqNr % REFITConfig.CHECKPOINT_INTERVAL) == 0) {
			createCheckpoint(agreementSeqNr);
		}
	}

	private void startFuzzyCheckpoint(long agreementSeqNr) {
		REFITLogger.logDebug(this, "Requesting fuzzy checkpoint start @" + agreementSeqNr);
		Set<?> changelist = null;
		if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			// send out old updates, if any
			sendUpdate();
		} else {
			if (REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING) {
				REFITLogger.logDebug(this, "getting changelist for fuzzy collection");
				changelist = application.takeChangelist();
			} else {
				REFITLogger.logDebug(this, "clearing changes");
				// reset changelist
				application.takeChangelist();
			}
		}
		fuzzyCheckpointStartSeqNr = agreementSeqNr;
		fuzzyCheckpointActive = true;
		fuzzyCheckpointCriticalSection = true;
		fuzzyCheckpointNotificationSender
				.broadcast(new REFITFuzzyCheckpointStartNotification(agreementSeqNr, changelist));
	}

	private void createCheckpoint(long agreementSeqNr) {
		int lastLeader = targetMinView % REFITConfig.TOTAL_NR_OF_REPLICAS;
		REFITLeaderRecommendation leaderRecommendation = systemApplication.queryLeader(lastLeader, targetProbeCounter);
		if (REFITConfig.APPLY_PROBING_RESULTS && leaderRecommendation != null && leaderRecommendation.leader != lastLeader) {
			int delta = leaderRecommendation.leader - lastLeader;
			if (delta < 0) delta += REFITConfig.TOTAL_NR_OF_REPLICAS;
			targetMinView += delta;
			targetAfterSeq = agreementSeqNr + REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
			targetProbeCounter = leaderRecommendation.probeCounter;
			REFITLogger.logDebug(this, "Planning switch to >" + targetMinView + "@" + targetAfterSeq);
		}

		if (REFITConfig.ENABLE_FUZZY_CHECKPOINTS) {
			if (!fuzzyCheckpointActive)
				throw new InternalError("Unexpected fuzzy checkpoint end @" + agreementSeqNr);
			REFITLogger.logDebug(this, "Notifying about finished interval up to " + agreementSeqNr);
			fuzzyCheckpointActive = false;
			REFITExecutorState state = collectMetadata();
			Map<Object, REFITCheckpointObject> changes = null;
			if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
				sendUpdate();
			} else {
				changes = collectChanges();
			}
			REFITCheckpointObject[] sysAppData = systemApplication.createCheckpoint();
			if (sysAppData.length != 1) {
				throw new AssertionError("Only one checkpoint object allowed for the system application");
			}
			fuzzyCheckpointNotificationSender
					.broadcast(new REFITFuzzyCheckpointFinishNotification(agreementSeqNr, state, sysAppData[0], changes));

			if (mustBlockExecution()) {
				fuzzyCheckpointWaitStart = REFITTime.currentTimeMillis.getAsLong();
				REFITLogger.logDebug(this, "Waiting for FCP @" + agreementSeqNr);
			} else {
				fuzzyCheckpointWaitStart = -1L;
				// the update for the blocked case is handled in setFuzzyCheckpointCompleted
				fuzzyCheckpointOffset = nextFuzzyCheckpointOffset;
			}
		} else if (REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING) {
			REFITExecutorState state = collectMetadata();
			Map<Object, REFITCheckpointObject> changes = collectChanges();
			REFITCheckpointObject[] sysAppData = systemApplication.createCheckpoint();
			if (sysAppData.length != 1) {
				throw new AssertionError("Only one checkpoint object allowed for the system application");
			}

			fuzzyCheckpointNotificationSender
					.broadcast(new REFITFuzzyCheckpointStartNotification(agreementSeqNr, null));
			fuzzyCheckpointNotificationSender
					.broadcast(new REFITFuzzyCheckpointFinishNotification(agreementSeqNr, state, sysAppData[0], changes));
			fuzzyCheckpointStartSeqNr = agreementSeqNr;

			if (agreementSeqNr == 0) {
				// must block the first time to give the fuzzy checkpoint stage the chance to collect a checkpoint
				fuzzyCheckpointCriticalSection = true;
				fuzzyCheckpointWaitStart = REFITTime.currentTimeMillis.getAsLong();
			}
		} else {
			createRegularCheckpoint(agreementSeqNr);
		}
	}

	private Map<Object, REFITCheckpointObject> collectChanges() {
		REFITLogger.logDebug(this, "getting changes");
		long start = REFITTime.currentTimeMillis.getAsLong();

		Map<Object, REFITCheckpointObject> changedObjects = application.createDiffCheckpoint(application.takeChangelist());

		long finishCollect = REFITTime.currentTimeMillis.getAsLong();
		REFITLogger.logDebug(this, (finishCollect - start) + "ms until " + changedObjects.size() + " objects of checkpoint data retrieved from application");

		return changedObjects;
	}

	private void createRegularCheckpoint(long agreementSeqNr) {
		long start = REFITTime.currentTimeMillis.getAsLong();
		if (REFITLogger.LOG_EXECUTION)
			REFITLogger.logExecution(this, "create checkpoint for agreement sequence number " + agreementSeqNr);

		// keep normal checkpoint implementation as baseline
		// an object based checkpoint should behave similar. However, for a proper comparison it seems to be better to use the default implementation
		REFITExecutorState checkpointMetadata = collectMetadata();

		REFITCheckpointObject[] data;
		if (createLightWeightCheckpoints) {
			data = new REFITCheckpointObject[]{checkpointMetadata};
		} else {
			REFITCheckpointObject[] checkpointData = application.createCheckpoint();
			REFITCheckpointObject[] sysAppData = systemApplication.createCheckpoint();
			if (sysAppData.length != 1) {
				throw new AssertionError("Only one checkpoint object allowed for the system application");
			}

			// merge data arrays
			data = new REFITCheckpointObject[checkpointData.length + 2];
			data[0] = checkpointMetadata;
			data[1] = sysAppData[0];
			System.arraycopy(checkpointData, 0, data, 2, checkpointData.length);
		}
		REFITCheckpointState checkpoint = new REFITCheckpointState(agreementSeqNr, replicaID, createLightWeightCheckpoints, data);
		checkpointSender.broadcast(checkpoint);

		long finishCollect = REFITTime.currentTimeMillis.getAsLong();
		if (data.length > 3) {
			// don't print for classical checkpoints
			REFITLogger.logDebug(this, (finishCollect - start) + "ms until " + data.length + " checkpoint objects for regular checkpoint retrieved from application");
		}
	}

	private REFITExecutorState collectMetadata() {
		REFITExecutorState checkpointMetadata = new REFITExecutorState(nodeProgresses, replyHashes, nodeSeqs,
				targetMinView, targetAfterSeq, targetProbeCounter);
		checkpointMetadata.serializeMessage();
		return checkpointMetadata;
	}

	public boolean mustBlockExecution() {
		return !fuzzyCheckpointActive && fuzzyCheckpointCriticalSection;
	}

	public void setFuzzyCheckpointCompleted(long agreementSeqNr) {
		long checkpointDuration = agreementSeqNr - fuzzyCheckpointStartSeqNr;
		REFITLogger.logDebug(this, "FCP took " + checkpointDuration + " seqs");
		if (agreementSeqNr > 1) {
			if (REFITConfig.DYNAMIC_FUZZY_CHECKPOINT_OFFSET) {
				nextFuzzyCheckpointOffset = (long) (checkpointDuration * REFITConfig.FUZZY_CHECKPOINT_BUFFER_SCALE) + REFITConfig.FUZZY_CHECKPOINT_BUFFER;
			} else {
				nextFuzzyCheckpointOffset = REFITConfig.FUZZY_CHECKPOINT_OFFSET;
			}
		}
		nextFuzzyCheckpointOffset = Math.min(nextFuzzyCheckpointOffset, maxOffset);
//		if (fuzzyCheckpointOffset != maxOffset) {
//			// limit adaption rate to reject noise
//			nextFuzzyCheckpointOffset = Math.max(fuzzyCheckpointOffset - REFITConfig.FUZZY_CHECKPOINT_BUFFER, nextFuzzyCheckpointOffset);
//		}

		if (mustBlockExecution()) {
			long time = REFITTime.currentTimeMillis.getAsLong() - fuzzyCheckpointWaitStart;
			REFITLogger.logDebug(this, "FCP delayed by " + time + "ms");
			fuzzyCheckpointWaitStart = -1L;
			// update offset here if execution was blocked
			fuzzyCheckpointOffset = nextFuzzyCheckpointOffset;
		}
		REFITLogger.logDebug(this, "Next offset: " + nextFuzzyCheckpointOffset);
		fuzzyCheckpointCriticalSection = false;
	}

	public void applyCheckpoint(REFITCheckpointState checkpoint) {
		if (fuzzyCheckpointCriticalSection)
			REFITLogger.logError(this, "Checkpoint application not allowed while in critical section");
		if (REFITLogger.LOG_EXECUTION) REFITLogger.logExecution(this, "apply checkpoint " + checkpoint);

		REFITCheckpointObject[] applicationData = new REFITCheckpointObject[checkpoint.data.length - 2];
		System.arraycopy(checkpoint.data, 2, applicationData, 0, applicationData.length);
		application.applyCheckpoint(applicationData);

		REFITExecutorState metadata = new REFITExecutorState(checkpoint.data[0].getBuffer());
		for (short i = 0; i < nodeProgresses.length; i++) {
			if (metadata.nodeProgresses[i] > nodeProgresses[i]) {
				// the checkpoint only contains reply hashes -> provide those to the client stage to block panics
				REFITReply reply = new REFITReply(new REFITUniqueID(i, metadata.nodeProgresses[i]), replicaID, (short) -1,
						(short) -1, metadata.nodeSeqs[i], false, ByteBuffer.wrap(metadata.replyHashes[i]), RequestMode.WRITE, (short) -1);
				for (REFITObserver<REFITReply> replySender : replySenders) {
					replySender.broadcast(reply);
				}
			}
		}
		nodeProgresses = metadata.nodeProgresses;
		nodeSeqs = metadata.nodeSeqs;

		REFITCheckpointObject[] sysAppData = {checkpoint.data[1]};
		systemApplication.applyCheckpoint(sysAppData);

		targetMinView = metadata.targetMinView;
		targetAfterSeq = metadata.targetAfterSeq;
		targetProbeCounter = metadata.targetProbeCounter;
		systemSender.broadcast(new REFITSystemNotification(targetMinView, targetAfterSeq));
	}

	// ###########
	// # UPDATES #
	// ###########

	private boolean enableUpdates;
	private final Queue<REFITUpdate> updateSendQueue;


	public void setEnableUpdates(boolean enableUpdates) {
		// Flush batched updates
		if (this.enableUpdates && !enableUpdates) sendUpdate();

		// Change setting
		REFITLogger.logEvent(this, "change update-creation setting to \"" + (enableUpdates ? "enabled" : "disabled") + "\"");
		this.enableUpdates = enableUpdates;
	}

	public void applyUpdate(REFITUpdateMessage update) {
		// Apply update (batch)
		long agreementSeqNr;
		if (update.type == REFITMessageType.BATCH_UPDATE) {
			REFITBatchUpdate batchUpdate = (REFITBatchUpdate) update;
			for (REFITUpdate upd : batchUpdate.updates) applySingleUpdate(upd);
			agreementSeqNr = batchUpdate.getMaxAgreementSeqNr();
		} else {
			applySingleUpdate((REFITUpdate) update);
			agreementSeqNr = update.agreementSeqNr();
		}

		// Check whether to create a checkpoint
		if (enableCheckpoints && (((agreementSeqNr + 1) % REFITConfig.CHECKPOINT_INTERVAL) == 0))
			createCheckpoint(agreementSeqNr + 1);
	}

	private void applySingleUpdate(REFITUpdate update) {
		if (!ensureExactlyOnce(update.uid)) return;
		if (REFITLogger.LOG_EXECUTION) REFITLogger.logExecution(this, "apply update " + update);
		ByteBuffer payload = update.getPayload();

		if (payload.remaining() > 0) {
			REFITApplicationServer server = (update.isSystem) ? systemApplication : application;
			server.applyUpdate(payload);
		}

		// Add reply to reply cache in client stage
		// FIXME correct view id
		int viewID = 0;
		int groupID = (REFITConfig.SEPARATE_EXECUTION_REPLICAS) ? -1 : REFITOrderGroups.getGroupID(update.agreementSeqNr());
		REFITReply reply = new REFITReply(update.uid, replicaID, (short) groupID, viewID, update.agreementSeqNr(), false,
				ByteBuffer.wrap(update.replyHash), RequestMode.WRITE, (short) -1);
		for (REFITObserver<REFITReply> replySender : replySenders) {
			replySender.broadcast(reply);
		}
		replyHashes[update.uid.nodeID] = update.replyHash;
		nodeSeqs[update.uid.nodeID] = update.agreementSeqNr();
	}

	private void createUpdate(REFITUniqueID uid, long agreementSeqNr, ByteBuffer payload, byte[] replyHash, boolean isSystem) {
		if (!REFITConfig.UPDATE_BASED_CHECKPOINTING) throw new AssertionError("Must not use updates!");
		if (payload == null) {
			payload = ByteBuffer.wrap(new byte[0]);
		}
		REFITUpdate update = new REFITUpdate(uid, replicaID, agreementSeqNr, payload, replyHash, isSystem);
		updateSendQueue.add(update);
	}

	private void sendUpdate() {
		if (updateSendQueue.isEmpty()) return;
		REFITUpdateMessage update;
		if (updateSendQueue.size() > 1) {
			// Assign batch update the UID and agreement sequence number of the first individual update
			REFITUpdate firstUpdate = updateSendQueue.peek();
			update = new REFITBatchUpdate(firstUpdate.agreementSeqNr(), replicaID, updateSendQueue);
			updateSendQueue.clear();
		} else {
			update = updateSendQueue.poll();
		}
		updateSender.broadcast(update);
	}


	// ############
	// # REQUESTS #
	// ############

	public void processRequest(REFITOrderRequest request, long agreementSeqNr, int viewID) {
		if (mustBlockExecution())
			throw new IllegalStateException("Request execution not allowed in the current state");
		// Debug output
		if (REFITLogger.LOG_EXECUTION)
			REFITLogger.logExecution(this, "process request " + request + " (" + agreementSeqNr + ")");
		if (agreementSeqNr == 0)
			REFITLogger.logEvent(this, String.format("Start time: %.3fs", REFITTime.timeSinceStart() * 0.001f));
		if ((agreementSeqNr % 1000) == 0)
			REFITLogger.logEvent(this, String.format("%10d @ %13d", agreementSeqNr, REFITTime.timeMicroUTC.getAsLong()));

		// Execute request
		switch (request.type) {
		case REQUEST:
			processSingleRequest((REFITRequest) request, agreementSeqNr, viewID);
			break;
		case BATCH:
			REFITBatch batch = (REFITBatch) request;
			for (int i = 0; i < batch.requests.length; i++) {
				REFITRequest req = batch.requests[i];
				if (req == null) {
					// If we didn't get an ordered read request, then just add a placeholder for it
					// Ideally we would tell the client that we can't provide the reply for its request
					REFITUniqueID uid = batch.requestTimestamps[i];
					ensureExactlyOnce(uid);

					// Create placeholder reply
					REFITReply reply = new REFITReply(uid, replicaID, batch.requestGroups[i], RequestMode.WRITE);
					replyHashes[uid.nodeID] = new byte[REFITPayload.HASH_SIZE];

					// Forward reply
					replySenders[(REFITConfig.SEPARATE_EXECUTION_REPLICAS) ? 0 : reply.groupID].broadcast(reply);
				} else {
					processSingleRequest(req, agreementSeqNr, viewID);
				}
			}
			break;
		case REQUEST_NO_OP:
		default:
			break;
		}

		// Create update and/or checkpoint
		if (enableUpdates && (((agreementSeqNr + 1) % REFITConfig.UPDATE_BATCH_INTERVAL) == 0)
				|| fuzzyCheckpointActive) {
			sendUpdate();
		}
		createCheckpointIfRequired(agreementSeqNr + 1);
	}

	private void processSingleRequest(REFITRequest request, long agreementSeqNr, int viewID) {
		// Check whether to process the request
		if (!ensureExactlyOnce(request.uid)) return;
		nodeSeqs[request.uid.nodeID] = agreementSeqNr;
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) request.getContext().statistics.track("PEX");

		// Process request
		REFITApplicationServer server = (request.mode.equals(RequestMode.SYSTEM)) ? systemApplication : application;
		REFITResult result = server.processRequest(request.uid, agreementSeqNr, request.getPayload(),
				enableUpdates || fuzzyCheckpointActive);

		// already acked internally
		if (request.mode.equals(RequestMode.WRITE_ACK)) return;

		// Create reply
		short groupID = (REFITConfig.SEPARATE_EXECUTION_REPLICAS) ? request.groupID : REFITOrderGroups.getGroupID(agreementSeqNr);
		REFITReply reply = new REFITReply(request.uid, replicaID, groupID, viewID, agreementSeqNr,
				true, result.reply, request.mode, request.replyReplicaID);
		if (REFITConfig.USE_HASHED_REPLIES) {
			if ((request.replyReplicaID >= 0) && (request.replyReplicaID != replicaID) && (reply.getPayload().remaining() > REFITPayload.HASH_SIZE))
				reply.sendHashReply();
		}
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) reply.getContext().statistics = request.getContext().statistics;

		// Create update
		if ((REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || fuzzyCheckpointActive) && REFITConfig.UPDATE_BASED_CHECKPOINTING
				|| enableUpdates)
			createUpdate(request.uid, agreementSeqNr, result.update, reply.getPayloadHash(), reply.mode.equals(RequestMode.SYSTEM));

		replyHashes[request.uid.nodeID] = reply.getPayloadHash();

		// Forward reply
		replySenders[(REFITConfig.SEPARATE_EXECUTION_REPLICAS) ? 0 : groupID].broadcast(reply);
	}

	public void processReadOnlyRequest(REFITRequest request, long agreementSeqNr) {
		// Debug output
		if (REFITLogger.LOG_EXECUTION) REFITLogger.logExecution(this, "process readonly request " + request);
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) request.getContext().statistics.track("REX");

		// Process request
		REFITReply reply;
		if (application.checkReadOnly(request.getPayload())) {
			REFITResult result = application.processRequest(request.uid, agreementSeqNr, request.getPayload(), false);

			// Create reply
			reply = new REFITReply(request.uid, replicaID, (short) -1, -1, agreementSeqNr, true, result.reply, request.mode, request.replyReplicaID);
			if (REFITConfig.USE_HASHED_REPLIES) {
				if ((request.replyReplicaID >= 0) && (request.replyReplicaID != replicaID) && (reply.getPayload().remaining() > REFITPayload.HASH_SIZE))
					reply.sendHashReply();
			}
		} else {
			// malicious request -> no reply, but block panics
			reply = new REFITReply(request.uid, replicaID, (short) -1, request.mode);
		}

		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) reply.getContext().statistics = request.getContext().statistics;
		// Forward reply
		replySenders[(REFITConfig.SEPARATE_EXECUTION_REPLICAS) ? 0 : request.groupID].broadcast(reply);
	}
}
