package refit.replica.mcclient;

import java.nio.ByteBuffer;

import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITChannelProgress;
import refit.message.REFITInstruction.REFITChannelSkip;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.replica.REFITStage;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.execution.REFITExecutorState;
import refit.replica.multichannel.REFITChannelItem;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITMCClientStage extends REFITStage {
	public final REFITObserver<REFITReply> replySender;
	public final REFITObserver<REFITChannelItem> channelRequestSender;
	public final REFITObserver<REFITChannelSkip> channelSkipSender;

	private final long[] seqNrLimit;
	private final REFITRequest[] lastRequest;
	private final boolean[] pendingRequest;

	// executor side
	public REFITMCClientStage(REFITBaseReplica replica) {
		super(REFITSchedulerTaskType.COUPLING_STAGE, replica);
		replySender = new REFITObserver<>();
		channelRequestSender = new REFITObserver<>();
		channelSkipSender = new REFITObserver<>();
		seqNrLimit = new long[REFITConfig.TOTAL_NR_OF_CLIENTS];
		lastRequest = new REFITRequest[REFITConfig.TOTAL_NR_OF_CLIENTS];
		pendingRequest = new boolean[REFITConfig.TOTAL_NR_OF_CLIENTS];
	}

	@Override
	public String toString() {
		return "MCLI-SRC";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REQUEST:
			handleRequest((REFITRequest) message);
			break;
		case CHECKPOINT_STATE:
			handleCheckpointState(((REFITCheckpointState) message));
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleRequest(REFITRequest message) {
		int clientIdx = message.from - REFITConfig.TOTAL_NR_OF_REPLICAS;
		REFITRequest lastReq = lastRequest[clientIdx];
		if (lastReq != null && lastReq.uid.equals(message.uid)) {
			// send requests only once
			REFITLogger.logWarning(this, "Dropping already processed client request " + message);
			return;
		}

		lastRequest[clientIdx] = message;
		channelSkipSender.broadcast(new REFITChannelSkip((short) clientIdx, message.uid.seqNr));
		pendingRequest[clientIdx] = true;

		if (message.uid.seqNr >= seqNrLimit[clientIdx]) {
			// no tokens, skip sending for now
//			REFITLogger.logWarning(this, "Dropping client request as no tokens are available " + message);
			return;
		}

		sendLastRequest(clientIdx);
	}

	private void sendLastRequest(int clientIdx) {
		REFITRequest message = lastRequest[clientIdx];
		channelRequestSender.broadcast(new REFITChannelItem(message.uid.seqNr, (short) clientIdx, message));
		pendingRequest[clientIdx] = false;
	}

	private void handleCheckpointState(REFITCheckpointState stableCheckpoint) {
		short executionGroupId = REFITReplicaGroups.getExecutionGroupId(replica.id);
		REFITExecutorState metadata = new REFITExecutorState(stableCheckpoint.data[0].getBuffer());

		for (short i = 0; i < metadata.nodeProgresses.length; i++) {
			// the checkpoint contains client request hashes, forward them to the client stage
			REFITUniqueID uid = new REFITUniqueID(i, metadata.nodeProgresses[i]);
			REFITReply reply = new REFITReply(uid, replica.id, executionGroupId, 0, metadata.nodeSeqs[i],
					false, ByteBuffer.wrap(metadata.replyHashes[i]), RequestMode.WRITE, (short) -1);
			replySender.broadcast(reply);
		}
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_PROGRESS:
			REFITChannelProgress progress = (REFITChannelProgress) message;
			short subChannel = progress.subChannelID;
			seqNrLimit[subChannel] = progress.baseSeqNr + progress.tokens;
			if (pendingRequest[subChannel]) {
				sendLastRequest(subChannel);
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.type);
		}
	}
}
