package refit.replica.multichannel;

import java.util.HashMap;
import java.util.Map;
import java.util.function.LongFunction;

import refit.config.REFITLogger;
import refit.message.REFITMessage;

public class REFITChannelInbox<T extends REFITMessage> {
	private final Map<Long, REFITChannelInboxCertificate<T>> certs;
	private final LongFunction<REFITChannelInboxCertificate<T>> certConstructor;

	private final int networkTokenCount;
	private long firstUnconfirmedSeqNr = 0;
	private int outputTokenCount = 0;
	private long nextSeqNr = 0;

	public REFITChannelInbox(LongFunction<REFITChannelInboxCertificate<T>> certConstructor, int networkTokenCount) {
		this.certConstructor = certConstructor;
		certs = new HashMap<>();

		this.networkTokenCount = networkTokenCount;
	}

	public boolean addShare(T share) {
		long seqNr = share.uid.seqNr;
		if (seqNr < nextSeqNr || seqNr >= nextSeqNr + networkTokenCount) {
			return false;
		}

		REFITChannelInboxCertificate<T> cert = ensureRequestCert(seqNr);
		if (cert.isStable()) {
			return false;
		}
		// it is no problem to keep a certificate that was only created for an invalid share
		// as the seq# range we accept is strictly limited
		boolean progress = cert.add(share);
		if (!progress) {
			REFITLogger.logWarning(this, "Ignoring invalid share " + share);
		}

		return cert.isStable() && seqNr == nextSeqNr;
	}

	private REFITChannelInboxCertificate<T> ensureRequestCert(long seqNr) {
		if (!certs.containsKey(seqNr)) {
			certs.put(seqNr, certConstructor.apply(seqNr));
		}
		return certs.get(seqNr);
	}

	public long getFirstUnconfirmedSeqNr() {
		return firstUnconfirmedSeqNr;
	}

	public long getNextSeqNr() {
		return nextSeqNr;
	}

	// call repeatedly until null is returned
	public T[] takeCompleted() {
		if (certs.containsKey(nextSeqNr) && certs.get(nextSeqNr).isStable() && nextSeqNr < firstUnconfirmedSeqNr + outputTokenCount) {
			T[] result = certs.remove(nextSeqNr).getResult();
			nextSeqNr++;
			return result;
		}
		return null;
	}

	public boolean isOutputBlocked() {
		return nextSeqNr >= firstUnconfirmedSeqNr + outputTokenCount;
	}

	public boolean moveStartTo(long seqNr, long interestSeqNr, int outputTokenCount) {
		firstUnconfirmedSeqNr = Math.max(firstUnconfirmedSeqNr, seqNr);
		this.outputTokenCount = outputTokenCount;

		interestSeqNr = Math.max(seqNr, interestSeqNr);

		if (interestSeqNr <= nextSeqNr) return false;
		nextSeqNr = Math.max(interestSeqNr, nextSeqNr);
		certs.keySet().removeIf(aLong -> aLong < nextSeqNr);
		return true;
	}
}
