package refit.replica.multichannel;

import java.util.function.LongFunction;

import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITChannelProgress;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITChannelReceiverStage extends REFITAbstractChannelReceiver {

	private static final int ACK_DELAY = 10;
	private final int MOVE_ACKS;

	private final boolean[] senderGroup;

	private final InboxWithMove<REFITShare>[] subChannels;
	private final short groupIdx;
	private REFITChannelAckOutbox ackOutbox;

	public REFITChannelReceiverStage(REFITBaseReplica replica, boolean[] senderGroup, short groupIdx, short subChannelCount, int networkTokenCount, int senderFaults) {
		super(REFITSchedulerTaskType.CHANNEL_RECEIVER_STAGE, replica);
		markInterleaved();
		MOVE_ACKS = senderFaults + 1;

		this.groupIdx = groupIdx;
		resultSender = new REFITObserver<>();
		skipSender = new REFITObserver<>();
		this.senderGroup = senderGroup;

		//noinspection unchecked
		this.subChannels = (InboxWithMove<REFITShare>[]) new InboxWithMove[subChannelCount];
		for (int i = 0; i < subChannels.length; i++) {
			final short subChannelId = (short) i;
			subChannels[i] = new InboxWithMove<>(
					(seqNr) -> new REFITFullShareCertificate(new REFITUniqueID(subChannelId, seqNr), replica.mac, senderGroup, senderFaults),
					networkTokenCount, MOVE_ACKS);
		}

		ackOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_ACK, senderGroup);
	}

	@Override
	public String toString() {
		return "CH-DST";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHANNEL_SHARE:
			handleShare((REFITShare) message);
			break;
		case CHANNEL_SENDER_ACK:
			handleSenderAck(((REFITChannelAck) message));
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleShare(REFITShare message) {
		if (!REFITReplicaGroups.checkInGroup(senderGroup, message.from)) {
			REFITLogger.logError(this, "Dropping message from unexpected sender " + message);
			return;
		}
		short subChannelID = message.uid.nodeID;
		if (subChannelID < 0 || subChannelID >= subChannels.length) {
			REFITLogger.logError(this, "FullShare for out of bounds subChannel " + message);
			return;
		}

		REFITChannelInbox<REFITShare> subChannel = subChannels[subChannelID];
		boolean progress = subChannel.addShare(message);
		if (progress) {
			sendFromChannel(subChannelID);
		}
	}

	private void sendFromChannel(short subChannelID) {
		REFITChannelInbox<REFITShare> subChannel = subChannels[subChannelID];
		while (true) {
			REFITShare[] share = subChannel.takeCompleted();
			if (share == null) {
				break;
			}
			resultSender.broadcast(new REFITChannelItem(share[0].uid.seqNr, subChannelID, share[0].getContent()));
			// ackOutbox.addProgress(subChannelID, subChannel.getStartSeqNr() - 1);
		}
	}

	private void handleSenderAck(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, senderGroup)) {
			return;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			long ackedSeqNr = message.changedAcks[i];
			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}
			InboxWithMove<REFITShare> channel = subChannels[subChannelID];
			if (channel.handleMove(message.from, ackedSeqNr)) {
				// requested later window start than its current position
				if (channel.getExpectedMove() > channel.getFirstUnconfirmedSeqNr()) {
					skipSender.broadcast(new REFITInstruction.REFITChannelSkip(subChannelID, channel.getExpectedMove()));
				}
			}
		}
	}

	@Override
	protected void stageComplete() {
		if (ackOutbox.hasNewEntries() && !hasTimeout()) {
			setTimeout(ACK_DELAY);
		}
	}

	@Override
	public void timeout() {
		ackOutbox.trySend();
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_PROGRESS:
			REFITChannelProgress progress = (REFITChannelProgress) message;
			REFITChannelInbox<REFITShare> subChannel = subChannels[progress.subChannelID];
			subChannel.moveStartTo(progress.baseSeqNr, progress.interestSeqNr, progress.tokens);
			if (subChannel.getFirstUnconfirmedSeqNr() > 0) {
				ackOutbox.addProgress(progress.subChannelID, subChannel.getFirstUnconfirmedSeqNr() - 1);
			}
			sendFromChannel(progress.subChannelID);
			break;
		case CHANNEL_CONFIG:
			REFITInstruction.REFITChannelConfig config = (REFITInstruction.REFITChannelConfig) message;
			if (groupIdx == config.groupIdx) {
				if (!config.isActive) throw new UnsupportedOperationException();
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.type);
		}
	}

	private static class InboxWithMove<T extends REFITMessage> extends REFITChannelInbox<T> {
		private final REFITChannelAckHandler moveHandler;

		public InboxWithMove(LongFunction<REFITChannelInboxCertificate<T>> certConstructor, int networkTokenCount, int MOVE_ACKS) {
			super(certConstructor, networkTokenCount);
			moveHandler = new REFITChannelAckHandler(MOVE_ACKS);
		}

		public boolean handleMove(short from, long seqNr) {
			REFITChannelAckHandler.AckStatus status = moveHandler.handleProgress(from, seqNr);
			if (status.equals(REFITChannelAckHandler.AckStatus.INVALID)) {
				REFITLogger.logWarning(this, "ignoring old sender-move from " + from + "@" + seqNr);
				return false;
			}

			return status.equals(REFITChannelAckHandler.AckStatus.CHANGED);
		}

		public long getExpectedMove() {
			return moveHandler.getMergedSeqNr();
		}
	}
}
