package refit.replica.multichannel;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.function.BiFunction;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITChannelProgress;
import refit.message.REFITInstruction.REFITChannelSkip;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITIncrementalOutbox;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;

public class REFITChannelSenderStage extends REFITAbstractChannelSender {

	private static final int SENDER_ACK_DELAY = 10;
	private final int REQUIRED_ACKS;

	private long nextMoveTimeout = -1;
	private ArrayDeque<REFITChannelItem> pendingRequests = new ArrayDeque<>();
	private boolean delayRequestUntilTimeout = false;

	private final boolean[][] receiverGroups;
	private final boolean[] mergedReceiverGroup;

	private final int networkTokenCount;
	private final int localTokenCount;
	private final int historyCount;
	private final SubChannel[] subChannels;
	private REFITChannelAckOutbox moveOutbox;

	private final BiFunction<REFITShare, boolean[], REFITShare> messageFilter;

	public REFITChannelSenderStage(REFITBaseReplica replica, boolean[][] receiverGroups, short subChannelCount,
	                               int tokenCount, int historyCount, BiFunction<REFITShare, boolean[], REFITShare> messageFilter,
	                               int receiverFaults) {
		super(REFITSchedulerTaskType.CHANNEL_SENDER_STAGE, replica);
		markInterleaved();
		REQUIRED_ACKS = receiverFaults + 1;
		this.messageFilter = messageFilter;
		progressSender = new REFITObserver<>();

		this.receiverGroups = receiverGroups;

		mergedReceiverGroup = new boolean[receiverGroups[0].length];
		for (int j = 0; j < receiverGroups.length; j++) {
			boolean[] receiverGroup = receiverGroups[j];
			if (REFITReplicaGroups.countGroupSize(receiverGroup) < REQUIRED_ACKS + receiverFaults) {
				throw new AssertionError("Expecting at least 2f+1 receiver replicas!");
			}
			for (int i = 0; i < receiverGroup.length; i++) {
				if (mergedReceiverGroup[i] && receiverGroup[i]) {
					throw new AssertionError("Receiver groups must be non-overlapping!");
				}
				mergedReceiverGroup[i] |= receiverGroup[i];
			}
		}

		this.networkTokenCount = tokenCount;
		this.localTokenCount = tokenCount;
		this.historyCount = historyCount;
		subChannels = new SubChannel[subChannelCount];
		for (short i = 0; i < subChannels.length; i++) {
			subChannels[i] = new SubChannel(i);
		}

		moveOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_SENDER_ACK, mergedReceiverGroup);
	}

	@Override
	public String toString() {
		return "CH-SRC";
	}

	@Override
	public void setGroupActive(int groupIdx, boolean isActive) {
		if (!isActive) throw new UnsupportedOperationException();
	}

	@Override
	public void init() {
		super.init();
		for (short i = 0; i < subChannels.length; i++) {
			progressSender.broadcast(new REFITChannelProgress(i, 0, -1, localTokenCount));
		}
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHANNEL_ITEM:
			pendingRequests.add(((REFITChannelItem) message));
			break;
		case CHANNEL_ACK:
			handleAck((REFITChannelAck) message);
			break;
		case INSTRUCTION:
			handleInstruction(((REFITInstruction) message));
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_SKIP:
			REFITChannelSkip skip = (REFITChannelSkip) message;
			if (subChannels[skip.subChannelID].ackedUpTo < skip.baseSeqNr) {
				moveOutbox.addProgress(skip.subChannelID, skip.baseSeqNr);
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.instructionType);
		}
	}

	private boolean handleRequest(REFITChannelItem message) {
		return subChannels[message.subChannelID].queueRequest(message.uid.seqNr, message.content);
	}

	private void handleAck(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, mergedReceiverGroup)) {
			return;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			long ackedSeqNr = message.changedAcks[i];
			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}
			subChannels[subChannelID].handleAck(message.from, ackedSeqNr);
		}
	}

	@Override
	protected void stageComplete() {
		super.stageComplete();

		if (moveOutbox.hasNewEntries() && nextMoveTimeout < 0) {
			nextMoveTimeout = REFITTime.currentTimeMillis.getAsLong() + SENDER_ACK_DELAY;
			if (!hasTimeout()) {
				setTimeoutAbsolute(nextMoveTimeout);
			}
		}

		while (!delayRequestUntilTimeout && !pendingRequests.isEmpty()) {
			// only execute one expensive request at a time
			REFITChannelItem request = pendingRequests.removeFirst();
			if (handleRequest(request)) {
				delayRequestUntilTimeout = true;
				setTimeout(0);
				break;
			}
		}
	}

	@Override
	public void timeout() {
		long currentTime = REFITTime.currentTimeMillis.getAsLong();
		if (nextMoveTimeout >= 0) {
			if (currentTime >= nextMoveTimeout) {
				moveOutbox.trySend();
				nextMoveTimeout = -1;
			} else {
				setTimeoutAbsolute(nextMoveTimeout);
			}
		}

		// wakeup will trigger another execution of stageComplete
		delayRequestUntilTimeout = false;
	}

	private class SubChannel {
		private final short subChannelID;
		private REFITIncrementalOutbox outbox;

		private long ackedUpTo = -1;
		private long[] ackedSeqNr = new long[REFITConfig.TOTAL_NR_OF_REPLICAS];

		private long[] groupProgresses;

		public SubChannel(short subChannelID) {
			this.subChannelID = subChannelID;
			this.outbox = new REFITIncrementalOutbox(replica, REFITChannelSenderStage.this, historyCount + localTokenCount, true);
			outbox.updateSendLimits(networkTokenCount);

			Arrays.fill(ackedSeqNr, -1);

			groupProgresses = new long[receiverGroups.length];
			Arrays.fill(groupProgresses, -1L);
		}

		@Override
		public String toString() {
			return "ACS-SRC[" + subChannelID + "]";
		}

		public boolean queueRequest(long seqNr, REFITMessage message) {
			// only serialize and authenticate message if someone is interested
			if (!outbox.isAfterGarbageLimit(seqNr)) return false;

			REFITShare fullShare = new REFITShare(REFITMessageType.CHANNEL_SHARE, seqNr, subChannelID, replica.id, message);
			fullShare.serializeMessage();
			replica.mac.appendSignature(fullShare);


			if (messageFilter != null) {
				for (boolean[] receiverGroup : receiverGroups) {
					REFITMessage m = messageFilter.apply(fullShare, receiverGroup);
					outbox.queue(m, receiverGroup);
				}
			} else {
				outbox.queue(fullShare, mergedReceiverGroup);
			}

			return true;
		}

		public void handleAck(short from, long ackSeqNr) {
			if (ackSeqNr <= ackedSeqNr[from]) {
				REFITLogger.logWarning(this, "ignoring old ack from " + from + "@" + ackSeqNr);
				return;
			}
			ackedSeqNr[from] = ackSeqNr;
			// once a replica has acked a request, we no longer need to keep it
			outbox.collectIndividualGarbageBefore(from, ackedSeqNr[from] + 1);

			for (int i = 0; i < receiverGroups.length; i++) {
				long[] acked = ackedSeqNr.clone();
				for (int j = 0; j < receiverGroups[i].length; j++) {
					if (!receiverGroups[i][j]) {
						// blank out nodes which are not in this group
						acked[j] = 0;
					}
				}
				Arrays.sort(acked);
				groupProgresses[i] = acked[acked.length - REQUIRED_ACKS];
			}

			long oldAckedUpTo = ackedUpTo;

			// use minimum progress over all channels
			ackedUpTo = Long.MAX_VALUE;
			for (int i = 0; i < groupProgresses.length; i++) {
				long groupProgress = groupProgresses[i];
				ackedUpTo = Math.min(ackedUpTo, groupProgress);
			}

			outbox.collectGarbageBefore(ackedUpTo - historyCount + 1);
			outbox.updateIndividualSendLimit(from, ackedSeqNr[from] + networkTokenCount + 1);

			if (oldAckedUpTo != ackedUpTo) {
				progressSender.broadcast(new REFITChannelProgress(subChannelID, ackedUpTo + 1, -1, localTokenCount));
			}
		}
	}
}
