package refit.replica.multichannel;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.function.Function;

import refit.message.*;
import refit.util.REFITPayload;

public class REFITShareBundle extends REFITMessage {

	private final REFITMessage content;
	private final short[] froms;
	private final byte[][] proof;

	private static short[] getFroms(REFITShare[] shares) {
		short[] froms = new short[shares.length];
		for (int i = 0; i < froms.length; i++) {
			froms[i] = shares[i].from;
		}
		return froms;
	}

	private static byte[][] getProofs(REFITShare[] shares) {
		byte[][] proofs = new byte[shares.length][];
		for (int i = 0; i < proofs.length; i++) {
			proofs[i] = shares[i].getPadding();
		}
		return proofs;
	}

	public REFITShareBundle(long seqNr, short subChannel, short from, REFITMessage content, REFITShare[] shares) {
		this(new REFITUniqueID(subChannel, seqNr), from, content, getFroms(shares), getProofs(shares));
	}

	private REFITShareBundle(REFITUniqueID uid, short from, REFITMessage content, short[] froms, byte[][] proof) {
		super(REFITMessageType.CHANNEL_SHARE_BUNDLE, uid, from);
		this.content = content;
		this.froms = froms;
		this.proof = proof;
//		if (proof.length != REFITShareCollectionCertificate.STABILITY_THRESHOLD)
//			throw new AssertionError("Execute certificate of wrong size: " + proof.length);
		for (byte[] bytes : proof) {
			if (bytes.length != REFITMessageAuthentication.SIGNATURE_SIZE) {
				throw new AssertionError("Invalid signature size");
			}
		}
	}

	public REFITShareBundle(ByteBuffer buffer) {
		super(REFITMessageType.CHANNEL_SHARE_BUNDLE, buffer);
		int proofLength = buffer.get();
		proof = new byte[proofLength][];
		froms = new short[proofLength];
		for (int i = 0; i < proof.length; i++) {
			froms[i] = buffer.getShort();
			byte[] signature = new byte[REFITMessageAuthentication.SIGNATURE_SIZE];
			buffer.get(signature);
			proof[i] = signature;
		}

		byte payloadType = buffer.get();
		if (payloadType == 1) {
			this.content = REFITNoOpRequest.INSTANCE;
		} else {
			this.content = REFITMessage.createMessage(buffer.slice());
			buffer.position(buffer.position() + content.getMessageSize());
		}
	}

	@Override
	public REFITMessageType groupType() {
		return REFITMessageType.CHANNEL_MSG;
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.MAC_AUTHENTICATOR;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%s|%s}", type, from, uid, REFITPayload.toString(content.getHash()));
	}

	public REFITMessage getContent() {
		return content;
	}

	public REFITShareBundle toFilteredBundle(Function<REFITMessage, REFITMessage> filter) {
		if (filter == null) return this;

		REFITMessage message = filter.apply(content);
		REFITShareBundle bundle = new REFITShareBundle(uid, from, message, froms, proof);
		bundle.setRecipientCount(recipientCount());
		bundle.serializeMessage();
		bundle.setPadding(getPadding());
		bundle.markVerified();
		return bundle;
	}

	public REFITShare[] getProof() {
		REFITShare[] shares = new REFITShare[proof.length];
		for (int i = 0; i < shares.length; i++) {
			REFITShare share = new REFITShare(REFITMessageType.CHANNEL_SHARE_PART, uid.seqNr, uid.nodeID, froms[i], content);
			share.serializeMessage();
			share.setPadding(proof[i]);
			shares[i] = share;
		}
		return shares;
	}

	@Override
	public boolean equals(Object object) {
		if (this == object) return true;
		if (object == null || getClass() != object.getClass()) return false;
		if (!super.equals(object)) return false;
		REFITShareBundle other = (REFITShareBundle) object;
		return Arrays.equals(proof, other.proof);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		buffer.put((byte) proof.length);
		for (int i = 0; i < proof.length; i++) {
			buffer.putShort(froms[i]);
			buffer.put(proof[i]);
		}
		if (content.type == REFITMessageType.REQUEST_NO_OP) {
			buffer.put((byte) 1);
		} else {
			buffer.put((byte) 0);
			content.serializeInto(buffer);
		}
	}

	@Override
	protected int getHeaderSize() {
		int size = Byte.BYTES + Byte.BYTES;
		for (byte[] bytes : proof) {
			size += Short.BYTES + bytes.length;
		}

		return (super.getHeaderSize() + size);
	}

	@Override
	protected int calculatePayloadSize() {
		int size;
		if (content.type == REFITMessageType.REQUEST_NO_OP) {
			size = 0;
		} else {
			size = content.calculateMessageSize();
		}
		size += Byte.BYTES + Byte.BYTES;
		for (byte[] bytes : proof) {
			size += Short.BYTES + bytes.length;
		}
		return (super.calculatePayloadSize() + size);
	}

	@Override
	protected byte[] calculateHash() {
		if (content.type != REFITMessageType.REQUEST_NO_OP) {
			content.getHash();
		}

		// Prepare message buffer
		ByteBuffer messageBuffer = getInternalBuffer();
		messageBuffer.position(REFITBaseMessage.HEADER_SIZE);
		messageBuffer.limit(getHeaderSize());

		// Calculate hash
		MessageDigest digest = REFITPayload.getDigest();
		digest.update(messageBuffer);
		if (content.type != REFITMessageType.REQUEST_NO_OP) {
			digest.update(content.getHash());
		}
		return digest.digest();
	}
}
