package refit.replica.multichannel;

import java.nio.ByteBuffer;

import refit.config.REFITLogger;
import refit.message.REFITBatch;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITUniqueID;
import refit.replica.REFITReplicaGroups;
import refit.util.REFITBallotBox;

public class REFITShareBundleCertificate implements REFITChannelInboxCertificate<REFITShareBundle> {

	private final int STABILITY_THRESHOLD;

	private final REFITMessageAuthentication mac;

	private final REFITUniqueID uid;
	private final int groupID;
	private boolean[] legitimateSenders;

	private REFITShareBundle[] result;

	public REFITShareBundleCertificate(REFITUniqueID uid, REFITMessageAuthentication mac, short replicaID, boolean[] legitimateSenders, int senderFaults) {
		this.mac = mac;

		STABILITY_THRESHOLD = senderFaults + 1;
		this.uid = uid;
		this.groupID = REFITReplicaGroups.getExecutionGroupId(replicaID);
		this.legitimateSenders = legitimateSenders;
	}

	@Override
	public boolean add(REFITShareBundle share) {
		// Check reply
		if (share == null)
			return false;
		if (!uid.equals(share.uid))
			return false;
		if (!mac.verifyMulticastMAC(share, legitimateSenders))
			return false;
		if (share.getProof().length != STABILITY_THRESHOLD)
			return false;

		REFITBallotBox<Short, ByteBuffer, REFITShare> ballot = new REFITBallotBox<>(STABILITY_THRESHOLD);

		// The REFITShareBundle must contain a full request by construction
		for (REFITShare proof : share.getProof()) {
			// NOte: If any preexecute fails validation, then we may assume that the sender wanted to attack us
			if (!mac.verifySignature(proof, legitimateSenders))
				return false;
			if (!proof.uid.equals(share.uid)) {
				REFITLogger.logWarning(this, "UID mismatch between REFITShareBundle and contained REFITShare " + share);
				return false;
			}
			boolean validVote = ballot.add(proof.from, ByteBuffer.wrap(proof.getPayloadHash()), proof);
			if (!validVote) {
				REFITLogger.logWarning(this, "REFITShareBundle contains two REFITShare from the same(?) replica " + share);
				return false;
			}
		}

		if (share.getContent() instanceof REFITBatch && !((REFITBatch) share.getContent()).verifyGroupOnly((short) groupID))
			return false;

		if (ballot.getDecision() == null || ballot.votesDiffer()) {
			REFITLogger.logWarning(this, "REFITShareBundle contains mismatched shares " + share);
			return false;
		}

		result = new REFITShareBundle[]{share};
		return true;
	}

	@Override
	public boolean isStable() {
		return (result != null);
	}

	@Override
	public REFITShareBundle[] getResult() {
		return result;
	}
}
