package refit.replica.multichannel;

import java.util.*;
import java.util.function.LongFunction;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITChannelConfig;
import refit.message.REFITInstruction.REFITChannelProgress;
import refit.message.REFITInstruction.REFITChannelSkip;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;

public class REFITSourceChannelReceiverStage extends REFITAbstractChannelReceiver {

	private static final long COUPLING_PROGRESS_TIMEOUT = 10000;
	private final int PROGRESS_ACKS;
	private final int MOVE_ACKS;
	private static final int ACK_DELAY = 10;

	private final boolean[] senderGroup;

	private final InboxWithMoveProgress<REFITShareBundle>[] subChannels;
	private Deque<TimeoutEntry> progressTimeouts;
	private final REFITChannelAckOutbox ackOutbox;
	private short selectedSender = 0;

	private final short groupIdx;
	private boolean isActive = false;
	private boolean checkProgress = false;
	private boolean gotInitialProgress = false;

	public REFITSourceChannelReceiverStage(REFITBaseReplica replica, boolean[] senderGroup, short groupIdx, short subChannelCount, int networkTokenCount, int senderFaults) {
		super(REFITSchedulerTaskType.CHANNEL_RECEIVER_STAGE, replica);
		markInterleaved();
		PROGRESS_ACKS = senderFaults + 1;
		MOVE_ACKS = senderFaults + 1;

		this.groupIdx = groupIdx;
		resultSender = new REFITObserver<>();
		skipSender = new REFITObserver<>();
		this.senderGroup = senderGroup;

		//noinspection unchecked
		this.subChannels = (InboxWithMoveProgress<REFITShareBundle>[]) new InboxWithMoveProgress[subChannelCount];
		for (int i = 0; i < subChannels.length; i++) {
			final short subChannelId = (short) i;
			subChannels[i] = new InboxWithMoveProgress<>(
					(seqNr) -> new REFITShareBundleCertificate(new REFITUniqueID(subChannelId, seqNr), replica.mac, replica.id, senderGroup, senderFaults),
					networkTokenCount, PROGRESS_ACKS, MOVE_ACKS);
		}

		progressTimeouts = new LinkedList<>();
		ackOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_ACK, senderGroup);
		// default to first replica in group
		for (int i = 0; i < senderGroup.length; i++) {
			if (senderGroup[i]) {
				selectedSender = (short) i;
				break;
			}
		}
		for (short i = 0; i < subChannels.length; i++) {
			setSelectedSender(i, selectedSender);
		}
	}

	@Override
	public String toString() {
		return "CH-DST";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHANNEL_SHARE_BUNDLE:
			handleShareBundle((REFITShareBundle) message);
			break;
		case CHANNEL_PROGRESS:
			handleProgress((REFITChannelAck) message);
			break;
		case CHANNEL_SENDER_ACK:
			handleSenderAck(((REFITChannelAck) message));
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleShareBundle(REFITShareBundle message) {
		if (!REFITReplicaGroups.checkInGroup(senderGroup, message.from)) {
			REFITLogger.logError(this, "Dropping message from unexpected sender " + message);
			return;
		}
		short subChannelID = message.uid.nodeID;
		if (subChannelID < 0 || subChannelID >= subChannels.length) {
			REFITLogger.logError(this, "FullShare for out of bounds subChannel " + message);
			return;
		}

		REFITChannelInbox<REFITShareBundle> subChannel = subChannels[subChannelID];
		boolean progress = subChannel.addShare(message);
		if (progress) {
			sendFromChannel(subChannelID);
		}
	}

	private void sendFromChannel(short subChannelID) {
		REFITChannelInbox<REFITShareBundle> subChannel = subChannels[subChannelID];
		while (true) {
			REFITShareBundle[] share = subChannel.takeCompleted();
			if (share == null) {
				break;
			}
			REFITShare fullShare = null;
			for (REFITShare proof : share[0].getProof()) {
				if (!proof.isHashOnly()) {
					fullShare = proof;
					break;
				}
			}
			if (fullShare == null) {
				throw new InternalError("Broken certificate implementation");
			}

			resultSender.broadcast(new REFITChannelItem(share[0].uid.seqNr, subChannelID, fullShare.getContent()));
			//ackOutbox.addProgress(subChannelID, subChannel.getStartSeqNr() - 1);
		}
	}

	private void setSelectedSender(short subChannelID, short selectedSender) {
		this.selectedSender = selectedSender;
		ackOutbox.addSpecialProgress((short) (-subChannelID - 1), selectedSender);
		REFITLogger.logDebug(this, "Switch sender to " + selectedSender);
	}

	private void handleProgress(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, senderGroup)) {
			return;
		}

		if (!gotInitialProgress) {
			// fully activate after receiving the first progress message
			isActive = true;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}
			InboxWithMoveProgress<REFITShareBundle> subChannel = subChannels[subChannelID];
			boolean madeProgress = subChannel.handleProgress(message.from, message.changedAcks[i]);
			if (madeProgress) {
				long timeout = REFITTime.currentTimeMillis.getAsLong() + COUPLING_PROGRESS_TIMEOUT;
				progressTimeouts.add(new TimeoutEntry(subChannelID, timeout, subChannel.getExpectedProgress()));
			}
		}
	}

	private boolean findProgressIssues(long time) {
		Iterator<TimeoutEntry> it = progressTimeouts.iterator();
		boolean hasIssues = false;
		while (it.hasNext()) {
			TimeoutEntry entry = it.next();
			if (time >= entry.timeout) {
				if (entry.expectedProgress < subChannels[entry.subChannelID].getNextSeqNr()) {
					// expected seqNr already reached -> timeout is fine
					it.remove();
				} else if (!subChannels[entry.subChannelID].isOutputBlocked()) {
					// don't blame the sender if we can't yet handle the message
					hasIssues = true;
				}
			} else {
				// timestamps are monotonic
				break;
			}
		}
		return hasIssues;
	}

	private void resetExpectedProgress(long oldTime) {
		ArrayDeque<TimeoutEntry> newTimeouts = new ArrayDeque<>();
		long time = REFITTime.currentTimeMillis.getAsLong();

		HashSet<Short> resetChannels = new HashSet<>();

		Iterator<TimeoutEntry> it = progressTimeouts.iterator();
		while (it.hasNext()) {
			TimeoutEntry entry = it.next();
			if (entry.timeout > oldTime) {
				// timeout entry did not expire yet
				if (resetChannels.contains(entry.subChannelID)) {
					// drop all other timeouts for channels that were reset
					it.remove();
				}
			} else {
				it.remove();
				// create a single timeout with the latest progress
				newTimeouts.add(new TimeoutEntry(entry.subChannelID, time + COUPLING_PROGRESS_TIMEOUT, subChannels[entry.subChannelID].getExpectedProgress()));
				resetChannels.add(entry.subChannelID);
			}
		}

		progressTimeouts.addAll(newTimeouts);
	}

	private void handleSenderAck(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, senderGroup)) {
			return;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			long ackedSeqNr = message.changedAcks[i];
			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}
			InboxWithMoveProgress<REFITShareBundle> channel = subChannels[subChannelID];
			if (channel.handleMove(message.from, ackedSeqNr)) {
				// requested later window start than its current position
				if (channel.getExpectedMove() > channel.getFirstUnconfirmedSeqNr()) {
					skipSender.broadcast(new REFITChannelSkip(subChannelID, channel.getExpectedMove()));
				}
			}
		}
	}

	@Override
	protected void stageComplete() {
		if (!isActive) return;

		if (checkProgress) {
			long time = REFITTime.currentTimeMillis.getAsLong();
			boolean foundIssues = findProgressIssues(time);
			if (foundIssues) {
				boolean[] resetChannel = new boolean[subChannels.length];
				for (TimeoutEntry entry : progressTimeouts) {
					if (entry.timeout > time) break;
					if (resetChannel[entry.subChannelID]) continue;

					long[] progress = subChannels[entry.subChannelID].getProgress();
					// switch to next sender that has claimed to have the request
					for (int i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
						short sender = REFITReplicaGroups.nextInGroup(senderGroup, selectedSender);
						if (progress[sender] >= entry.expectedProgress) {
							setSelectedSender(entry.subChannelID, sender);
							resetChannel[entry.subChannelID] = true;
							break;
						}
					}
				}
				resetExpectedProgress(time);
			}

			ackOutbox.trySend();
			checkProgress = false;
		}

		if (ackOutbox.hasNewEntries() && !hasTimeout()) {
			setTimeout(ACK_DELAY);
		}
		if (!progressTimeouts.isEmpty() && !hasTimeout()) {
			setTimeout(ACK_DELAY);
		}
	}

	@Override
	public void timeout() {
		checkProgress = true;
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_PROGRESS:
			REFITChannelProgress progress = (REFITChannelProgress) message;
			REFITChannelInbox<REFITShareBundle> subChannel = subChannels[progress.subChannelID];
			subChannel.moveStartTo(progress.baseSeqNr, progress.interestSeqNr, progress.tokens);
			if (subChannel.getFirstUnconfirmedSeqNr() > 0) {
				ackOutbox.addProgress(progress.subChannelID, subChannel.getFirstUnconfirmedSeqNr() - 1);
			}
			sendFromChannel(progress.subChannelID);
			break;
		case CHANNEL_CONFIG:
			REFITChannelConfig config = (REFITChannelConfig) message;
			if (groupIdx == config.groupIdx) {
				isActive = config.isActive;
				gotInitialProgress = true;
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.type);
		}
	}

	private static class InboxWithMoveProgress<T extends REFITMessage> extends REFITChannelInbox<T> {
		private final REFITChannelAckHandler moveHandler;
		private final REFITChannelAckHandler progressHandler;

		public InboxWithMoveProgress(LongFunction<REFITChannelInboxCertificate<T>> certConstructor, int networkTokenCount, int PROGRESS_ACKS, int MOVE_ACKS) {
			super(certConstructor, networkTokenCount);
			progressHandler = new REFITChannelAckHandler(PROGRESS_ACKS);
			moveHandler = new REFITChannelAckHandler(MOVE_ACKS);
		}

		public boolean handleProgress(short from, long seqNr) {
			REFITChannelAckHandler.AckStatus status = progressHandler.handleProgress(from, seqNr);
			if (status.equals(REFITChannelAckHandler.AckStatus.INVALID)) {
				REFITLogger.logWarning(this, "ignoring old progress from " + from + "@" + seqNr);
				return false;
			}

			return status.equals(REFITChannelAckHandler.AckStatus.CHANGED);
		}

		public long getExpectedProgress() {
			return progressHandler.getMergedSeqNr();
		}

		public long[] getProgress() {
			return progressHandler.getProgress();
		}

		public boolean handleMove(short from, long seqNr) {
			REFITChannelAckHandler.AckStatus status = moveHandler.handleProgress(from, seqNr);
			if (status.equals(REFITChannelAckHandler.AckStatus.INVALID)) {
				REFITLogger.logWarning(this, "ignoring old sender-move from " + from + "@" + seqNr);
				return false;
			}

			return status.equals(REFITChannelAckHandler.AckStatus.CHANGED);
		}

		public long getExpectedMove() {
			return moveHandler.getMergedSeqNr();
		}
	}

	private static class TimeoutEntry {
		public TimeoutEntry(short subChannelID, long timeout, long expectedProgress) {
			this.subChannelID = subChannelID;
			this.timeout = timeout;
			this.expectedProgress = expectedProgress;
		}

		final short subChannelID;
		final long timeout;
		final long expectedProgress;
	}
}
