package refit.replica.multichannel;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.function.BiFunction;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITChannelConfig;
import refit.message.REFITInstruction.REFITChannelProgress;
import refit.message.REFITInstruction.REFITChannelSkip;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITIncrementalOutbox;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;

public class REFITSourceChannelSenderStage extends REFITAbstractChannelSender {

	private final int REQUIRED_ACKS;
	private static final long ACK_DELAY = 10;

	private long nextAckTimeout = -1;
	private ArrayDeque<REFITChannelItem> pendingRequests = new ArrayDeque<>();
	private boolean delayRequestUntilTimeout = false;

	private final boolean[] senderGroup;
	private final boolean[] nonLocalSenderGroup;
	private final boolean[][] receiverGroups;
	private final boolean[] mergedReceiverGroup;
	private final short mergedReceiverGroupSize;
	private final int[] reverseGroupsMap;

	private final boolean[] activeGroups;

	private final int networkTokenCount;
	private final int localTokenCount;
	private final int historyCount;
	private final REFITIncrementalOutbox[] localSendSubChannels;
	private final REFITChannelInbox<REFITShare>[] localReceiveSubChannels;
	private final SubChannel[] subChannels;
	private REFITChannelAckOutbox moveOutbox;

	private final REFITChannelAckOutbox localAckOutbox;
	private final REFITChannelAckOutbox progressOutbox;

	private final BiFunction<REFITShareBundle, boolean[], REFITShareBundle> messageFilter;

	public REFITSourceChannelSenderStage(REFITBaseReplica replica, boolean[] senderGroup, boolean[][] receiverGroups, short subChannelCount,
	                                     int tokenCount, int historyCount, BiFunction<REFITShareBundle, boolean[], REFITShareBundle> messageFilter,
	                                     int senderFaults, int receiverFaults) {
		super(REFITSchedulerTaskType.CHANNEL_SENDER_STAGE, replica);
		markInterleaved();
		REQUIRED_ACKS = receiverFaults + 1;
		this.messageFilter = messageFilter;
		progressSender = new REFITObserver<>();

		this.senderGroup = senderGroup;
		this.nonLocalSenderGroup = senderGroup.clone();
		nonLocalSenderGroup[replica.id] = false;
		this.receiverGroups = receiverGroups;

		mergedReceiverGroup = new boolean[receiverGroups[0].length];
		reverseGroupsMap = new int[receiverGroups[0].length];
		Arrays.fill(reverseGroupsMap, -1);
		for (int j = 0; j < receiverGroups.length; j++) {
			boolean[] receiverGroup = receiverGroups[j];
			if (REFITReplicaGroups.countGroupSize(receiverGroup) < REQUIRED_ACKS + receiverFaults) {
				throw new AssertionError("Expecting at least 2f+1 receiver replicas!");
			}
			for (int i = 0; i < receiverGroup.length; i++) {
				if (mergedReceiverGroup[i] && receiverGroup[i]) {
					throw new AssertionError("Receiver groups must be non-overlapping!");
				}
				mergedReceiverGroup[i] |= receiverGroup[i];
				if (receiverGroup[i]) {
					reverseGroupsMap[i] = j;
				}
			}
		}
		mergedReceiverGroupSize = (short) REFITReplicaGroups.countGroupSize(mergedReceiverGroup);

		this.networkTokenCount = tokenCount;
		this.localTokenCount = tokenCount;
		this.historyCount = historyCount;
		subChannels = new SubChannel[subChannelCount];
		localSendSubChannels = new REFITIncrementalOutbox[subChannelCount];
		//noinspection unchecked
		localReceiveSubChannels = (REFITChannelInbox<REFITShare>[]) new REFITChannelInbox[subChannelCount];
		for (short i = 0; i < subChannels.length; i++) {
			localSendSubChannels[i] = new REFITIncrementalOutbox(replica, this, historyCount + localTokenCount, true);
			localSendSubChannels[i].updateSendLimits(networkTokenCount);
			final short subChannelId = i;
			localReceiveSubChannels[i] = new REFITChannelInbox<>(
					(seqNr) -> new REFITShareCollectionCertificate(new REFITUniqueID(subChannelId, seqNr), replica.mac, senderGroup, senderFaults),
					networkTokenCount);
			localReceiveSubChannels[i].moveStartTo(0, -1, localTokenCount);
			subChannels[i] = new SubChannel(i);
		}

		localAckOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_LOCAL_ACK, senderGroup);
		progressOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_PROGRESS, mergedReceiverGroup);
		moveOutbox = new REFITChannelAckOutbox(replica, this, REFITMessageType.CHANNEL_SENDER_ACK, mergedReceiverGroup);

		activeGroups = new boolean[receiverGroups.length];
		Arrays.fill(activeGroups, true);
	}

	@Override
	public String toString() {
		return "CH-SRC";
	}

	@Override
	public void init() {
		super.init();
		for (short i = 0; i < subChannels.length; i++) {
			progressSender.broadcast(new REFITChannelProgress(i, 0, -1, localTokenCount));
		}
	}

	public void setGroupActive(int groupIdx, boolean isActive) {
		activeGroups[groupIdx] = isActive;
		for (int i = 0; i < receiverGroups[groupIdx].length; i++) {
			if (!receiverGroups[groupIdx][i]) continue;
			progressOutbox.setTargetDisable((short) i, !isActive);
		}
		REFITLogger.logWarning(this, "Setting channel group " + groupIdx + " to status " + isActive);
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHANNEL_ITEM:
			pendingRequests.add(((REFITChannelItem) message));
			break;
		case CHANNEL_SHARE_PART:
			handleShare((REFITShare) message);
			break;
		case CHANNEL_LOCAL_ACK:
			handleLocalAck((REFITChannelAck) message);
			break;
		case CHANNEL_ACK:
			handleAck((REFITChannelAck) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_CONFIG:
			handleChannelConfig((REFITChannelConfig) message);
			break;
		case CHANNEL_SKIP:
			REFITChannelSkip skip = (REFITChannelSkip) message;
			if (subChannels[skip.subChannelID].ackedUpTo < skip.baseSeqNr) {
				moveOutbox.addProgress(skip.subChannelID, skip.baseSeqNr);
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.instructionType);
		}
	}

	private void handleChannelConfig(REFITChannelConfig message) {
		setGroupActive(message.groupIdx, message.isActive);
	}

	private boolean handleRequest(REFITChannelItem message) {
		REFITIncrementalOutbox subChannel = localSendSubChannels[message.subChannelID];
		long seqNr = message.uid.seqNr;

		// only serialize and authenticate message if someone is interested
		if (!subChannel.isAfterGarbageLimit(seqNr) && localReceiveSubChannels[message.subChannelID].getFirstUnconfirmedSeqNr() > seqNr)
			return false;

		REFITShare fullShare = new REFITShare(REFITMessageType.CHANNEL_SHARE_PART, seqNr, message.subChannelID, replica.id, message.content);
		fullShare.serializeMessage();
		replica.mac.appendSignature(fullShare);
		// distribute hash-only share
		REFITShare hashShare = fullShare.toHashOnly();
		subChannel.queue(hashShare, nonLocalSenderGroup);
		handleShare(fullShare);
		return true;
	}

	private void handleShare(REFITShare message) {
		// the signature check happens inside the channel
		short subChannelID = message.uid.nodeID;
		if (subChannelID < 0 || subChannelID >= localReceiveSubChannels.length) {
			REFITLogger.logWarning(this, "Out of bounds subchannel id " + message);
			return;
		}
		REFITChannelInbox<REFITShare> channel = localReceiveSubChannels[subChannelID];
		channel.addShare(message);
		sendFromChannel(subChannelID);
	}

	private void sendFromChannel(short subChannelID) {
		REFITChannelInbox<REFITShare> subChannel = localReceiveSubChannels[subChannelID];
		while (true) {
			REFITShare[] shares = subChannel.takeCompleted();
			if (shares == null) {
				break;
			}
			REFITShareBundle bundle = new REFITShareBundle(shares[0].uid.seqNr, subChannelID, replica.id, shares[0].getContent(), shares);
			subChannels[subChannelID].queueRequest(bundle.uid.seqNr, bundle);
		}
		localAckOutbox.addProgress(subChannelID, subChannel.getNextSeqNr() - 1);
		progressOutbox.addProgress(subChannelID, subChannel.getNextSeqNr() - 1);
	}

	private void handleLocalAck(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, senderGroup)) {
			return;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			long ackedSeqNr = message.changedAcks[i];
			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}

			REFITIncrementalOutbox subChannel = localSendSubChannels[subChannelID];
			subChannel.updateIndividualSendLimit(message.from, ackedSeqNr + networkTokenCount + 1);
			subChannel.collectIndividualGarbageBefore(message.from, ackedSeqNr + 1);
		}
	}

	private void handleAck(REFITChannelAck message) {
		if (!replica.mac.verifyUnicastMAC(message, mergedReceiverGroup)) {
			return;
		}

		for (int i = 0; i < message.changedSubChannels.length; i++) {
			short subChannelID = message.changedSubChannels[i];
			long ackedSeqNr = message.changedAcks[i];

			if (subChannelID >= -subChannels.length && subChannelID < 0) {
				boolean isSelectedSender = (replica.id == message.changedAcks[i]);
				subChannelID = (short) -(subChannelID + 1);
				subChannels[subChannelID].setSenderStatus(message.from, isSelectedSender);
				continue;
			}

			if (subChannelID < 0 || subChannelID >= subChannels.length) {
				continue;
			}
			subChannels[subChannelID].handleAck(message.from, ackedSeqNr);

			long ackedUpTo = subChannels[subChannelID].getAckedUpTo();
			localSendSubChannels[subChannelID].collectGarbageBefore(ackedUpTo - historyCount + 1);

			// adjust window to only collect bundles that can still be useful
			REFITChannelInbox<REFITShare> receiveSubChannel = localReceiveSubChannels[subChannelID];
			receiveSubChannel.moveStartTo(ackedUpTo - historyCount + 1, -1, historyCount + localTokenCount);

			// no longer interested in requests older than the start of the receiveSubChannel window
			localAckOutbox.addProgress(subChannelID, receiveSubChannel.getNextSeqNr() - 1);
			progressOutbox.addProgress(subChannelID, receiveSubChannel.getNextSeqNr() - 1);

			// with the moved window we might have enough tokens to send further requests
			sendFromChannel(subChannelID);
		}
	}

	@Override
	protected void stageComplete() {
		super.stageComplete();

		if ((localAckOutbox.hasNewEntries() || progressOutbox.hasNewEntries() || moveOutbox.hasNewEntries()) && nextAckTimeout < 0) {
			nextAckTimeout = REFITTime.currentTimeMillis.getAsLong() + ACK_DELAY;
			if (!hasTimeout()) {
				setTimeoutAbsolute(nextAckTimeout);
			}
		}

		while (!delayRequestUntilTimeout && !pendingRequests.isEmpty()) {
			// only execute one expensive request
			REFITChannelItem request = pendingRequests.removeFirst();
			if (handleRequest(request)) {
				delayRequestUntilTimeout = true;
				setTimeout(0);
				break;
			}
		}
	}

	@Override
	public void timeout() {
		long currentTime = REFITTime.currentTimeMillis.getAsLong();
		if (nextAckTimeout >= 0) {
			if (currentTime >= nextAckTimeout) {
				localAckOutbox.trySend();
				progressOutbox.trySend();
				moveOutbox.trySend();
				nextAckTimeout = -1;
			} else {
				setTimeoutAbsolute(nextAckTimeout);
			}
		}

		// wakeup will trigger another execution of stageComplete
		delayRequestUntilTimeout = false;
	}

	private class SubChannel {
		private final short subChannelID;
		private REFITIncrementalOutbox outbox;

		private long ackedUpTo = -1;
		private long[] ackedSeqNr = new long[REFITConfig.TOTAL_NR_OF_REPLICAS];
		private boolean[] activeSender = new boolean[REFITConfig.TOTAL_NR_OF_REPLICAS];

		private long[] groupProgresses;

		public SubChannel(short subChannelID) {
			this.subChannelID = subChannelID;
			this.outbox = new REFITIncrementalOutbox(replica, REFITSourceChannelSenderStage.this, historyCount + localTokenCount, true);

			Arrays.fill(ackedSeqNr, -1);
			groupProgresses = new long[receiverGroups.length];
			Arrays.fill(groupProgresses, -1L);
		}

		@Override
		public String toString() {
			return "ACS-SRC[" + subChannelID + "]";
		}

		public void queueRequest(long seqNr, REFITShareBundle bundle) {
			// only serialize and authenticate message if someone is interested
			if (!outbox.isAfterGarbageLimit(seqNr)) return;

			bundle.setRecipientCount(mergedReceiverGroupSize);
			bundle.serializeMessage();
			replica.mac.appendMulticastMAC(bundle, mergedReceiverGroup);

			if (messageFilter != null) {
				for (boolean[] receiverGroup : receiverGroups) {
					REFITMessage message = messageFilter.apply(bundle, receiverGroup);
					outbox.queue(message, receiverGroup);
				}
			} else {
				outbox.queue(bundle, mergedReceiverGroup);
			}
		}

		public void handleAck(short from, long ackSeqNr) {
			if (ackSeqNr <= ackedSeqNr[from]) {
				REFITLogger.logWarning(this, "ignoring old ack from " + from + "@" + ackSeqNr);
				return;
			}
			ackedSeqNr[from] = ackSeqNr;
			// once a replica has acked a request, we no longer need to keep it
			outbox.collectIndividualGarbageBefore(from, ackedSeqNr[from] + 1);

			for (int i = 0; i < receiverGroups.length; i++) {
				long[] acked = ackedSeqNr.clone();
				for (int j = 0; j < receiverGroups[i].length; j++) {
					if (!receiverGroups[i][j]) {
						// blank out nodes which are not in this group
						acked[j] = 0;
					}
				}
				Arrays.sort(acked);
				groupProgresses[i] = acked[acked.length - REQUIRED_ACKS];
			}

			long oldAckedUpTo = ackedUpTo;

			// use minimum progress over all channels
			ackedUpTo = Long.MAX_VALUE;
			for (int i = 0; i < groupProgresses.length; i++) {
				if (!activeGroups[i]) continue;
				long groupProgress = groupProgresses[i];
				ackedUpTo = Math.min(ackedUpTo, groupProgress);
			}

			outbox.collectGarbageBefore(ackedUpTo - historyCount + 1);
			updateSendLimit(from);

			if (oldAckedUpTo != ackedUpTo) {
				progressSender.broadcast(new REFITChannelProgress(subChannelID, ackedUpTo + 1, -1, localTokenCount));
			}
		}

		public void setSenderStatus(short from, boolean active) {
			activeSender[from] = active;
			updateSendLimit(from);
		}

		protected void updateSendLimit(short from) {
			if (activeSender[from] && activeGroups[reverseGroupsMap[from]]) {
				outbox.updateIndividualSendLimit(from, ackedSeqNr[from] + networkTokenCount + 1);
			}
		}

		public long getAckedUpTo() {
			return ackedUpTo;
		}
	}
}
