package refit.replica.probe;

import java.nio.ByteBuffer;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITUniqueID;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbeCommit;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrePrepare;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrepare;
import refit.util.REFITBallotBox;


public class REFITProbePBFTCertificate {


	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private final boolean[] legitimateSenders;

	private REFITUniqueID uid;

	public REFITProbePrePrepare prePrepare;
	public REFITMessage message;
	private REFITBallotBox<Short, ByteBuffer, REFITProbeOrderMessage> prepares;
	private REFITBallotBox<Short, ByteBuffer, REFITProbeCommit> commits;


	public REFITProbePBFTCertificate(int groupSize, boolean[] legitimateSenders) {
		// calculates ceil((n+f+1)/2)
		int PREPARED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		int COMMITTED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		this.legitimateSenders = legitimateSenders;
		if (groupSize < PREPARED_THRESHOLD || groupSize < COMMITTED_THRESHOLD) {
			throw new AssertionError(
					String.format("Group size %d is too small for the expected number of faults %d",
							groupSize, REFITConfig.FAULTS_TO_TOLERATE));
		}
		this.prepares = new REFITBallotBox<>(PREPARED_THRESHOLD);
		this.commits = new REFITBallotBox<>(COMMITTED_THRESHOLD);
	}


	@Override
	public String toString() {
		return "{" + uid + ": " + isPrepared() + " " + isCommitted() + "}";
	}

	public void init(REFITUniqueID uid) {
		this.uid = uid;
		this.prePrepare = null;
		this.message = null;
		this.prepared = false;
		this.committed = false;

		prepares.clear();
		commits.clear();
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public boolean addPrePrepare(REFITProbePrePrepare prePrepare, REFITMessageAuthentication mac) {
		if (prePrepare == null) return false;
		if (this.prePrepare != null) return false;
		if (message != null) return false;
		if (!uid.equals(prePrepare.message.uid)) {
			REFITLogger.logWarning(this, "got preprepare uid " + prePrepare.message.uid + " expected uid " + uid);
			return false;
		}
		if (!mac.verifyMulticastMAC(prePrepare, legitimateSenders)) return false;
		if (!mac.verifyMulticastMACRequest(prePrepare.message)) return false;
		this.prePrepare = prePrepare;
		message = prePrepare.message;
		prepares.add(prePrepare.from, ByteBuffer.wrap(new byte[0]), prePrepare);

		checkPrepared();
		return true;
	}

	public boolean addPrepare(REFITProbePrepare prepare, REFITMessageAuthentication mac) {
		if (prepare == null) return false;
		if (!uid.equals(prepare.uid)) {
			REFITLogger.logWarning(this, "got prepare uid " + prepare.uid + " expected uid " + uid);
			return false;
		}
		if (prepares.hasVoted(prepare.from)) return false;
		if (!mac.verifyMulticastMAC(prepare, legitimateSenders)) return false;
		prepares.add(prepare.from, ByteBuffer.wrap(new byte[0]), prepare);

		checkPrepared();
		return true;
	}


	public AddResult addCommit(REFITProbeCommit commit, REFITMessageAuthentication mac) {
		if (commit == null) return AddResult.INVALID;
		if (!uid.equals(commit.uid)) {
			REFITLogger.logWarning(this, "got commit uid " + commit.uid + " expected uid " + uid);
			return AddResult.INVALID;
		}
		if (commits.hasVoted(commit.from)) return AddResult.DUPLICATE;
		if (!mac.verifyMulticastMAC(commit, legitimateSenders)) return AddResult.INVALID;

		commits.add(commit.from, ByteBuffer.wrap(new byte[0]), commit);

		checkCommitted();
		return AddResult.ADDED;
	}

	// ######################
	// # CERTIFICATE STATES #
	// ######################

	private boolean prepared;
	private boolean committed;

	public boolean hasPrePrepare() {
		return prePrepare != null;
	}

	public boolean isPrepared() {
		return prepared;
	}

	public boolean preparedVotesDiffer() {
		return prepares.votesDiffer();
	}

	public boolean isCommitted() {
		return committed;
	}

	private void checkPrepared() {
		if (prepared) return;
		if (prePrepare == null) return;
		ByteBuffer decision = prepares.getDecision();
		if (decision == null) return;
		if (!prepares.getDecidingBallots().contains(prePrepare)) return;
		prepared = true;
	}

	private void checkCommitted() {
		if (committed) return;
		if (!prepared) return;
		if (prePrepare == null) return;

		ByteBuffer decision = commits.getDecision();
		if (decision == null) return;
		committed = true;
	}

	public boolean isComplete() {
		return committed && prepares.getVoteCount() == REFITConfig.TOTAL_NR_OF_REPLICAS
				&& commits.getVoteCount() == REFITConfig.TOTAL_NR_OF_REPLICAS;
	}
}
