package refit.replica.probe;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;

import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.replica.REFITBaseReplica;
import refit.replica.order.REFITOrderGroups;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbeCommit;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrePrepare;
import refit.replica.probe.REFITProbeProtocolMessages.REFITProbePrepare;

public class REFITProbeSlot {

	private static class MessageWithTimestamp {
		public final long timestamp;
		public final REFITProbeOrderMessage message;

		MessageWithTimestamp(long timestamp, REFITProbeOrderMessage message) {
			this.timestamp = timestamp;
			this.message = message;
		}
	}

	private enum ProbeStep {
		PRE_PREPARE,
		PROPOSE,
		PREPARE,
		COMMIT,
		STRAGGLER,
		COMPLETE
	}

	private final boolean[] orderReplicas;

	private MessageList[] messageBuckets;
	private REFITRequest probe;
	private ProbeStep step;
	private REFITUniqueID currentUID;
	private REFITBaseReplica replica;
	private REFITProbeStage stage;
	private REFITProbePBFTCertificate certificate;
	private REFITProbePrePrepare prePrepare;
	private long timeout;

	private long prePrepareTime;
	private short[] preparedTime;
	private short[] commitTime;
	private byte[] probeRiddle;
	private byte[][] prepareRiddles;
	private byte[][] commitRiddles;

	private static class MessageList extends LinkedList<MessageWithTimestamp> {
	}


	public REFITProbeSlot(REFITBaseReplica replica, REFITProbeStage stage) {
		this.replica = replica;
		this.stage = stage;
		messageBuckets = new MessageList[REFITMessageType.values().length];
		for (int i = 0; i < messageBuckets.length; i++) {
			messageBuckets[i] = new MessageList();
		}

		step = ProbeStep.PRE_PREPARE;
		preparedTime = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		commitTime = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];

		orderReplicas = REFITOrderGroups.getGroupByID(0);
		certificate = new REFITProbePBFTCertificate(REFITConfig.TOTAL_NR_OF_REPLICAS, orderReplicas);
		prepareRiddles = new byte[REFITConfig.TOTAL_NR_OF_REPLICAS][];
		commitRiddles = new byte[REFITConfig.TOTAL_NR_OF_REPLICAS][];
	}

	public void setProposal(REFITRequest probe) {
		if (step == ProbeStep.PRE_PREPARE || acceptNewProbe(probe)) {
			this.probe = probe;
			this.step = ProbeStep.PROPOSE;
		}
	}

	public boolean insertMessage(REFITProbeOrderMessage message) {
		// FIXME limit number of messages with new uid that are kept
		messageBuckets[message.type.getMagic()].add(new MessageWithTimestamp(System.currentTimeMillis(), message));
		return true;
	}

	private MessageWithTimestamp takeMessage(REFITMessageType type) {
		return messageBuckets[type.getMagic()].poll();
	}

	private MessageWithTimestamp takeCurrentMessage(REFITMessageType type, REFITUniqueID uid) {
		MessageList bucket = messageBuckets[type.getMagic()];
		Iterator<MessageWithTimestamp> iter = bucket.iterator();
		while (iter.hasNext()) {
			MessageWithTimestamp message = iter.next();
			if (message.message.uid.seqNr <= uid.seqNr) {
				iter.remove();
				return message;
			}
		}
		return null;
	}

	public void execute() {
		handlePropose();
		handlePrePrepare();
		if (currentUID == null) return;
		handlePrepare();
		handleCommit();

		if (timeout != 0 && timeout <= System.currentTimeMillis()) {
			handleTimeout();
		}
	}

	private boolean acceptNewProbe(REFITRequest probe) {
		if (!replica.mac.verifyMulticastMAC(probe, REFITMessageAuthentication.ANY_SENDER)) return false;
		if (!probe.mode.equals(RequestMode.PROBE)) return false;

		if (currentUID == null) return true;
		// FIXME check if enough time has passed
		return probe.uid.seqNr > currentUID.seqNr;
	}


	private void handlePropose() {
		if (step == ProbeStep.PROPOSE) {
			ByteBuffer payload = probe.getPayload();
			probeRiddle = new byte[payload.remaining()];
			payload.get(probeRiddle);
			byte[] stepKey = replica.mac.calcStepKey(probeRiddle, 0, probe.from, replica.id);
			byte[] hashChain = replica.mac.deriveDirectHashChain(stepKey, probeRiddle);
			REFITProbePrePrepare prePrepare = new REFITProbePrePrepare(probe.uid, replica.id, probe, hashChain);

			certificate.init(probe.uid);
			probe = null;
			prePrepare.serializeMessage();
			replica.mac.appendMulticastMAC(prePrepare);
			replica.replicaMulticast(prePrepare);
			boolean acceptedOwnPrePrepare = certificate.addPrePrepare(prePrepare, replica.mac);
			if (!acceptedOwnPrePrepare) {
				throw new InternalError("Rejected own preprepare " + prePrepare);
			}

			acceptPrePrepare(prePrepare);
		}
	}

	private void acceptPrePrepare(REFITProbePrePrepare prePrepare) {
		step = ProbeStep.PREPARE;
		currentUID = prePrepare.message.uid;
		this.prePrepare = prePrepare;
		timeout = 0;

		prePrepareTime = System.currentTimeMillis();
		Arrays.fill(preparedTime, REFITConfig.PROBE_TIMEOUT);
		preparedTime[prePrepare.from] = 0;
		Arrays.fill(commitTime, REFITConfig.PROBE_TIMEOUT);

		ByteBuffer payload = prePrepare.message.getPayload();
		probeRiddle = new byte[payload.remaining()];
		payload.get(probeRiddle);
		Arrays.fill(prepareRiddles, null);
		prepareRiddles[prePrepare.from] = prePrepare.riddlePart;
		Arrays.fill(commitRiddles, null);
	}

	private void handlePrePrepare() {
		REFITProbePrePrepare newPrePrepare = null;

		while (true) {
			MessageWithTimestamp tm = takeMessage(REFITMessageType.PROBE_PBFT_PRE_PREPARE);
			if (tm == null) {
				break;
			}
			REFITProbePrePrepare prePrepare = (REFITProbePrePrepare) tm.message;

			if (certificate.hasPrePrepare() && !acceptNewProbe(prePrepare.message)) {
				REFITLogger.logProbe(this, "Too early new preprepare " + prePrepare);
				continue;
			}

			certificate.init(prePrepare.message.uid);
			step = ProbeStep.PRE_PREPARE;

			// FIXME reset logic?
			boolean preprepareAccepted = certificate.addPrePrepare(prePrepare, replica.mac);
			if (!preprepareAccepted) {
				REFITLogger.logProbe(this, "Ignoring preprepare with invalid message " + prePrepare);
				continue;
			}

			newPrePrepare = prePrepare;
		}

		if (newPrePrepare != null && step == ProbeStep.PRE_PREPARE) {
			acceptPrePrepare(newPrePrepare);

			// send own prepares
			byte[] stepKey = replica.mac.calcStepKey(probeRiddle, 1, currentUID.nodeID, replica.id);
			byte[] hashChain = replica.mac.deriveDirectHashChain(stepKey, prePrepare.riddlePart);
			prepareRiddles[replica.id] = hashChain;

			REFITProbePrepare ownPrepare = new REFITProbePrepare(newPrePrepare, replica.id, hashChain);
			ownPrepare.serializeMessage();
			replica.mac.appendMulticastMAC(ownPrepare);
			replica.replicaMulticast(ownPrepare);
			boolean acceptedOwnPrepare = certificate.addPrepare(ownPrepare, replica.mac);
			if (!acceptedOwnPrepare) {
				throw new InternalError("Rejected own prepare " + ownPrepare);
			}

			// own prepare also counts in the message certificate
			preparedTime[replica.id] = 0;
		}
	}

	private void handlePrepare() {
		boolean addedPrepare = false;

		while (true) {
			MessageWithTimestamp tm = takeCurrentMessage(REFITMessageType.PROBE_PBFT_PREPARE, currentUID);
			if (tm == null) {
				break;
			}
			REFITProbePrepare prepare = (REFITProbePrepare) tm.message;

			boolean prepareAdded = certificate.addPrepare(prepare, replica.mac);
			if (!prepareAdded) {
				REFITLogger.logProbe(this, "Ignoring prepare with invalid message " + prepare);
				continue;
			}
			ByteBuffer payload = prepare.getCommand();
			byte[] riddle = new byte[payload.remaining()];
			payload.get(riddle);

			prepareRiddles[prepare.from] = riddle;
			preparedTime[prepare.from] = (short) (tm.timestamp - prePrepareTime);
			addedPrepare = true;
		}

		if (step == ProbeStep.PREPARE && certificate.isPrepared()) {
			step = ProbeStep.COMMIT;

			// send own commit
			byte[] hashChain = updateOwnCommitRiddle();

			REFITProbeCommit ownCommit = new REFITProbeCommit(prePrepare, replica.id, hashChain);
			ownCommit.serializeMessage();
			replica.mac.appendMulticastMAC(ownCommit);
			replica.replicaMulticast(ownCommit);
			REFITProbePBFTCertificate.AddResult ownCommitResult = certificate.addCommit(ownCommit, replica.mac);
			if (ownCommitResult != REFITProbePBFTCertificate.AddResult.ADDED) {
				throw new InternalError("Rejected own commit " + ownCommit);
			}
			commitTime[ownCommit.from] = (short) (System.currentTimeMillis() - prePrepareTime);
		} else if (step.ordinal() > ProbeStep.PREPARE.ordinal() && step != ProbeStep.COMPLETE && addedPrepare) {
			// update own commit
			updateOwnCommitRiddle();
			checkResend();
		}
	}

	private byte[] updateOwnCommitRiddle() {
		byte[] stepKey = replica.mac.calcStepKey(probeRiddle, 2, currentUID.nodeID, replica.id);
		byte[] hashChain = replica.mac.deriveHashChain(stepKey, prepareRiddles);
		commitRiddles[replica.id] = hashChain;
		return hashChain;
	}

	private void handleCommit() {
		boolean addedCommit = false;

		while (true) {
			MessageWithTimestamp tm = takeCurrentMessage(REFITMessageType.PROBE_PBFT_COMMIT, currentUID);
			if (tm == null) {
				break;
			}
			REFITProbeCommit commit = (REFITProbeCommit) tm.message;

			REFITProbePBFTCertificate.AddResult addResult = certificate.addCommit(commit, replica.mac);
			if (addResult != REFITProbePBFTCertificate.AddResult.ADDED) {
				REFITLogger.logProbe(this, "Ignoring commit with invalid message " + commit);
				continue;
			}
			ByteBuffer payload = commit.getCommand();
			byte[] riddle = new byte[payload.remaining()];
			payload.get(riddle);

			commitRiddles[commit.from] = riddle;
			commitTime[commit.from] = (short) (tm.timestamp - prePrepareTime);
			addedCommit = true;
		}

		if (step == ProbeStep.COMMIT && certificate.isCommitted()) {
			step = ProbeStep.STRAGGLER;
			sendProbeReply();

//            printTimes();
		} else if (step == ProbeStep.STRAGGLER && addedCommit) {
			checkResend();
//            printTimes();
		}
	}

	private void sendProbeReply() {
		// send reply to client
		byte[] stepKey = replica.mac.calcStepKey(probeRiddle, 3, currentUID.nodeID, replica.id);
		byte[] hashChain = replica.mac.deriveHashChain(stepKey, commitRiddles);

		REFITProbeResult result = new REFITProbeResult(hashChain, preparedTime, commitTime);
		result.serializeMessage();

		REFITReply ownReply = new REFITReply(currentUID, replica.id, (short) -1, (short) -1, -1, true,
				result.getBuffer(), RequestMode.PROBE, (short) -1);
		ownReply.serializeMessage();
		replica.mac.appendUnicastMAC(ownReply, currentUID.nodeID);
		replica.replicaUnicast(ownReply, currentUID.nodeID);
	}

	private void printTimes() {
		System.out.println("Preprepare time " + prePrepareTime);
		for (short aPreparedTime : preparedTime) {
			System.out.println("Prepare time " + aPreparedTime);
		}
		for (short aCommitTime : commitTime) {
			System.out.println("Commit time " + aCommitTime);
		}
	}

	private void checkResend() {
		if (!certificate.isComplete()) {
			timeout = prePrepareTime + REFITConfig.PROBE_TIMEOUT;
			stage.addSlotTimeout(this, REFITConfig.PROBE_TIMEOUT);
		} else {
			sendFinalProbe();
		}
	}

	private void sendFinalProbe() {
		if (step.ordinal() < ProbeStep.STRAGGLER.ordinal()) {
			return;
		}
		sendProbeReply();
		timeout = 0;
		step = ProbeStep.COMPLETE;
	}

	private void handleTimeout() {
		sendFinalProbe();
	}

}
