package refit.replica.update;

import java.util.HashMap;
import java.util.Map;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITConfigurationNotification;
import refit.message.REFITInstruction.REFITProgressNotification;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.scheduler.REFITSchedulerTaskType;


public class REFITUpdateStage extends REFITStage {

	public final REFITObserver<REFITUpdateMessage> stableUpdateSender;

	private final Map<Long, REFITUpdateCertificate> certificates;
	private long executionStageProgress;
	private boolean[] updateRecipients;
	private final boolean[] updateSenders;


	public REFITUpdateStage(REFITBaseReplica replica, boolean[] updateSenders) {
		super(REFITSchedulerTaskType.UPDATE_STAGE, replica);
		this.certificates = new HashMap<>();
		this.executionStageProgress = -1L;
		this.updateRecipients = null;
		this.updateSenders = updateSenders;

		stableUpdateSender = new REFITObserver<>();
	}


	@Override
	public String toString() {
		return "UPDTE";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case UPDATE:
		case BATCH_UPDATE:
			handleUpdate((REFITUpdateMessage) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case PROGRESS_NOTIFICATION:
			// Perform garbage collection and update local progress information
			REFITProgressNotification progressNotification = (REFITProgressNotification) instruction;
			long newProgress = Math.max(0, progressNotification.agreementSeqNr + REFITConfig.INSTANCES_AFTER_CHECKPOINT - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS);

			for (long seqNr = executionStageProgress; seqNr < newProgress; seqNr++)
				certificates.remove(seqNr);
			executionStageProgress = newProgress;
			break;
		case CONFIGURATION_NOTIFICATION:
			// Update the information on where to send updates
			REFITConfigurationNotification configurationNotification = (REFITConfigurationNotification) instruction;
			updateRecipients = configurationNotification.protocol.getUpdateRecipients();
			if (!configurationNotification.protocol.hasUpdateRecipients()) updateRecipients = null;
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	private boolean isPrimary() {
		// FIXME account for view changes!!!
		return (replica.id == 0);
	}

	private boolean isHashedFollower(long agreementSeqNr) {
		// TODO: only 1 follower replica has to send the full reply
		// FIXME generalize + view changes
		return (agreementSeqNr % 2) == (replica.id - 1);
	}

	private void handleUpdate(REFITUpdateMessage update) {
		// Do not process old updates
		if (update.agreementSeqNr() < executionStageProgress) return;
		// nothing to do if there are no update recipients
		if (updateRecipients == null) return;

		// Distribute hash update based on local full update
		if (update.from == replica.id && update.isInternal()) {
			// Serialize updates in batch update
			if (update.type == REFITMessageType.BATCH_UPDATE) {
				for (REFITUpdate upd : ((REFITBatchUpdate) update).updates) upd.serializeMessage();
			}

			// Serialize and distribute update
			REFITUpdateMessage ownUpdate = update;
			if (REFITConfig.USE_HASHED_UPDATES) {
				if (isPrimary() || isHashedFollower(update.agreementSeqNr())) {
					ownUpdate = update.toHashedUpdate();
				}
			}
			ownUpdate.setRecipientCount((short) REFITConfig.PASSIVE_REPLICAS.length);
			ownUpdate.serializeMessage();
			replica.mac.appendMulticastMAC(ownUpdate, updateRecipients);
			replica.replicaMulticast(ownUpdate, updateRecipients);
		}

		// Only store update if the local replica is an update recipient
		if (!updateRecipients[replica.id]) return;
		if (REFITLogger.LOG_UPDATE) REFITLogger.logUpdate(this, "received update " + update);

		// Get update certificate
		REFITUpdateCertificate certificate = certificates.get(update.agreementSeqNr());
		if (certificate == null) {
			certificate = new REFITUpdateCertificate(update.uid, updateSenders);
			certificates.put(update.agreementSeqNr(), certificate);

			// Debug check
			if (REFITConfig.ENABLE_DEBUG_CHECKS) {
				int mapSize = certificates.size();
				if ((mapSize > 1000) && ((mapSize % 1000) == 0))
					REFITLogger.logDebug(this, "update-certificate store reached " + mapSize + " entries");
			}
		}

		// Store update
		boolean success = certificate.add(update, replica.mac);
		if (!success) {
			REFITLogger.logWarning(this, "bad UPDATE (" + update + ")");
			return;
		}

		// Check whether the update became stable
		if (!certificate.isStable()) return;

		// Handle stable update
		handleStableUpdate(certificate.result);
	}

	private void handleStableUpdate(REFITUpdateMessage stableUpdate) {
		// Stable update reached
		certificates.remove(stableUpdate.agreementSeqNr());
		if (REFITLogger.LOG_UPDATE) REFITLogger.logUpdate(this, "stable update for " + stableUpdate.agreementSeqNr());

		// Forward update operation
		stableUpdateSender.broadcast(stableUpdate);
	}

}
