package refit.scheduler;

import javax.management.InstanceNotFoundException;
import javax.management.openmbean.CompositeData;
import java.io.IOException;
import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.*;
import java.util.concurrent.CountDownLatch;

import com.sun.management.GarbageCollectionNotificationInfo;
import com.sun.management.GcInfo;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.util.REFITTime;


public class REFITScheduler extends Thread {

	// ####################
	// # SCHEDULER OBJECT #
	// ####################

	private static final int STATISTICS_INTERVAL = 10000;
	private static final int TASK_TYPE_SUB_ITERATION_LIMIT = 16;
	private static final int HANG_TIMEOUT_US = 500 * 1000;

	private final Selector selector;

	private final List<REFITSchedulerTask> assignedTasks;
	private boolean[] readyTasks;
	private REFITSchedulerTask[] tasks;
	private boolean[] isQueued;
	private EnumMap<REFITSchedulerTaskType, ArrayDeque<SelectionKey>> queues;
	private int queueSize;

	private final CountDownLatch startupSync;
	private final REFITHangDetector hangDetector;

	private final LinkedList<SlowTask> slowTasks = new LinkedList<>();
	private volatile long lastGCDuration;
	private volatile long lastGCEndTime;


	public REFITScheduler(CountDownLatch sync, REFITHangDetector detector) {
		// Create list of assigned tasks
		this.assignedTasks = new LinkedList<>();
		this.taskTimeouts = new TreeMap<>();
		startupSync = sync;
		hangDetector = detector;

		// Create selector
		Selector s = null;
		try {
			s = Selector.open();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		this.selector = s;
	}


	// ###################
	// # TASK ASSIGNMENT #
	// ###################

	public void assignTasks(REFITSchedulerGroup group, REFITSchedulerTaskType taskType) {
		Collection<REFITSchedulerTask> entry = group.takeTasks(taskType);
		if (entry == null) return;
		for (REFITSchedulerTask task : entry) assignTask(task);
	}

	protected void assignTask(REFITSchedulerTask task) {
		assignedTasks.add(task);
		task.setScheduler(this);
	}

	public SelectionKey registerIOTask(REFITSchedulerTask task, SelectableChannel channel, int operations) throws IOException {
		if (REFITConfig.ENABLE_DEBUG_CHECKS && !assignedTasks.contains(task)) {
			throw new IllegalArgumentException("Task not owned by this scheduler wants to register as io task " + task.toString());
		}
		return channel.register(selector, operations, task);
	}

	public void deregisterIOTask(SelectableChannel channel) {
		SelectionKey key = channel.keyFor(selector);
		if (key == null) {
			throw new IllegalArgumentException("Channel is not registered with this selector!");
		}
		key.cancel();
	}


	// ##################
	// # TASK EXECUTION #
	// ##################

	private final TreeMap<Long, LinkedList<REFITSchedulerTask>> taskTimeouts;

	public void notifyProgress(REFITSchedulerTask task) {
		readyTasks[task.taskID()] = true;
		if (this == Thread.currentThread()) return;
		selector.wakeup();
	}


	public void notifySetTimeout(REFITSchedulerTask task, long timeout) {
		if (this != Thread.currentThread()) {
			throw new IllegalArgumentException("Timeout must only be set by the task itself!");
		}
		if (!taskTimeouts.containsKey(timeout)) {
			taskTimeouts.put(timeout, new LinkedList<>());
		}
		taskTimeouts.get(timeout).add(task);
	}

	@Override
	public void run() {
		initRun();
		if (REFITConfig.TRACK_SLOW_STAGES) {
			subscribeToGCEvents();
		}

		if (startupSync != null) {
			startupSync.countDown();
			try {
				startupSync.await();
			} catch (InterruptedException e) {
				e.printStackTrace();
				System.exit(1);
			}
		}

		// Allow for initialization on the correct thread
		for (REFITSchedulerTask task : tasks) task.init();

		// For a single stage it's usually ok to not return to the scheduler
		REFITHangDetector.Entry watchdog = (hangDetector != null && tasks.length > 1)
				? hangDetector.register(getName(), HANG_TIMEOUT_US) : null;

		if (REFITConfig.COLLECT_STAGE_STATISTICS) runDebug(watchdog);
		else runDefault(watchdog);
	}

	private void initRun() {
		// Prepare tasks
		tasks = assignedTasks.toArray(new REFITSchedulerTask[0]);
		for (short i = 0; i < tasks.length; i++) tasks[i].setTaskID(i);

		readyTasks = new boolean[tasks.length];

		// store task reference instead of taskID to avoid autoboxing of int to Integer
		queues = new EnumMap<>(REFITSchedulerTaskType.class);
		for (REFITSchedulerTaskType type : REFITSchedulerTaskType.values()) {
			queues.put(type, new ArrayDeque<>());
		}
		isQueued = new boolean[tasks.length];
	}

	private void runDefault(REFITHangDetector.Entry watchdog) {
		selector.wakeup();

		while (!isInterrupted()) {
			// Collect tasks that are able to perform I/O operations
			if (watchdog != null) watchdog.disable();
			selectReadyTasks();
			if (watchdog != null) watchdog.refresh();

			// Only allow handling timeouts in the first readyTasks iterations
			// This prevents timeouts from keeping a stage permanently active
			boolean allowTimeout = true;

			// Handle runnable tasks; Called "sub-iteration", process internal messages exhaustively
			// before reading new messages from the network
			while (!isInterrupted()) {
				// Execute runnable tasks
				boolean progress = false;
				long lastStart = (REFITConfig.TRACK_SLOW_STAGES) ? System.nanoTime() : 0;
				for (int i = 0; i < tasks.length; i++) {
					if (!readyTasks[i]) continue;
					readyTasks[i] = false;
					progress = true;
					tasks[i].work(allowTimeout);

					if (REFITConfig.TRACK_SLOW_STAGES) {
						lastStart = trackSlowStage(lastStart, tasks[i]);
					}
				}

				// Break if none of the tasks has been executed in this loop
				if (!progress) break;
				allowTimeout = false;
			}

			if (REFITConfig.TRACK_SLOW_STAGES) {
				printSlowStages();
			}
		}
		if (watchdog != null) watchdog.disable();
	}

	private long trackSlowStage(long lastStart, REFITSchedulerTask task) {
		long tss = REFITTime.timeSinceStart();
		long taskEndTime = System.nanoTime();
		if (taskEndTime - lastStart > 5 * 1000 * 1000) {
			slowTasks.add(new SlowTask(taskEndTime - lastStart,
					tss, "Slow task " + task));
		}
		lastStart = taskEndTime;
		return lastStart;
	}

	private void printSlowStages() {
		while (!slowTasks.isEmpty()) {
			SlowTask head = slowTasks.peekFirst();
			if (head.endTimeInMs - head.durationInNs / 1000 / 1000 - 1 < lastGCEndTime) {
				// Discount GC time from slow stage runs
				long compensatedDurationInNs = head.durationInNs - lastGCDuration * 1000 * 1000;
				if (head.endTimeInMs + 1 < lastGCEndTime) {
					compensatedDurationInNs = head.durationInNs;
				}
				if (compensatedDurationInNs > 5 * 1000 * 1000) {
					REFITLogger.logWarning(this, head.msg + " took " + compensatedDurationInNs / 1000 + " us");
				}
				slowTasks.removeFirst();
			} else {
				break;
			}
		}
	}

	private void runDebug(REFITHangDetector.Entry watchdog) {
		// Prepare statistics
		int epochCounter = 0;
		long epochStartTimestamp = REFITTime.currentTimeMillis.getAsLong();
		long selectDurations = 0L;
		long selectDurationsMax = 0L;
		int[] executions = new int[tasks.length];
		long[] executionDurations = new long[tasks.length];
		long[] executionDurationsMax = new long[tasks.length];
		long[] throughputs = new long[tasks.length];
		long[] totalThroughputs = new long[tasks.length];
		long scheduleLoopMax = 0L;

		selector.wakeup();
		while (!isInterrupted()) {
			// Start a new epoch
			epochCounter++;
			long selectDuration = -REFITTime.nanoTime.getAsLong();

			// Collect tasks that are able to perform I/O operations
			if (watchdog != null) watchdog.disable();
			selectReadyTasks();
			if (watchdog != null) watchdog.refresh();
			// Only allow handling timeouts in the first readyTasks iterations
			// This prevents timeouts from keeping a stage permanently active
			boolean allowTimeout = true;

			// Collect select statistics
			selectDuration += REFITTime.nanoTime.getAsLong();
			selectDurations += selectDuration;
			selectDurationsMax = Math.max(selectDurationsMax, selectDuration);

			long scheduleLoopDuration = -REFITTime.nanoTime.getAsLong();

			// Handle runnable tasks
			while (!isInterrupted()) {
				// Execute runnable tasks
				boolean progress = false;
				for (int i = 0; i < tasks.length; i++) {
					if (!readyTasks[i]) continue;
					readyTasks[i] = false;
					progress = true;

					// Execute task
					long duration = -REFITTime.nanoTime.getAsLong();
					tasks[i].work(allowTimeout);
					duration += REFITTime.nanoTime.getAsLong();
					executions[i]++;
					executionDurations[i] += duration;
					executionDurationsMax[i] = Math.max(executionDurationsMax[i], duration);
					long throughput = tasks[i].collectThroughput();
					throughputs[i] += throughput;
					totalThroughputs[i] += throughput;
				}

				// Break if none of the tasks has been executed in this loop
				if (!progress) break;
				allowTimeout = false;
			}

			scheduleLoopDuration += REFITTime.nanoTime.getAsLong();
			scheduleLoopMax = Math.max(scheduleLoopMax, scheduleLoopDuration);

			// Handle statistics
			if (REFITTime.currentTimeMillis.getAsLong() < (epochStartTimestamp + STATISTICS_INTERVAL)) continue;
			StringBuilder sb = new StringBuilder(45 + 38 + 82 * tasks.length);
			sb.append(String.format("[STAGE] Epoch %04d: %6.0f (%10d)\n", epochCounter, selectDurations / (float) epochCounter, selectDurationsMax));
			sb.append(String.format("[STAGE] Iteration max: (%10d)\n", scheduleLoopMax));
			for (int i = 0; i < tasks.length; i++) {
				if (executions[i] == 0) continue;
				sb.append(String.format(
						"[STAGE]  %-10s | %6d | %7.0f (%10d) | %12d %12d | \n",
						tasks[i],
						executions[i],
						executionDurations[i] / (float) executions[i],
						executionDurationsMax[i],
						throughputs[i],
						totalThroughputs[i]));
			}
			REFITLogger.logDebug(this, sb.toString());

			// Reset statistics
			epochCounter = 0;
			epochStartTimestamp = REFITTime.currentTimeMillis.getAsLong();
			selectDurations = 0L;
			selectDurationsMax = 0L;
			executions = new int[tasks.length];
			executionDurations = new long[tasks.length];
			executionDurationsMax = new long[tasks.length];
			throughputs = new long[tasks.length];
			scheduleLoopMax = 0L;
		}
		if (watchdog != null) watchdog.disable();
	}

	private void selectReadyTasks() {
		// timeout of zero blocks indefinitely
		long timeout = 0;
		if (!taskTimeouts.isEmpty()) {
			timeout = Math.max(0, taskTimeouts.firstKey() - REFITTime.currentTimeMillis.getAsLong());
		}
		// Block until at least one task becomes runnable
		try {
			if (!taskTimeouts.isEmpty() && timeout == 0 || queueSize > 0) {
				selector.selectNow();
			} else {
				selector.select(Math.max(0, timeout));
			}
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return;
		}

		// Collect tasks that are able to perform I/O operations
		for (SelectionKey key : selector.selectedKeys()) {
			REFITSchedulerTask task = (REFITSchedulerTask) key.attachment();
			if (REFITConfig.ENABLE_DEBUG_CHECKS && (task.taskID() >= tasks.length || tasks[task.taskID()] != task)) {
				throw new IllegalArgumentException("Got task from different scheduler " + task.toString());
			}
			if (key.readyOps() != SelectionKey.OP_READ) {
				// always process non-read operations immediately to avoid blocking our output
				task.selectKey(key);
				readyTasks[task.taskID()] = true;
			} else if (!isQueued[task.taskID()]) {
				queues.get(task.taskType).add(key);
				isQueued[task.taskID()] = true;
				queueSize++;
			}
		}
		selector.selectedKeys().clear();

		// Pick only a limited number of tasks per taskType
		for (ArrayDeque<SelectionKey> queue : queues.values()) {
			for (int i = 0; i < TASK_TYPE_SUB_ITERATION_LIMIT && !queue.isEmpty(); i++) {
				SelectionKey key = queue.removeFirst();
				// the queue might have stored keys that were canceled in the meantime
				if (!key.isValid()) {
					i--;
					continue;
				}

				REFITSchedulerTask task = (REFITSchedulerTask) key.attachment();
				task.selectKey(key);
				readyTasks[task.taskID()] = true;
				isQueued[task.taskID()] = false;
				queueSize--;
			}
		}

		// Collect timed out tasks
		long timeMillis = REFITTime.currentTimeMillis.getAsLong();
		while (!taskTimeouts.isEmpty() && taskTimeouts.firstKey() <= timeMillis) {
			LinkedList<REFITSchedulerTask> timedOutTasks = taskTimeouts.pollFirstEntry().getValue();
			for (REFITSchedulerTask task : timedOutTasks) {
				readyTasks[task.taskID()] = true;
			}
		}
	}

	private void subscribeToGCEvents() {
		for (GarbageCollectorMXBean bean : ManagementFactory.getGarbageCollectorMXBeans()) {
			try {
				ManagementFactory.getPlatformMBeanServer().addNotificationListener(bean.getObjectName(), (notification, handback) -> {
					if (notification.getType().equals(GarbageCollectionNotificationInfo.GARBAGE_COLLECTION_NOTIFICATION)) {
						CompositeData cd = (CompositeData) notification.getUserData();
						GarbageCollectionNotificationInfo gcNotificationInfo = GarbageCollectionNotificationInfo.from(cd);
						GcInfo gcInfo = gcNotificationInfo.getGcInfo();
						lastGCDuration = gcInfo.getDuration();
						lastGCEndTime = gcInfo.getEndTime();
//						REFITLogger.logWarning(this, notification.getType() + " end " + lastGCEndTime);
					}
				}, null, null);
			} catch (InstanceNotFoundException e) {
				throw new InternalError(e);
			}
		}
	}


	private static class SlowTask {
		public final long durationInNs;
		public final long endTimeInMs;
		public final String msg;

		public SlowTask(long durationInNs, long endTimeInMs, String msg) {
			this.durationInNs = durationInNs;
			this.endTimeInMs = endTimeInMs;
			this.msg = msg;
		}
	}
}
