package refit.test;

import java.nio.ByteBuffer;

import refit.message.*;

public class SignatureMicrobenchmark {

	public static void main(String[] args) {
		// ed25519 got approx 40k sign/s, 20k verify/s, 64 byte signatures
		// using implementation from https://github.com/str4d/ed25519-java/releases/tag/v0.3.0
		//
		// rsa 1024-bit got 3.8k sign/s,  52k verify/s, 128 byte signatures
		// rsa 2048-bit got 0.7k sign/s, 18k verify/s, 256 byte signatures
		// rsa 3072-bit got 0.2k sign/s,  9k verify/s, 384 byte signatures -> similar security level to ed25519

		REFITMessageAuthentication mac = new REFITMessageAuthentication((short) 1);

		long mainStart = System.currentTimeMillis();
		long duration = 140 * 1000;

		while (System.currentTimeMillis() - mainStart < duration) {
			final long iterations = 1000;
			long startTime = System.nanoTime();
			SigRequest message = null;
			for (int i = 0; i < iterations; i++) {
				message = new SigRequest(new REFITUniqueID((short) 1, 1), (short) 1, ByteBuffer.allocate(0));
				message.serializeMessage();
				mac.appendSignature(message);
			}
			long midTime = System.nanoTime();
			for (int i = 0; i < iterations; i++) {
				REFITMessage decoded = new SigRequest(message.getBuffer());
				if (!mac.verifySignature(decoded, REFITMessageAuthentication.ANY_SENDER)) {
					throw new InternalError("Broken signatures!");
				}
			}
			long endTime = System.nanoTime();
			System.out.println((long) (iterations / ((midTime - startTime) / 1e9)) + " Sign / s");
			System.out.println((long) (iterations / ((endTime - midTime) / 1e9)) + " Verify / s");
		}
	}

	private static class SigRequest extends REFITOrderRequest {
		private static final int REQUEST_HEADER_SIZE = Short.BYTES;
		private ByteBuffer payload;

		public SigRequest(REFITUniqueID uid, short from, ByteBuffer payload) {
			super(REFITMessageType.REQUEST, uid, from);
			this.payload = payload;
		}

		public SigRequest(ByteBuffer buffer) {
			super(REFITMessageType.REQUEST, buffer);
			short payloadSize = buffer.getShort();
			this.payload = buffer.slice();
			this.payload.limit(payloadSize);
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.SIGNATURE;
		}

		@Override
		public boolean equals(Object object) {
			if (this == object) return true;
			if (object == null || getClass() != object.getClass()) return false;
			if (!super.equals(object)) return false;
			SigRequest other = (SigRequest) object;
			return payload.equals(other.payload);
		}

		@Override
		protected void serialize(ByteBuffer buffer) {
			super.serialize(buffer);
			int payloadSize = payload.remaining();
			buffer.putShort((short) payloadSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newPayload = buffer.slice();
			buffer.put(payload);
			newPayload.limit(payloadSize);
			payload = newPayload;
		}

		@Override
		protected int getHeaderSize() {
			return (super.getHeaderSize() + REQUEST_HEADER_SIZE);
		}

		@Override
		protected int calculatePayloadSize() {
			return (super.calculatePayloadSize() + REQUEST_HEADER_SIZE + payload.remaining());
		}
	}
}
