package threshsig;

import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Random;

/**
 * A Secret Key Share for an RSA (k,l) Threshold Scheme
 * <p>
 * Reference: "Practical Threshold Signatures",<br>
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99<BR>
 *
 * @author Steve Weis <sweis@mit.edu>
 */
// Modification by Michael Eischer: Optimize signature calculation and give more access to internal variables
public class KeyShare {
	// Constants and variables
	//............................................................................
	/**
	 * Secret key value
	 */
	private BigInteger secret;
	/**
	 * Verifier used to authenticate self to other shares
	 */
	private BigInteger verifier;
	private BigInteger groupVerifier;

	private byte[] verifiern;
	private byte[] groupVerifiern;

	// TODO: This information should be moved to a GroupKey object
	// It's redundant to put into every share
	private BigInteger n;
	private final BigInteger delta;
	private final BigInteger fourdelta;

	/**
	 * The secret key value used to sign messages. Should not be exposed
	 */
	private BigInteger signVal;
	// TODO: Maybe get a better id scheme. Use verifier?
	private int id;

	// TODO-Rand: This is klunky.
	// Maybe use a utility class with a RNG for everyone
	private static SecureRandom random;
	private MessageDigest md;

	static {
		final byte[] randSeed = new byte[20];
		(new Random()).nextBytes(randSeed);
		random = new SecureRandom(randSeed);
	}

	// Constructors
	//............................................................................

	/**
	 * Create a new share
	 *
	 * @param id     - the identifier of this share
	 * @param secret - a secret value generated by a Dealer
	 * @param n      - the modulo of the group public key
	 * @param delta  - l! (group size factorial)
	 */
	public KeyShare(final int id, final BigInteger secret, final BigInteger n, final BigInteger delta) {
		this.id = id;
		this.secret = secret;

		// Verifier must be set later using setVerifier()
		// TODO: Merge Dealer.generateShares and Dealer.generateVerifiers
		// and generate them simultaneously
		verifier = null;

		this.n = n;
		this.delta = delta;
		this.fourdelta = ThreshUtil.FOUR.multiply(delta);
		this.signVal = fourdelta.multiply(secret);
		try {
			md = MessageDigest.getInstance("SHA");
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			throw new InternalError();
		}
	}

	// Public Methods
	//............................................................................

	public int getId() {
		return id;
	}

	public BigInteger getDelta() {
		return delta;
	}

	public BigInteger getSecret() {
		return secret;
	}

	public void setVerifiers(final BigInteger verifier, final BigInteger groupVerifier) {
		this.verifier = verifier;
		this.groupVerifier = groupVerifier;

		this.verifiern = verifier.mod(n).toByteArray();
		this.groupVerifiern = groupVerifier.mod(n).toByteArray();
	}

	public BigInteger getVerifier() {
		return verifier;
	}

	public BigInteger getGroupVerifier() {
		return groupVerifier;
	}

	public BigInteger getSignVal() {
		return signVal;
	}

	@Override
	public String toString() {
		return "KeyShare[" + id + "]";
	}

	/**
	 * Create a SigShare and a Verifier for byte[] b<BR>
	 * <p>
	 * Refer to Shoup pg. 8 <BR>
	 *
	 * @param b The array of bytes to produce a signature share for.
	 * @return a sig share with a verifier
	 */
	public SigShare sign(final byte[] b) {
		final BigInteger x = (new BigInteger(b)).mod(n);

		final int randbits = n.bitLength() + 2 * ThreshUtil.L1;

		// r \elt (0, 2^L(n)+2*l1)
		final BigInteger r = (new BigInteger(randbits, random));
		final BigInteger vprime = groupVerifier.modPow(r, n);
		final BigInteger xtilde = x.modPow(fourdelta, n);
		final BigInteger xprime = xtilde.modPow(r, n);

		// Try to generate C and Z
		BigInteger xsignValn = x.modPow(signVal, n);
		md.reset();

		// debug("v: " + groupVerifier.mod(n));
		md.update(groupVerifiern);

		// debug("xtilde: " + xtilde);
		md.update(xtilde.toByteArray());

		// debug("vi: " + verifier.mod(n));
		md.update(verifiern);

		// debug("xi^2: " + x.modPow(signVal,n).modPow(TWO,n));
		md.update(xsignValn.modPow(ThreshUtil.TWO, n).toByteArray());

		// debug("v': "+ vprime);
		md.update(vprime.toByteArray());

		// debug("x': " + xprime);
		md.update(xprime.toByteArray());
		BigInteger c = new BigInteger(md.digest()).mod(n);
		BigInteger z = (c.multiply(secret)).add(r);

		final Verifier ver = new Verifier(z, c, verifier, groupVerifier);

		return new SigShare(this.id, xsignValn, ver);
	}

	// Debugging
	//............................................................................
	private static void debug(final String s) {
		System.err.println("KeyShare: " + s);
	}
}
