package refit.application.rezk;

import java.nio.ByteBuffer;
import java.util.Arrays;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import refit.application.REFITApplicationServer;
import refit.config.REFITConfig;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointContainer;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertFalse;

@RunWith(Parameterized.class)
public class REZKFuzzyCheckpointTest extends REZKFuzzyTest {
	private REFITApplicationServer application;
	private int requestSeqNr;

	@Override
	public void setUp() throws Exception {
		if (!REFITConfig.UPDATE_BASED_CHECKPOINTING) {
			throw new AssertionError("This test requires updates!");
		}
		application = new REZKTestServerSqlite();
		application.init();
		app = application;
		requestSeqNr = prepareApplication();
	}

	@Test
	public void testMergeOnce() {
		System.out.println("Start " + description);

		System.out.println("Initial checkpoint");
		REZKFuzzyCheckpointContainer initialFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();
		ByteBuffer updateBytes = executeRequest(requestSeqNr, true);

		System.out.println("After request checkpoint");
		REZKFuzzyCheckpointContainer completeFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();

		assertFalse(Arrays.equals(initialFuzzyCheckpoint.serialize(), completeFuzzyCheckpoint.serialize()));

		System.out.println("Once merged checkpoint");
		initialFuzzyCheckpoint.mergeUpdate(updateBytes);
		REFITCheckpointObject[] initialCheckpointBytes = initialFuzzyCheckpoint.serialize();

		System.out.println("Deserialized checkpoint");
		REZKFuzzyCheckpointContainer deserializedCheckpoint = new REZKFuzzyCheckpointContainer(initialCheckpointBytes);

		REFITCheckpointObject[] serialize = completeFuzzyCheckpoint.serialize();
		assertArrayEquals(serialize, initialCheckpointBytes);
	}

	@Test
	public void testMergeTwice() {
		System.out.println("Start " + description);

		System.out.println("Initial checkpoint");
		REZKFuzzyCheckpointContainer initialFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();
		ByteBuffer updateBytes = executeRequest(requestSeqNr, true);

		System.out.println("After request checkpoint");
		REZKFuzzyCheckpointContainer completeFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();

		assertFalse(Arrays.equals(initialFuzzyCheckpoint.serialize(), completeFuzzyCheckpoint.serialize()));

		System.out.println("Twice merged checkpoint");
		for (int i = 0; i < 2; i++) {
			initialFuzzyCheckpoint.mergeUpdate(updateBytes);
		}
		REFITCheckpointObject[] initialCheckpointBytes = initialFuzzyCheckpoint.serialize();

		System.out.println("Deserialized checkpoint");
		REZKFuzzyCheckpointContainer deserializedCheckpoint = new REZKFuzzyCheckpointContainer(initialCheckpointBytes);

		assertArrayEquals(completeFuzzyCheckpoint.serialize(), initialCheckpointBytes);
	}

	@Test
	public void testMergeAfterExecute() {
		System.out.println("Start " + description);
		ByteBuffer updateBytes = executeRequest(requestSeqNr, true);

		System.out.println("After request checkpoint");
		REZKFuzzyCheckpointContainer completeFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();
		REZKFuzzyCheckpointContainer mergedAfterExecuteFuzzyCheckpoint = (REZKFuzzyCheckpointContainer) application.createFuzzyCheckpoint();

		System.out.println("Twice merged checkpoint");
		mergedAfterExecuteFuzzyCheckpoint.mergeUpdate(updateBytes);

		System.out.println("Deserialized checkpoint");
		REFITCheckpointObject[] mergeAfterExecuteCheckpointBytes = mergedAfterExecuteFuzzyCheckpoint.serialize();
		REZKFuzzyCheckpointContainer deserializedCheckpoint = new REZKFuzzyCheckpointContainer(mergeAfterExecuteCheckpointBytes);

		assertArrayEquals(completeFuzzyCheckpoint.serialize(), mergeAfterExecuteCheckpointBytes);
	}

	@Test
	public void testCheckpointAndApply() {
		System.out.println("Start " + description);
		executeRequest(requestSeqNr, true);

		System.out.println("After request checkpoint");
		REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> checkpoint = application.createFuzzyCheckpoint();
		application.applyCheckpoint(checkpoint.serialize());

		System.out.println("After apply checkpoint");
		REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> appliedCheckpoint = application.createFuzzyCheckpoint();

		assertArrayEquals(checkpoint.serialize(), appliedCheckpoint.serialize());
	}

	@Test
	public void testNormalCheckpointAndApply() {
		System.out.println("Start " + description);
		executeRequest(requestSeqNr, true);

		System.out.println("After request checkpoint");
		REFITCheckpointObject[] checkpoint = application.createCheckpoint();
		application.applyCheckpoint(checkpoint);

		System.out.println("After apply checkpoint");
		REFITCheckpointObject[] appliedCheckpoint = application.createCheckpoint();

		assertArrayEquals(checkpoint, appliedCheckpoint);
	}

	@Test
	public void testMixedCheckpointAndApply() {
		System.out.println("Start " + description);
		executeRequest(requestSeqNr, true);

		// A fuzzy checkpoint has sorted objects avoid confusion
		application.applyCheckpoint(application.createFuzzyCheckpoint().serialize());

		System.out.println("After request checkpoint");
		REFITCheckpointObject[] checkpoint = application.createCheckpoint();
		REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> fuzzyCheckpoint = application.createFuzzyCheckpoint();
		application.applyCheckpoint(fuzzyCheckpoint.serialize());

		System.out.println("After apply checkpoint");
		REFITCheckpointObject[] appliedCheckpoint = application.createCheckpoint();
		REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> appliedFuzzyCheckpoint = application.createFuzzyCheckpoint();

		assertArrayEquals(checkpoint, appliedCheckpoint);
		assertArrayEquals(fuzzyCheckpoint.serialize(), appliedFuzzyCheckpoint.serialize());
	}
}
