package refit.application.rezk;

import java.util.Arrays;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import refit.application.rezk.common.REZKException;
import refit.application.rezk.common.REZKStat;
import refit.application.rezk.datatree.REZKCachedDataTree;
import refit.application.rezk.datatree.REZKDataTree;
import refit.application.rezk.datatree.REZKInMemoryDataTree;
import refit.application.rezk.datatree.REZKSQLiteDataTree;
import refit.client.REFITLocalClientLibrary;
import refit.scheduler.REFITSchedulerGroup;
import refit.util.REFITIntervalStatistics;
import refit.util.REFITTime;


@RunWith(Parameterized.class)
public class REZKTest {
	@Parameters
	public static Iterable<?> data() {
		// Create data trees
		REZKDataTree[] trees = new REZKDataTree[]{
				new REZKInMemoryDataTree(),
				new REZKSQLiteDataTree("/tmp/rezk.db"),
				new REZKCachedDataTree(new REZKSQLiteDataTree("/tmp/rezk.db")),
//				new REZKMySQLDataTree("refit"),
//				new REZKCachedDataTree(new REZKMySQLDataTree("refit"))
		};
		return Arrays.asList(trees);
	}

	@SuppressWarnings("WeakerAccess")
	@Parameter
	public REZKDataTree dataTree;

	@Test
	public void test() throws Exception {
		// Evaluate different data sizes
		for (int dataLength = 0; dataLength <= REZKServer.MAXIMUM_DATA_LENGTH; dataLength += 256) {
			System.out.println("\n---------- " + dataTree.getClass().getSimpleName() + " (" + dataLength + ") ----------");
			System.out.println("[#NDs SIZE]    Operation      ops/s   op in us");
			REZKServer server = new REZKServer(dataTree);
			server.init();
			REFITLocalClientLibrary connection = new REFITLocalClientLibrary(server, null,
					new REFITSchedulerGroup(false),
					new REFITIntervalStatistics(1000, null, null),
					null, true, null);
			REZKClient service = new REZKClient(connection);
			runBenchmark(service, server, 30000, 10, dataLength);
			server.shutdown();
		}
	}


	private static void printResult(String command, int runs, int nrOfNodes, int dataLength, long duration) {
		System.out.println(String.format("[%3d %5d] %12s %10d %5.3f", nrOfNodes, dataLength, command,
				(int) (runs / (duration * 1e-9f)), duration * 1e-3f / runs));
	}

	private static void runBenchmark(REZKClient service, REZKServer server, int runs, int nrOfNodes, int dataLength) throws REZKException {
		// Preparations
		REZKStat stat = new REZKStat();
		byte[] data = new byte[dataLength];
		String[] paths = new String[nrOfNodes];
		for (int i = 0; i < paths.length; i++) paths[i] = String.format("%s%05d", "/node", i);
		String counter = "/counter";
		String queue = "/queue";

		// Create
		long duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.create("/n", data, true);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("CREATE", runs, nrOfNodes, dataLength, duration);

		// Delete
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.delete(String.format("/n%010d", i));
		duration += REFITTime.nanoTime.getAsLong();
		printResult("DELETE", runs, nrOfNodes, dataLength, duration);

		// Create nodes for the rest of the benchmark
		for (int i = 0; i < nrOfNodes; i++) service.create(paths[i], data, false);
		service.create(counter, new byte[Integer.BYTES], false);
		service.create(queue, null, false);

		// Exists
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.exists(paths[i % nrOfNodes], false, true);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("EXISTS", runs, nrOfNodes, dataLength, duration);

		// Get children
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.getChildren("/", false, stat, true);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("GET_CHILDREN", runs, nrOfNodes, dataLength, duration);

		// Get data
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.getData(paths[i % nrOfNodes], false, stat, true);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("GET_DATA", runs, nrOfNodes, dataLength, duration);

		// Set data
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.setData(paths[i % nrOfNodes], data, -1);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("SET_DATA", runs, nrOfNodes, dataLength, duration);

		// Increment counter
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) service.increment(counter);
		duration += REFITTime.nanoTime.getAsLong();
		printResult("INCREMENT", runs, nrOfNodes, dataLength, duration);

		// Queue
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) {
			service.create(queue + "/element", data, true);
			service.dequeue(queue, stat);
		}
		duration += REFITTime.nanoTime.getAsLong();
		printResult("QUEUE", runs, nrOfNodes, dataLength, duration);

		// Checkpoints
		duration = -REFITTime.nanoTime.getAsLong();
		for (int i = 0; i < runs; i++) {
			server.createCheckpoint();
		}
		duration += REFITTime.nanoTime.getAsLong();
		printResult("CHECKPOINT", runs, nrOfNodes, dataLength, duration);

		// Delete nodes
		for (int i = 0; i < nrOfNodes; i++) service.delete(paths[i]);
		service.delete(counter);
		service.delete(queue);
	}

}
