package refit.application.rezk.common;

import java.nio.ByteBuffer;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import org.junit.Test;
import refit.application.rezk.util.REFITByteString;

import static org.junit.Assert.*;
import static refit.application.rezk.common.REZKStatTest.assertStatEquals;


public class REZKReplyTest {
	private static <T> void assertCollectionEquals(Collection<T> a, Collection<T> b) {
		if (a == null && b == null) return;
		assertFalse(a == null || b == null);

		Iterator<T> iteratorA = a.iterator();
		Iterator<T> iteratorB = b.iterator();
		while (iteratorA.hasNext()) {
			assertTrue(iteratorB.hasNext());
			assertEquals(iteratorA.next(), iteratorB.next());
		}
		assertFalse(iteratorB.hasNext());
	}

	private static void assertReplyEquals(REZKReply a, REZKReply b) {
		assertEquals(a.operation, b.operation);
		assertEquals(a.message, b.message);
		if (a.message != null) return;

		assertArrayEquals(a.data, b.data);
		assertStatEquals(a.stat, b.stat);
		assertEquals(a.path, b.path);
		assertEquals(a.exists, b.exists);
		assertCollectionEquals(a.children, b.children);
	}

	private static void assertCopy(REZKReply reply) {
		reply.serializeMessage();

		ByteBuffer buffer = reply.getBuffer();
		REZKReply replyMod = new REZKReply(buffer);
		assertFalse(buffer.hasRemaining());
		assertReplyEquals(reply, replyMod);
	}

	@Test
	public void message() {
		REZKReply reply = new REZKReply(REZKOperation.CREATE);
		reply.setException(new REZKException("Hello!"));
		assertCopy(reply);
	}

	@Test
	public void create() {
		REZKReply reply = new REZKReply(REZKOperation.CREATE);
		reply.setPath(new REFITByteString("test"));
		assertCopy(reply);
	}

	@Test
	public void delete() {
		REZKReply reply = new REZKReply(REZKOperation.DELETE);
		assertCopy(reply);
	}

	@Test
	public void exists() {
		REZKReply reply = new REZKReply(REZKOperation.EXISTS);
		reply.setExists(true);
		assertCopy(reply);
	}

	private static REZKStat getExampleStat() {
		REZKStat stat = new REZKStat();
		stat.setDataLength(4);
		stat.setNrOfChildren(2);
		stat.setVersion(42);
		return stat;
	}

	@Test
	public void getChildren() {
		REZKReply reply = new REZKReply(REZKOperation.GET_CHILDREN);
		reply.setStat(getExampleStat());
		HashSet<REFITByteString> children = new HashSet<>();
		children.add(new REFITByteString("abc"));
		children.add(new REFITByteString("foo"));
		children.add(new REFITByteString("bar"));
		reply.setChildren(children);
		assertCopy(reply);
	}

	@Test
	public void getData() {
		REZKOperation[] operations = new REZKOperation[]{REZKOperation.GET_DATA, REZKOperation.DEQUEUE};
		for (REZKOperation op : operations) {
			REZKReply reply = new REZKReply(op);
			reply.setStat(getExampleStat());
			reply.setData(new byte[]{17, 18, 19});
			assertCopy(reply);
		}
	}

	@Test
	public void setData() {
		REZKReply reply = new REZKReply(REZKOperation.SET_DATA);
		reply.setStat(getExampleStat());
		assertCopy(reply);
	}

	@Test
	public void increment() {
		REZKReply reply = new REZKReply(REZKOperation.INCREMENT);
		reply.setData(new byte[]{0, 1, 2, 3, 4});
		assertCopy(reply);
	}
}
