package refit.replica.multichannel;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.junit.BeforeClass;
import org.junit.Test;
import refit.client.REFITClientLibrary.RequestMode;
import refit.message.*;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertTrue;

public class REFITShareBundleTest {

	private static final int STABILITY_THRESHOLD = 1 + 1;
	private static final short REPLICA_ID = 2;

	private static REFITMessageAuthentication mac;

	@BeforeClass
	public static void before() {
		mac = new REFITMessageAuthentication(REPLICA_ID);
	}

	private static void assertBundleEquals(REFITShareBundle a, REFITShareBundle b) {
		MessageHelper.assertBasicsEquals(a, b);
		assertArrayEquals(a.getProof(), b.getProof());
	}

	private static void assertCopy(REFITShareBundle bundle) {
		REFITShareBundle bundleCopy = MessageHelper.copyMessage(bundle, REFITShareBundle::new);
		assertBundleEquals(bundle, bundleCopy);
		assertValidProofs(bundleCopy);
	}

	private static void assertValidProofs(REFITShareBundle bundle) {
		REFITShare[] proofs = bundle.getProof();
		for (REFITShare proof : proofs) {
			assertTrue(mac.verifySignature(proof, REFITMessageAuthentication.ANY_SENDER));
		}
	}

	@Test
	public void regularShareSerialization() {
		REFITMessage msg = createRequest(27, 1, false);
		REFITShareBundle bundle = createShareBundle(msg);
		assertCopy(bundle);
		assertValidProofs(bundle);
	}

	private static REFITRequest createRequest(long seqNr, int groupID, boolean isReadOnly) {
		REFITRequest request = new REFITRequest(new REFITUniqueID((short) 1, seqNr), REPLICA_ID, (short) groupID,
				-1, ByteBuffer.wrap(new byte[17]), (short) 0, isReadOnly ? RequestMode.READ : RequestMode.WRITE);
		request.serializeMessage();
		return request;
	}

	private static REFITShareBundle createShareBundle(REFITMessage msg) {
		REFITShare[] proof = new REFITShare[STABILITY_THRESHOLD];
		for (int i = 0; i < proof.length; i++) {
			proof[i] = createShare(15, msg);
		}

		return new REFITShareBundle(15, (short) 0, (short) 8, msg, proof);
	}

	private static REFITShare createShare(long seqNr, REFITMessage message) {
		REFITShare share = new REFITShare(REFITMessageType.CHANNEL_SHARE_PART, seqNr, (short) 0, REPLICA_ID, message);
		share.serializeMessage();
		mac.appendSignature(share);
		return share;
	}

	@Test
	public void noopShareSerialization() {
		REFITMessage msg = REFITNoOpRequest.INSTANCE;
		REFITShareBundle bundle = createShareBundle(msg);
		assertValidProofs(bundle);
		assertCopy(bundle);
	}

	@Test
	public void filteredBundle() {
		List<REFITRequest> requests = new ArrayList<>();
		requests.add(createRequest(41, 1, false));
		requests.add(createRequest(42, 0, false));
		requests.add(createRequest(43, 0, true));

		REFITBatch batch = new REFITBatch(43, (short) 1, requests);
		batch.serializeMessage();

		REFITShareBundle bundle = createShareBundle(batch);
		assertCopy(bundle);
		assertValidProofs(bundle);

		REFITShareBundle filteredBundle = bundle.toFilteredBundle((msg) -> ((REFITBatch) msg).toGroupOnly((short) 1));
		assertCopy(filteredBundle);
		assertValidProofs(filteredBundle);

		assertArrayEquals(bundle.getContent().getHash(), filteredBundle.getContent().getHash());

		MessageHelper.assertHeaderEquals(bundle, filteredBundle);
		assertArrayEquals(bundle.getHash(), filteredBundle.getHash());
	}
}
