package threshsig.testing;

import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;
import threshsig.*;

import static org.junit.Assert.*;

// Modification by Michael Eischer: Modernize / fix tests, add test for signature combination and signature copying
public class ThreshTest {
	private static final int KEYSIZE = 1024;
	private static final int K = 11;
	private static final int L = 16;
	private static Dealer d;
	private static GroupKey gk;
	private static KeyShare[] keys;
	private static final byte[] data = new byte[1024];
	private static byte[] b;
	private static final SigShare[] sigs = new SigShare[K];

	@Before
	public final void setUp() throws NoSuchAlgorithmException {
		(new Random()).nextBytes(data);
		final MessageDigest md = MessageDigest.getInstance("SHA-1");
		b = md.digest(data);
	}

	@Before
	public final void testInitDealer() {
		// Initialize a dealer with a keysize
		d = new Dealer(KEYSIZE);

		final long start = System.currentTimeMillis();
		long elapsed;
		// Generate a set of key shares
		d.generateKeys(K, L);

		elapsed = System.currentTimeMillis() - start;
		System.out.println("\tKey Gen total (ms): " + elapsed);

		// This is the group key common to all shares, which
		// is not assumed to be trusted. Treat like a Public Key
		gk = d.getGroupKey();

		// The Dealer has the shares and is assumed trusted
		// This should be destroyed, unless you want to reuse the
		// Special Primes of the group key to generate a new set of
		// shares
		keys = d.getShares();

		Arrays.fill(sigs, null);
	}

	@Test
	public void testVerifySignatures() {
		System.out.println("Attempting to verify a valid set of signatures...");
		// Pick a set of shares to attempt to verify
		// These are the indices of the shares
		final int[] S = {3, 5, 1, 2, 10, 7, 4, 6, 8, 14, 15};

		for (int i = 0; i < S.length; i++)
			sigs[i] = keys[S[i]].sign(b);

		assertTrue(SigShare
				.verify(b, sigs, K, L, gk.getModulus(), gk.getExponent()));
	}

	@Test
	public void testVerifySignaturesAgain() {
		System.out.println("Attempting to verify a different set of shares...");

		// Create k sigs to verify using different keys
		final int[] T = {8, 9, 7, 6, 1, 12, 4, 2, 3, 14, 15};
		for (int i = 0; i < K; i++)
			sigs[i] = keys[T[i]].sign(b);

		assertTrue(SigShare
				.verify(b, sigs, K, L, gk.getModulus(), gk.getExponent()));
	}

	@Test
	public void testVerifyBadSignature() {
		// Create k sigs to verify using different keys
		final int[] T = {8, 9, 7, 6, 1, 12, 5, 2, 4, 14, 15};
		for (int i = 0; i < K; i++)
			sigs[i] = keys[T[i]].sign(b);

		b = "corrupt data".getBytes();
		sigs[3] = keys[3].sign(b);
		assertFalse(SigShare.verify(b, sigs, K, L, gk.getModulus(), gk
				.getExponent()));
	}

	@Test
	public void testPerformance() {
		for (int z = 0; z < 5; z++) {
			final int RUNS = 100;
			final int[] S = {3, 5, 1, 2, 10, 7, 4, 6, 8, 14, 15};

			long start = System.currentTimeMillis(), elapsed;
			for (int i = 0; i < RUNS; i++)
				sigs[i % K] = keys[S[i % K]].sign(b);
			elapsed = System.currentTimeMillis() - start;
			System.out.println("Signing total (" + RUNS + " sigs) (ms): " + elapsed
					+ " Average: " + ((float) elapsed / RUNS));

			for (int i = 0; i < K; i++)
				sigs[i] = keys[S[i]].sign(b);

			start = System.currentTimeMillis();
			for (int i = 0; i < RUNS; i++)
				SigShare.combine(b, sigs, gk.getK(), gk.getL(), gk.getModulus(), gk.getExponent());
			elapsed = System.currentTimeMillis() - start;
			System.out.println("Combine total (" + RUNS + " sigs) (ms): "
					+ elapsed + " Average: " + ((float) elapsed / RUNS));

			start = System.currentTimeMillis();
			for (int i = 0; i < RUNS; i++)
				if (!SigShare.verify(b, sigs, K, L, gk.getModulus(), gk.getExponent()))
					System.out.println("Sig Failed to verify correctly");
			elapsed = System.currentTimeMillis() - start;
			System.out.println("Verification total (" + RUNS + " sigs) (ms): "
					+ elapsed + " Average: " + ((float) elapsed / RUNS / K));
		}
	}

	@Test
	public void testCombinedSignature() {
		System.out.println("Attempting to combine and verify a valid set of signatures...");
		// Pick a set of shares to attempt to verify
		// These are the indices of the shares
		final int[] S = {3, 5, 1, 2, 10, 7, 4, 6, 8, 14, 15};

		for (int i = 0; i < S.length; i++)
			sigs[i] = keys[S[i]].sign(b);

		BigInteger combine = SigShare.combine(b, sigs, gk.getK(), gk.getL(), gk.getModulus(), gk.getExponent());
		assertNotNull(combine);

		assertTrue(SigShare.verifyCombined(b, combine, gk.getModulus(), gk.getExponent()));
	}

	@Test
	public void testVerifySignatureTransfer() {
		System.out.println("Attempting to verify a valid set of signatures...");
		// Pick a set of shares to attempt to verify
		// These are the indices of the shares
		final int[] S = {3, 5, 1, 2, 10, 7, 4, 6, 8, 14, 15};

		for (int i = 0; i < S.length; i++)
			sigs[i] = keys[S[i]].sign(b);

		GroupKey gk2 = new GroupKey(gk.getK(), gk.getL(), 0, null, gk.getExponent(), gk.getModulus());
		byte[] b2 = Arrays.copyOf(b, b.length);

		SigShare[] sigs2 = new SigShare[sigs.length];
		for (int i = 0; i < sigs.length; i++) {
			SigShare sig = sigs[i];
			Verifier sv = sig.getSigVerifier();
			KeyShare key = keys[sig.getId() - 1];
			Verifier sigv = new Verifier(sv.getZ(), sv.getC(), key.getVerifier(), key.getGroupVerifier());
			sigs2[i] = new SigShare(sig.getId(), new BigInteger(sig.getBytes()), sigv);
		}

		assertTrue(SigShare
				.verify(b2, sigs2, gk2.getK(), gk2.getL(), gk2.getModulus(), gk2.getExponent()));
	}

}
