package refit.replica.fuzzycheckpoint;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import refit.application.REFITApplicationServer;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification.REFITFuzzyCheckpointFinishNotification;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification.REFITFuzzyCheckpointStartNotification;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification.REFITFuzzyCollectionDoneNotification;
import refit.replica.update.REFITBatchUpdate;
import refit.replica.update.REFITUpdate;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;

public class REFITFuzzyCheckpointStage extends REFITStage {
	public final REFITObserver<REFITFuzzyCheckpointNotification> fuzzyCheckpointNotificationSender;
	public final REFITObserver<REFITCheckpointState> checkpointSender;

	private final REFITApplicationServer application;
	private final short replicaID;
	private boolean fuzzyActive = false;
	private HashMap<Object, REFITCheckpointObject> updateChangeList = new HashMap<>();

	private REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> fcContainer;

	public REFITFuzzyCheckpointStage(REFITBaseReplica replica, REFITApplicationServer application) {
		super(REFITSchedulerTaskType.FUZZY_CHECKPOINT_STAGE, replica);

		replicaID = replica.id;

		this.fuzzyCheckpointNotificationSender = new REFITObserver<>();
		this.checkpointSender = new REFITObserver<>();

		this.application = application;
	}

	@Override
	public String toString() {
		return "FCP";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case UPDATE:
			handleUpdate((REFITUpdate) message);
			break;
		case BATCH_UPDATE:
			handleBatchUpdate((REFITBatchUpdate) message);
			break;
		case FUZZY_CHECKPOINT_NOTIFICATION:
			handleNotification((REFITFuzzyCheckpointNotification) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleUpdate(REFITUpdate update) {
		if (fcContainer == null) {
			return;
		}
		if (fuzzyActive) {
			fcContainer.mergeUpdate(update.getPayload());
		} else {
			fcContainer.mergeUpdateMarker(updateChangeList, update.getPayload());
		}
	}

	private void handleBatchUpdate(REFITBatchUpdate batchUpdate) {
		for (REFITUpdate update : batchUpdate.updates) {
			handleUpdate(update);
		}
	}

	private void handleNotification(REFITFuzzyCheckpointNotification notification) {
		switch (notification.notificationType) {
		case START:
			handleStartNotification((REFITFuzzyCheckpointStartNotification) notification);
			break;
		case FINISH:
			handleFinishNotification((REFITFuzzyCheckpointFinishNotification) notification);
			break;
		default:
			REFITLogger.logError(this, "drop notification of unexpected type " + notification.notificationType);
			break;
		}
	}

	private void handleStartNotification(REFITFuzzyCheckpointStartNotification notification) {
		REFITLogger.logFuzzyCheckpoint(this, "Fuzzy collection started @" + notification.startSeqNr);
		long start = REFITTime.currentTimeMillis.getAsLong();

		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || fcContainer == null) {
			REFITLogger.logFuzzyCheckpoint(this, "Collecting full checkpoint");
			fcContainer = application.createFuzzyCheckpoint();
		} else {
			REFITLogger.logFuzzyCheckpoint(this, "Collecting diff checkpoint");
			if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
				HashMap<Object, REFITCheckpointObject> changeList = this.updateChangeList;
				this.updateChangeList = new HashMap<>();

				Collection<REFITCheckpointObject> diffCheckpoint = application.createFuzzyUpdateCheckpoint(changeList);
				REFITLogger.logDebug(this, "Collected " + diffCheckpoint.size() + " objects of checkpoint data");
				for (REFITCheckpointObject object : diffCheckpoint) {
					fcContainer.mergeUpdate(object.getBuffer());
				}
			} else {
				Set<?> changelist = notification.changelist;
				Map<Object, REFITCheckpointObject> diffCheckpoint = application.createFuzzyDiffCheckpoint(changelist);
				REFITLogger.logDebug(this, "Collected " + diffCheckpoint.size() + " objects of checkpoint data");
				fcContainer.applyDiff(diffCheckpoint);
			}
		}
		fuzzyActive = true;
		fuzzyCheckpointNotificationSender.broadcast(new REFITFuzzyCollectionDoneNotification());

		REFITLogger.logFuzzyCheckpoint(this, "Collection done after " + (REFITTime.currentTimeMillis.getAsLong() - start) + "ms");
	}

	private void handleFinishNotification(REFITFuzzyCheckpointFinishNotification notification) {
		if (fcContainer == null) REFITLogger.logError(this, "Checkpoint end without start!");
		long lastSeqNr = notification.lastSeqNr;

		REFITLogger.logFuzzyCheckpoint(this, "Finishing checkpoint @" + lastSeqNr);
		long start = REFITTime.currentTimeMillis.getAsLong();

		if (notification.changedObjects != null) {
			fcContainer.applyDiff(notification.changedObjects);
		}
		REFITCheckpointObject[] checkpointData = fcContainer.serialize();

		// merge data arrays
		REFITCheckpointObject[] data = new REFITCheckpointObject[checkpointData.length + 2];
		data[0] = notification.metaData;
		data[1] = notification.sysAppCheckpoint;
		System.arraycopy(checkpointData, 0, data, 2, checkpointData.length);
		REFITCheckpointState checkpoint = new REFITCheckpointState(lastSeqNr, replicaID, false, data);
		checkpointSender.broadcast(checkpoint);

		long end = REFITTime.currentTimeMillis.getAsLong() - start;
		REFITLogger.logFuzzyCheckpoint(this, "Serializing and broadcasting done for " + data.length
				+ " objects of checkpoint @ " + lastSeqNr + " after " + end + "ms");

		if (!REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING) {
			fcContainer = null;
		}
		fuzzyActive = false;
	}
}
